using System;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.AI;
using TankyTank.Game;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// A base class for user-created AI bots.
	/// </summary>
	public class AIBase : IBot
	{
		#region Constructors

		/// <summary>
		/// Creates and initializes the base class for a new AI bot.
		/// </summary>
		/// <param name="game">The game currently being played.</param>
		public AIBase(IGame game)
		{
			if(game == null)
				throw new ArgumentNullException("game");

			_game = game;
		}

		#endregion
		#region Fields

		protected readonly IGame _game;
		protected readonly Serializer _ser;
		protected ITank _curTank;

		#endregion
		#region Public Methods

		/// <summary>
		/// Gets the tank input for the current frame. This method is required by the IBot interface.
		/// </summary>
		/// <param name="forward">Whether the tank is moving forward.</param>
		/// <param name="reverse">Whether the tank is moving backward.</param>
		/// <param name="left">Whether the tank is rotating left.</param>
		/// <param name="right">Whether the tank is rotating right.</param>
		/// <param name="fireMouseLeft">Whether the tank should fire its weapons bound to the left mouse button if possible.</param>
		/// <param name="fireMouseRight">Whether the tank should fire its weapons bound to the right mouse button if possible.</param>
		/// <param name="aimLoc">The location of the map at which the turret(s) are aimed.</param>
		public virtual void GetInput(ref bool forward, ref bool reverse, ref bool left, ref bool right, ref bool fireMouseLeft,
		                             ref bool fireMouseRight, ref Vector aimLoc)
		{
			// Do not change input.
			// Functionality should be overridden.
		}

		/// <summary>
		/// Called when this bot is spawned as the specified tank.
		/// </summary>
		public virtual void Spawn(ITank tank)
		{
			_curTank = tank;
		}

		/// <summary>
		/// Returns a string containing the debug information for this AI bot, or null to display no debug info.
		/// </summary>
		public virtual string GetDebugInfo()
		{
			string nameStr = "";
			if(_curTank != null)
			{
				string playerName = _game.Players[_curTank.Player].Name;
				nameStr = " \"" + playerName + "\"";
			}
			return GetType().Name + nameStr;
		}

		#endregion
	}
}