﻿using TankyTank.Game.Tanks;
using TankyTank.Game;
using Halfling.Geometry.Dim2;
using TankyTank.Game.Doodads;
using TankyTank.Game.Bullets;
using System;

namespace TankyTank.AIBots
{
	/// <summary>
	/// Provides utility methods for use with AI bots and behaviors.
	/// </summary>
	public static class AITools
	{
		#region Public Static Methods

		/// <summary>
		/// Returns whether the specified tank currently exists in the game.
		/// </summary>
		public static bool Exists(ITank tank, IGame game)
		{
			return game.Tanks.Exists(tank.UniqueID) && game.Tanks[tank.UniqueID] == tank;
		}

		/// <summary>
		/// Returns whether the specified bullet currently exists in the game.
		/// </summary>
		public static bool Exists(IBullet bullet, IGame game)
		{
			return game.Bullets.Exists(bullet.UniqueID) && game.Bullets[bullet.UniqueID] == bullet;
		}

		/// <summary>
		/// Returns whether the specified doodad currently exists in the game.
		/// </summary>
		public static bool Exists(IDoodad doodad, IGame game)
		{
			return game.Doodads.Exists(doodad.UniqueID) && game.Doodads[doodad.UniqueID] == doodad;
		}

		/// <summary>
		/// Returns whether the specified tank is able to fire at least one weapon this frame.
		/// </summary>
		public static bool CanFire(ITank tank)
		{
			for(int i = 0; i < tank.Rules.Turrets.Length; i++)
			{
				if(tank.CanFire(i))
					return true;
			}

			return false;
		}

		/// <summary>
		/// Returns the rules for the first turret that can currently fire, or null if none can fire.
		/// </summary>
		public static TurretRules GetTurretRules(ITank tank)
		{
			for(int i = 0; i < tank.Rules.Turrets.Length; i++)
			{
				if(tank.CanFire(i))
					return tank.Rules.Turrets[i];
			}

			return null;
		}

		/// <summary>
		/// Returns whether the specified tank is within range of the current tank.
		/// </summary>
		public static bool IsInRange(ITank tank, ITank target)
		{
			double range = GetTurretRules(tank).Bullet.Range;
			double radius = target.Position.Radius;
			double rangePlusRadius = range + radius;
			return tank.Location.DistanceSquaredTo(target.Location) < rangePlusRadius * rangePlusRadius;
		}

		/// <summary>
		/// Returns whether the two specified tanks are allied.
		/// </summary>
		public static bool AreAllied(ITank tank1, ITank tank2)
		{
			return tank1.Team == tank2.Team && tank1.Team != 0;
		}

		/// <summary>
		/// Returns whether line-of-sight is blocked between the two specified tanks.
		/// </summary>
		public static bool IsLosBlocked(IGame game, ITank tank, ITank target)
		{
			Segment seg = new Segment(tank.Location, target.Location);
			return game.Map.IsTall(seg) ||
			       !game.Roofs.PointCanSeePoint(seg.Point1, seg.Point2);
		}

		/// <summary>
		/// Returns whether the specified doodad is a desirable powerup for the specified tank.
		/// </summary>
		public static bool IsDesirablePowerup(IDoodad doodad, ITank tank)
		{
			return IsDesirablePowerup(doodad, tank, 1, 1, false, false, false, false);
		}

		/// <summary>
		/// Returns whether the specified doodad is a desirable powerup for the specified tank.
		/// </summary>
		public static bool IsDesirablePowerup(IDoodad doodad, ITank tank, double minHealthRatio,
		                                      double minAmmoRatio, bool ignoreInvisibility, bool ignoreSpeed,
		                                      bool ignoreFirepower, bool ignoreArmor)
		{
			if(doodad.Rules.Collectable)
			{
				if(doodad.Rules.CollectionBonuses.Health > 0 && (double)tank.Health / tank.Rules.Combat.MaxHealth <= minHealthRatio)
					return true;
				else if(doodad.Rules.CollectionBonuses.Ammo > 0 && (double)tank.Ammo / tank.Rules.Combat.MaxAmmo <= minAmmoRatio)
					return true;
				else if(doodad.Rules.CollectionBonuses.InvisibilityDuration > 0 && !ignoreInvisibility)
					return true;
				else if(doodad.Rules.CollectionBonuses.SpeedFactor > 1 && doodad.Rules.CollectionBonuses.SpeedDuration > 0 && !ignoreSpeed)
					return true;
				else if(doodad.Rules.CollectionBonuses.PowerFactor > 1 && doodad.Rules.CollectionBonuses.PowerDuration > 0 && !ignoreFirepower)
					return true;
				else if(doodad.Rules.CollectionBonuses.ArmorFactor > 1 && doodad.Rules.CollectionBonuses.ArmorDuration > 0 && !ignoreArmor)
					return true;
				else
					return false;
			}
			else
			{
				return false;
			}
		}

		/// <summary>
		/// Returns whether the specified tank needs ammo.
		/// </summary>
		public static bool NeedsAmmo(ITank tank)
		{
			for(int i = 0; i < tank.Rules.Turrets.Length; i++)
			{
				if(tank.Rules.Turrets[i].Ammo <= tank.Ammo)
					return false;
			}

			return true;
		}

		/// <summary>
		/// Returns the range of the weapon with the shortest range.
		/// </summary>
		public static double ShortestRange(ITank tank)
		{
			double ret = double.MaxValue;
			foreach(TurretRules tRules in tank.Rules.Turrets)
			{
				BulletRules bRules = tRules.Bullet;
				if(bRules.Velocity >= 1 || bRules.Radius < 1)
					ret = Math.Min(ret, bRules.Range);
			}

			return ret;
		}

		/// <summary>
		/// Returns the range of the weapon with the longest minimum range.
		/// </summary>
		public static double LongestMinRange(ITank tank)
		{
			double ret = double.MinValue;
			foreach(TurretRules tRules in tank.Rules.Turrets)
			{
				BulletRules bRules = tRules.Bullet;
				ret = Math.Max(ret, bRules.MinRange);
			}

			return ret;
		}

		/// <summary>
		/// Returns the team opposing the specified CTF team.
		/// </summary>
		public static int OtherCTFTeam(int team)
		{
			if(team == 1)
				return 2;
			else if(team == 2)
				return 1;
			else
				throw new ArgumentException("team");
		}

		/// <summary>
		/// Returns the score for the specified range. The score will be computed from the other parameters.
		/// </summary>
		/// <param name="range">The range whose score will be returned.</param>
		/// <param name="rangeLow">The lower-bound of the range, at which scoreLow will be returned.</param>
		/// <param name="rangeHigh">The upper-bound of the range, at which scoreHigh will be returned.</param>
		/// <param name="scoreLow">The lowest possible score.</param>
		/// <param name="scoreHigh">The highest possible score.</param>
		/// <param name="bounded">Whether the returned score will be bounded by scoreLow, scoreHigh.</param>
		public static double GetRangeBasedScore(double range, double rangeLow, double rangeHigh,
		                                        double scoreLow, double scoreHigh, bool bounded)
		{
			double factor = (range - rangeLow) / (rangeHigh - rangeLow);
			if(bounded)
				factor = Math.Min(1, Math.Max(0, factor));
			return factor * (scoreHigh - scoreLow) + scoreLow;
		}


        /// <summary>
        /// Returns whether the indicated is a repair tank by searching for weapons that do negative damage.
        /// </summary>
        /// <param name="tank"></param>
        /// <returns></returns>
        public static bool IsRepairTank(ITank tank)
        {
            // Search for any weapons that do negative damage.
            foreach (TurretRules tRules in tank.Rules.Turrets)
            {
                BulletRules bRules = tRules.Bullet;
                if (bRules.Damage < 0)
                    return true;
                // Or else keep on looking.
            }
            // If the search gets to this point, there are no repair weapons.
            return false;
        }

        /// <summary>
        /// Returns whether this tank has any weapons that do positive damage.
        /// </summary>
        public static bool IsAttackTank(ITank tank)
        {
            // Search for any weapons that do positive damage.
            foreach (TurretRules tRules in tank.Rules.Turrets)
            {
                BulletRules bRules = tRules.Bullet;
                if (bRules.Damage > 0)
                    return true;
                // Or else keep on looking.
            }
            // If the search gets to this point, there are no attack weapons.
            return false;
        }
        
        /// <summary>
        /// Returns a number between 0.0 and 7.0 indicating the level of the rank of the tank in question.
        /// e.g. Private is 0.0
        ///      Major is   6.0
        ///      Human is   7.0
        /// </summary>
        public static double MilitaryRank(IGame game, ITank tank)
        {
            string threeLetter = game.Players[tank.Player].Name.Substring(0, 3);
            switch (threeLetter)
            {
                case "Dum":
                    // Dummy
                    return 0;
                case "PVT":
                    // Private
                    return 0;
                case "PFC":
                    // Private First Class
                    return 1;
                case "CPL":
                    // Corporal
                    return 2;
                case "SGT":
                    // Sergeant
                    return 3;
                case "LT ":
                    // Lieutenant
                    return 4;
                case "CPT":
                    // Captain
                    return 5;
                case "MAJ":
                    // Major
                    return 6;
                default:
                    // Human player or unknown ranking.  Assume human.
                    return 7;
            }
        }

		#endregion
	}
}