﻿using TankyTank.Game;
using TankyTank.Game.Tanks;
namespace TankyTank.AIBots
{
	/// <summary>
	/// Base class for attack behaviors.
	/// </summary>
	public abstract class AttackBehaviorBase : IAttackBehavior
	{
		#region Constructors

		/// <summary>
		/// Initializes a new AttackBehaviorBase.
		/// </summary>
		public AttackBehaviorBase(IGame game)
		{
			_game = game;
		}

		#endregion
		#region Fields

		protected readonly IGame _game;
		protected ITank _curTank;
		private double? _score;
		private double _behaviorChangeThreshold;

		#endregion
		#region IAttackBehavior Implementations

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		void IAttackBehavior.OnTankChanged(ITank tank)
		{
			_curTank = tank;
			OnTankChanged(tank);
		}

		/// <summary>
		/// Called when this behavior is selected as the attack behavior.
		/// </summary>
		void IAttackBehavior.OnSelected()
		{
			OnSelected();
		}

		/// <summary>
		/// Called when this behavior is deselected as the attack behavior.
		/// </summary>
		void IAttackBehavior.OnDeselected()
		{
			OnDeselected();
		}

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		double IAttackBehavior.BehaviorChangeThreshold
		{
			get { return _behaviorChangeThreshold = GetBehaviorChangeThreshold(); }
		}

		/// <summary>
		/// Gets the attack action score for this behavior, or null to skip this behavior.
		/// </summary>
		double? IAttackBehavior.GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			if(IsValidBehavior())
				return _score = GetScore(curMoveBehavior, curAttackBehavior);
			else
				return _score = null;
		}

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// Return null to perform no action.
		/// </summary>
		AttackAction IAttackBehavior.GetAction(AttackAction prevAttackAction)
		{
			if(IsValidBehavior())
				return GetAction(prevAttackAction);
			else
				return null;
		}

		/// <summary>
		/// Called when this behavior is the current attack behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="attackBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		double? IAttackBehavior.ModifyScore(IAttackBehavior attackBehavior, double? score)
		{
			return ModifyScore(attackBehavior, score);
		}

		/// <summary>
		/// Returns the debug info for this behavior, or null to print no info.
		/// </summary>
		string IAttackBehavior.GetDebugInfo()
		{
			return GetDebugInfo();
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected virtual void OnTankChanged(ITank tank)
		{

		}

		/// <summary>
		/// Called when this behavior is selected as the attack behavior.
		/// </summary>
		protected virtual void OnSelected()
		{

		}

		/// <summary>
		/// Called when this behavior is deselected as the attack behavior.
		/// </summary>
		protected virtual void OnDeselected()
		{

		}

		/// <summary>
		/// Returns the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected virtual double GetBehaviorChangeThreshold()
		{
			return 0;
		}

		/// <summary>
		/// Returns whether this behavior is currently valid.
		/// </summary>
		protected virtual bool IsValidBehavior()
		{
			return true;
		}

		/// <summary>
		/// Gets the attack action score for this behavior, or null to skip this behavior.
		/// </summary>
		protected abstract double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior);

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// Return null to perform no action.
		/// </summary>
		protected abstract AttackAction GetAction(AttackAction prevAttackAction);

		/// <summary>
		/// Called when this behavior is the current attack behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="attackBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		protected virtual double? ModifyScore(IAttackBehavior attackBehavior, double? score)
		{
			return score;
		}

		/// <summary>
		/// Returns the debug info for this behavior.
		/// </summary>
		protected virtual string GetDebugInfo()
		{
			string scoreStr = _score != null ? ((int)_score).ToString() : "null";
			return GetType().Name + " (" + scoreStr + "/+" + _behaviorChangeThreshold + ")";
		}

		#endregion
	}
}