﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Drawing;
using System.Text;
using Halfling.Pathfinding;
using Halfling.Geometry;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Bullets;
using TankyTank.Game.Maps;
using TankyTank.Game.Tanks;
using Tile =
	Halfling.Game.Maps.Tile
		<Halfling.Game.Maps.Terrain.TypeKey, Halfling.Game.Maps.Terrain.UnambiguousType<TankyTank.Game.Maps.GeneralType>>;

namespace TankyTank.AIBots
{
	/// <summary>
	/// Complex score-based tank AI behavior.
	/// </summary>
	public class BehaviorBotBase : AIBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for a BehaviorBotBase.
		/// </summary>
		[AutoSerializable, DeserializeOnce] public class Settings
		{
			[Serial] public readonly int BehaviorInterval;

			[Serial] public readonly int AimLocFindInterval;
			[Serial] public readonly double AimLocError;
			[Serial] public readonly double TargetLeadError;
			[Serial] public readonly Angle TargetFireConeSatisfied;

			[Serial] public readonly double TargetLocationReachedSatisfied;
			[Serial] public readonly Angle TargetDirectionReachedSatisfied;
			[Serial] public readonly Angle TargetPursueCone;
			[Serial] public readonly Angle TargetPursueConeSatisfied;

			[Serial, Optional(0)] public readonly int PrimaryHesitateFramesBeforeShotMin;
			[Serial, Optional(0)] public readonly int PrimaryHesitateFramesBeforeShotMax;
			[Serial, Optional(0)] public readonly int SecondaryHesitateFramesBeforeShotMin;
			[Serial, Optional(0)] public readonly int SecondaryHesitateFramesBeforeShotMax;

			[Serial] public readonly int StuckDetectFrames;
			[Serial] public readonly double StuckDetectDistance;
			public double StuckDetectDistanceSquared { get { return StuckDetectDistance * StuckDetectDistance; } }
			[Serial] public readonly int StuckBackupFrames;
		}

		#endregion
		#region Fields

		protected const int STUCK_DETECT_FRAMES = 40; // How long it will take before triggering stuck behavior if the tank doesn't move far enough.
		protected const double STUCK_DETECT_DISTANCE = 5; // How far a tank has to move to avoid trigging stuck behavior.
		protected const double STUCK_DETECT_DISTANCE_SQUARED = STUCK_DETECT_DISTANCE * STUCK_DETECT_DISTANCE;

		protected readonly Settings _settings;

		protected readonly int _behaviorOffset;
		protected readonly List<IMoveBehavior> _moveBehaviors = new List<IMoveBehavior>();
		protected readonly List<IAttackBehavior> _attackBehaviors = new List<IAttackBehavior>();
		protected IMoveBehavior _curMoveBehavior;
		protected IAttackBehavior _curAttackBehavior;
		protected MoveAction _curMoveAction;
		protected AttackAction _curAttackAction;

		protected Vector _aimLocOffset = Vector.Zero;
		protected Vector _perfectAimLocOffset = Vector.Zero;
		protected readonly int _aimLocFindOffset;

		protected Vector _targetLoc;
		protected Path _pathToTargetLoc;

		protected Vector _lastStuckDetectLoc;
		protected int _lastStuckDetectFrame;
		protected int _stuckBackupUntil;

		protected int? _primaryHesitateCountdown;
		protected int? _secondaryHesitateCountdown;

		#endregion
		#region Properties

		/// <summary>
		/// Gets the behaviors that control the actions of the tank.
		/// </summary>
		public ReadOnlyCollection<IMoveBehavior> MoveBehaviors
		{
			get { return _moveBehaviors.AsReadOnly(); }
		}

		public ReadOnlyCollection<IAttackBehavior> AttackBehaviors
		{
			get { return _attackBehaviors.AsReadOnly(); }
		}

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new BehaviorBotBase.
		/// </summary>
		/// <param name="game">The game currently being played.</param>
		/// <param name="node">The node from which this bot was loaded.</param>
		/// <param name="ser">Serializer used to deserialize settings.</param>
		public BehaviorBotBase(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
			_behaviorOffset = game.Rand.Int32(0, _settings.BehaviorInterval - 1);
			_aimLocFindOffset = game.Rand.Int32(0, _settings.AimLocFindInterval - 1);
		}

		/// <summary>
		/// Creates a new BehaviorBotBase.
		/// </summary>
		/// <param name="game">The game currently being played.</param>
		/// <param name="settings">The bot settings.</param>
		public BehaviorBotBase(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
			_behaviorOffset = game.Rand.Int32(0, _settings.BehaviorInterval - 1);
			_aimLocFindOffset = game.Rand.Int32(0, _settings.AimLocFindInterval - 1);
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Adds a move behavior to this BehaviorBotBase.
		/// </summary>
		public void AddBehavior(IMoveBehavior behavior)
		{
			_moveBehaviors.Add(behavior);
			behavior.OnTankChanged(_curTank);
		}

		/// <summary>
		/// Adds an attack behavior to this BehaviorBotBase.
		/// </summary>
		public void AddBehavior(IAttackBehavior behavior)
		{
			_attackBehaviors.Add(behavior);
			behavior.OnTankChanged(_curTank);
		}

		/// <summary>
		/// Adds a move/attack behavior to this BehaviorBotBase.
		/// </summary>
		public void AddBehavior(IMoveAttackBehavior behavior)
		{
			AddBehavior((IMoveBehavior)behavior);
			AddBehavior((IAttackBehavior)behavior);
		}

		/// <summary>
		/// Gets the tank input for the current frame. This method is required by the IBot interface.
		/// </summary>
		/// <param name="forward">Whether the tank is moving forward.</param>
		/// <param name="reverse">Whether the tank is moving backward.</param>
		/// <param name="left">Whether the tank is rotating left.</param>
		/// <param name="right">Whether the tank is rotating right.</param>
		/// <param name="fireMouseLeft">Whether the tank should fire its weapons bound to the left mouse button if possible.</param>
		/// <param name="fireMouseRight">Whether the tank should fire its weapons bound to the right mouse button if possible.</param>
		/// <param name="aimLoc">The location of the map at which the turret(s) are aimed.</param>
		public override void GetInput(ref bool forward, ref bool reverse, ref bool left, ref bool right,
		                              ref bool fireMouseLeft, ref bool fireMouseRight, ref Vector aimLoc)
		{
			bool isBehaviorFrame = _game.Frame % _settings.BehaviorInterval == _behaviorOffset;

			if(isBehaviorFrame)
			{
				IMoveBehavior curMoveBehavior = _curMoveBehavior;
				IAttackBehavior curAttackBehavior = _curAttackBehavior;
				SelectMoveBehavior(curMoveBehavior, curAttackBehavior);
				SelectAttackBehavior(curMoveBehavior, curAttackBehavior);
			}

			// Get move input.
			if(_curMoveBehavior != null)
			{
				// Update move action if we need to, it's time to, or the last move action was an input action.
				if(isBehaviorFrame || _curMoveAction == null || _curMoveAction is MoveAction.Input)
					_curMoveAction = _curMoveBehavior.GetAction(_curMoveAction);

				// Get input from the current move action.
				if(_curMoveAction is MoveAction.Input)
				{
					MoveAction.Input input = (MoveAction.Input)_curMoveAction;
					forward = input.Forward;
					reverse = input.Reverse;
					left = input.Left;
					right = input.Right;
				}
				else if(_curMoveAction is MoveAction.Target)
				{
					MoveToTarget((MoveAction.Target)_curMoveAction, ref forward, ref reverse, ref left, ref right);
				}
				else if(_curMoveAction == null)
				{
					forward = false;
					reverse = false;
					left = false;
					right = false;
				}
				else
				{
					throw new Exception("Unknown type of move action.");
				}
			}
			else
			{
				forward = false;
				reverse = false;
				left = false;
				right = false;
			}

			// Get attack input.
			if(_curAttackBehavior != null)
			{
				// Update attack action if we need to, it's time to, or the last attack action was an input action.
				if(isBehaviorFrame || _curAttackAction == null || _curAttackAction is AttackAction.Input)
					_curAttackAction = _curAttackBehavior.GetAction(_curAttackAction);

				// Get input from the current attack action.
				if(_curAttackAction is AttackAction.Input)
				{
					AttackAction.Input input = (AttackAction.Input)_curAttackAction;
					fireMouseLeft = input.FireMouseLeft;
					fireMouseRight = input.FireMouseRight;
					aimLoc = input.AimLoc;
				}
				else if(_curAttackAction is AttackAction.Target)
				{
					AttackTarget((AttackAction.Target)_curAttackAction, ref fireMouseLeft, ref fireMouseRight, ref aimLoc);
				}
				else if(_curAttackAction == null)
				{
					fireMouseLeft = false;
					fireMouseRight = false;
				}
				else
				{
					throw new Exception("Unknown type of attack action.");
				}
			}
			else
			{
				fireMouseLeft = false;
				fireMouseRight = false;
			}
		}

		/// <summary>
		/// Called when this bot is spawned as the specified tank.
		/// </summary>
		public override void Spawn(ITank tank)
		{
			base.Spawn(tank);

			foreach(IMoveBehavior behavior in _moveBehaviors)
				behavior.OnTankChanged(tank);
			foreach(IAttackBehavior behavior in _attackBehaviors)
				behavior.OnTankChanged(tank);

			_curMoveBehavior = null;
			_curAttackBehavior = null;
			_curMoveAction = null;
			_curAttackAction = null;
			_aimLocOffset = Vector.Zero;
			_perfectAimLocOffset = Vector.Zero;
			_targetLoc = Vector.Zero;
			_pathToTargetLoc = null;
			_primaryHesitateCountdown = null;
			_secondaryHesitateCountdown = null;
		}

		/// <summary>
		/// Returns a string containing the debug information for this AI bot, or null to display no debug info.
		/// </summary>
		public override string GetDebugInfo()
		{
			StringBuilder sb = new StringBuilder();
			sb.Append(base.GetDebugInfo());
			sb.Append("\nMove Behaviors:\n");
			foreach(IMoveBehavior moveBehavior in _moveBehaviors)
			{
				if(_curMoveBehavior == moveBehavior)
					sb.Append(">> " + moveBehavior.GetDebugInfo() + "\n");
				else
					sb.Append("- " + moveBehavior.GetDebugInfo() + "\n");
			}
			sb.Append("Attack Behaviors:\n");
			foreach(IAttackBehavior attackBehavior in _attackBehaviors)
			{
				if(_curAttackBehavior == attackBehavior)
					sb.Append(">> " + attackBehavior.GetDebugInfo() + "\n");
				else
					sb.Append("- " + attackBehavior.GetDebugInfo() + "\n");
			}
			return sb.ToString();
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Sets the _curMoveBehavior field to the highest-scored behavior.
		/// </summary>
		protected virtual void SelectMoveBehavior(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			IMoveBehavior bestMoveBehavior = _curMoveBehavior;
			double bestMoveScore = _curMoveBehavior != null ?
				(_curMoveBehavior.GetScore(curMoveBehavior, curAttackBehavior) ?? double.MinValue) :
				double.MinValue;
			double changeThreshold = _curMoveBehavior != null ?
				_curMoveBehavior.BehaviorChangeThreshold :
				0;

			// Find best behavior.
			foreach(IMoveBehavior behavior in _moveBehaviors)
			{
				if(behavior == _curMoveBehavior)
					continue;

				double? score = behavior.GetScore(curMoveBehavior, curAttackBehavior);
				score = _curMoveBehavior != null ? _curMoveBehavior.ModifyScore(behavior, score) : score;
				if(score != null && score.Value > bestMoveScore + changeThreshold)
				{
					bestMoveBehavior = behavior;
					bestMoveScore = score.Value;
					changeThreshold = 0;
				}
			}

			// Set current behavior, calling selection/deselection methods as necessary.
			if(_curMoveBehavior != bestMoveBehavior)
			{
				if(_curMoveBehavior != null)
					_curMoveBehavior.OnDeselected();
				if(bestMoveBehavior != null)
					bestMoveBehavior.OnSelected();
				_curMoveBehavior = bestMoveBehavior;
			}
		}

		/// <summary>
		/// Sets the _curAttackBehavior field to the highest-scored behavior.
		/// </summary>
		protected virtual void SelectAttackBehavior(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			IAttackBehavior bestAttackBehavior = _curAttackBehavior;
			double bestAttackScore = _curAttackBehavior != null ?
				(_curAttackBehavior.GetScore(curMoveBehavior, curAttackBehavior) ?? double.MinValue) :
				double.MinValue;
			double changeThreshold = _curAttackBehavior != null ?
				_curAttackBehavior.BehaviorChangeThreshold :
				0;

			// Find best behavior.
			foreach(IAttackBehavior behavior in _attackBehaviors)
			{
				if(behavior == _curAttackBehavior)
					continue;

				double? score = behavior.GetScore(curMoveBehavior, curAttackBehavior);
				score = _curAttackBehavior != null ? _curAttackBehavior.ModifyScore(behavior, score) : score;
				if(score != null && score.Value > bestAttackScore + changeThreshold)
				{
					bestAttackBehavior = behavior;
					bestAttackScore = score.Value;
					changeThreshold = 0;
				}
			}

			// Set current behavior, calling selection/deselection methods as necessary.
			if(_curAttackBehavior != bestAttackBehavior)
			{
				if(_curAttackBehavior != null)
					_curAttackBehavior.OnDeselected();
				if(bestAttackBehavior != null)
					bestAttackBehavior.OnSelected();
				_curAttackBehavior = bestAttackBehavior;
			}
		}

		/// <summary>
		/// Gets the tank input for moving to the specified target.
		/// </summary>
		/// <param name="target">The target location and direction.</param>
		/// <param name="forward">Whether the tank is moving forward.</param>
		/// <param name="reverse">Whether the tank is moving backward.</param>
		/// <param name="left">Whether the tank is rotating left.</param>
		/// <param name="right">Whether the tank is rotating right.</param>
		protected virtual void MoveToTarget(MoveAction.Target target, ref bool forward, ref bool reverse, ref bool left,
		                                    ref bool right)
		{
			forward = false;
			reverse = false;
			left = false;
			right = false;

			// Backup if stuck?
			if(_pathToTargetLoc != null)
			{
				if(_game.Frame < _stuckBackupUntil)
				{
					reverse = true;
					return;
				}
				else if(_stuckBackupUntil != 0)
				{
					_stuckBackupUntil = 0;
					_pathToTargetLoc = null; // Forces recomputation of path after done backing up.
				}
			}
			else
			{
				_stuckBackupUntil = 0;
			}

			// Get path to target.
			if(target.Location == null)
				_pathToTargetLoc = null;
			else if(_pathToTargetLoc == null || _targetLoc != target.Location)
			{
				_targetLoc = target.Location.Value;
				try
				{
					_pathToTargetLoc = _game.Map.FindShortPath(_curTank.Location, _targetLoc);
				}
				catch(CoarseGridPathfinderWrapper.InconsistentGridException)
				{
					// Just leave path null.
				}
			}

			bool doRotation = false;
			if(_pathToTargetLoc != null)
			{
				// Advance to next node if it has been reached.
				double targetLocationReachedSatisfiedSquared = _settings.TargetLocationReachedSatisfied *
				                                               _settings.TargetLocationReachedSatisfied;
				while(!_pathToTargetLoc.IsFinished &&
				      _curTank.Location.DistanceSquaredTo(_pathToTargetLoc.CurrentNode) <= targetLocationReachedSatisfiedSquared)
					_pathToTargetLoc.GetNextNode();

				// Move towards node unless we're close enough.
				if(_curTank.Location.DistanceSquaredTo(_pathToTargetLoc.CurrentNode) > targetLocationReachedSatisfiedSquared)
				{
					Direction dirToTarget = _curTank.Location.DirectionTo(_pathToTargetLoc.CurrentNode);
					Angle angleToTarget = _curTank.ChassisRot.AngularDistanceTo(dirToTarget);

					if(angleToTarget > _settings.TargetPursueConeSatisfied / 2)
					{
						if(_curTank.ChassisRot.AngularSignTo(dirToTarget) < 0)
							left = true;
						else
							right = true;
					}

					if(angleToTarget <= _settings.TargetPursueCone / 2)
					{
						forward = true;

						// Detect if stuck?
						if(_lastStuckDetectFrame == 0 || _lastStuckDetectLoc.DistanceSquaredTo(_curTank.Location) >= _settings.StuckDetectDistanceSquared)
						{
							_lastStuckDetectLoc = _curTank.Location;
							_lastStuckDetectFrame = _game.Frame;
						}
						else if(_game.Frame > _lastStuckDetectFrame + _settings.StuckDetectFrames)
						{
							_stuckBackupUntil = _game.Frame + _settings.StuckBackupFrames;
							_lastStuckDetectFrame = 0;
						}
					}
					else
					{
						_lastStuckDetectFrame = 0;
					}
				}

				// If we've reached the node, rotate until we're at the desired direction.
				else
				{
					doRotation = true;
					_lastStuckDetectFrame = 0;
				}
			}
			else
			{
				_lastStuckDetectFrame = 0;
			}

			if(doRotation && target.Direction != null)
			{
				Direction dirToTarget = target.Direction.Value;
				if(_curTank.ChassisRot.AngularDistanceTo(dirToTarget) > _settings.TargetPursueConeSatisfied / 2)
				{
					if(_curTank.ChassisRot.AngularSignTo(dirToTarget) < 0)
						left = true;
					else
						right = true;
				}
			}
		}

		/// <summary>
		/// Gets the tank input for the current frame. This method is required by the IBot interface.
		/// </summary>
		/// <param name="target">The target tank.</param>
		/// <param name="fireMouseLeft">Whether the tank should fire its weapons bound to the left mouse button if possible.</param>
		/// <param name="fireMouseRight">Whether the tank should fire its weapons bound to the right mouse button if possible.</param>
		/// <param name="aimLoc">The location of the map at which the turret(s) are aimed.</param>
		protected virtual void AttackTarget(AttackAction.Target target, ref bool fireMouseLeft, ref bool fireMouseRight,
		                                    ref Vector aimLoc)
		{
			// Find aim location offset.
			if(_game.Frame % _settings.AimLocFindInterval == _aimLocFindOffset)
			{
				double distToTarget = _curTank.Location.DistanceTo(target.Tank.Location);
				TurretRules tRules = AITools.GetTurretRules(_curTank);
				BulletRules bRules = tRules != null ? tRules.Bullet : _curTank.Rules.Turrets[0].Bullet;
				double leadError = _game.Rand.Double(-_settings.TargetLeadError, _settings.TargetLeadError);
				_aimLocOffset = (target.Tank.Velocity + target.Tank.Velocity * leadError) * (distToTarget / bRules.Velocity);
				Direction errorDir = _game.Rand.Direction();
				double errorMag = _game.Rand.Double(0, _settings.AimLocError);
				_aimLocOffset += Vector.FromAngle(errorDir) * errorMag;

				// Find perfect aim location for use in the "should shoot" algorithm.
				_perfectAimLocOffset = target.Tank.Velocity * (distToTarget / bRules.Velocity);
			}

			// Find aim locations based on offsets.
			aimLoc = target.Tank.Location + _aimLocOffset;
			Vector perfectAimLoc = target.Tank.Location + _perfectAimLocOffset;

			// Fire?
			fireMouseLeft = ShouldShoot(target, FireButton.MouseLeft, perfectAimLoc);
			fireMouseRight = ShouldShoot(target, FireButton.MouseRight, perfectAimLoc);

			// Delay firing primary.
			if(fireMouseLeft)
			{
				if(_primaryHesitateCountdown == null)
					_primaryHesitateCountdown = GetPrimaryHesitateFrames();

				Debug.Assert(_primaryHesitateCountdown.Value >= 0);
				if(_primaryHesitateCountdown.Value == 0)
					_primaryHesitateCountdown = null;
				else
				{
					fireMouseLeft = false;
					_primaryHesitateCountdown = _primaryHesitateCountdown.Value - 1;
				}
			}
			else
			{
				_primaryHesitateCountdown = null;
			}

			// Delay firing secondary.
			if(fireMouseRight)
			{
				if(_secondaryHesitateCountdown == null)
					_secondaryHesitateCountdown = GetSecondaryHesitateFrames();

				Debug.Assert(_secondaryHesitateCountdown.Value >= 0);
				if(_secondaryHesitateCountdown.Value == 0)
					_secondaryHesitateCountdown = null;
				else
				{
					fireMouseRight = false;
					_secondaryHesitateCountdown = _secondaryHesitateCountdown.Value - 1;
				}
			}
			else
			{
				_secondaryHesitateCountdown = null;
			}
		}

		/// <summary>
		/// Returns the number of frames to wait before firing the primary weapons.
		/// </summary>
		protected int GetPrimaryHesitateFrames()
		{
			return _game.Rand.Int32(_settings.PrimaryHesitateFramesBeforeShotMin,
			                        _settings.PrimaryHesitateFramesBeforeShotMax);
		}

		/// <summary>
		/// Returns the number of frames to wait before firing the secondary weapons.
		/// </summary>
		protected int GetSecondaryHesitateFrames()
		{
			return _game.Rand.Int32(_settings.SecondaryHesitateFramesBeforeShotMin,
			                        _settings.SecondaryHesitateFramesBeforeShotMax);
		}

		/// <summary>
		/// Returns whether the weapons bound to the specified button should be shot.
		/// </summary>
		protected virtual bool ShouldShoot(AttackAction.Target target, FireButton button, Vector aimLoc)
		{
			// Look for all weapons bound to this FireButton (e.g. left or right click).
            for(int i = 0; i < _curTank.Rules.Turrets.Length; i++)
			{
				TurretRules tRules = _curTank.Rules.Turrets[i];
				if(_curTank.CanFire(i) && tRules.FireButton == button)
				{
					// Don't heal an enemy or shoot a friend on purpose.
                    bool allies = AITools.AreAllied(_curTank, target.Tank);
                    if ((tRules.Bullet.Damage < 0 && !allies) ||
                        (tRules.Bullet.Damage > 0 && allies))
                        return false; // Even if other weapons are appropriate.

                    // If any weapons are good, then fire at will!
                    Segment seg = new Segment(_curTank.Location, aimLoc);
					Angle dist = ((Direction)seg.Point1.DirectionTo(seg.Point2)).AngularDistanceTo(_curTank.ChassisRot);
					if(dist <= tRules.Arc / 2 + _settings.TargetFireConeSatisfied / 2 &&
					   AITools.IsInRange(_curTank, target.Tank) && ShouldShoot(seg))
					{
						return true;  // Even if other weapons can't fire from where they are.
					}
				}
			}

			return false;
		}

		/// <summary>
		/// Returns whether the specified segment is a good firing path.
		/// </summary>
		protected virtual bool ShouldShoot(Segment seg)
		{
			object obstacle = null;
			double distanceSquared = double.MaxValue;

			// Search terrain.
			var potentialIntersectingTiles = _game.Map.FindTall(seg);
			foreach(Point tile in potentialIntersectingTiles)
			{
				Rect pos = _game.Map.GetTileRect(tile.X, tile.Y);
				Segment? intersection = pos.FindIntersection(seg);
				if(intersection != null)
				{
					Vector closest = seg.Point1.FindClosest(intersection.Value.Point1, intersection.Value.Point2);
					double d = closest.DistanceSquaredTo(seg.Point1);
					if(d < distanceSquared)
					{
						obstacle = tile;
						distanceSquared = d;
					}
				}
			}

			// Search tanks.
			var potentialIntersectingTanks = _game.Tanks.FindIntersectsWith(seg, _curTank.UniqueID);
			foreach(ITank tank in potentialIntersectingTanks)
			{
				Segment? intersection = tank.Position.FindIntersection(seg);
				if(intersection != null)
				{
					Vector closest = seg.Point1.FindClosest(intersection.Value.Point1, intersection.Value.Point2);
					double d = closest.DistanceSquaredTo(seg.Point1);
					if(d < distanceSquared)
					{
						obstacle = tank;
						distanceSquared = d;
					}
				}
			}

			// Return.
			return obstacle == null ||
			       (obstacle is ITank && !AITools.AreAllied(_curTank, (ITank)obstacle)) ||
			       !(obstacle is Tile);
		}

		#endregion
	}
}