using System.Collections.Generic;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Doodads;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// A move behavior that causes the tank to seek out and collect powerups.
	/// </summary>
	public class CollectPowerupBehavior : MoveBehaviorBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for a CollectPowerupBehavior.
		/// </summary>
		[AutoSerializable, DeserializeOnce] public class Settings
		{
			[Serial] public readonly double ScoreLow;
			[Serial] public readonly double ScoreHigh;
			[Serial] public readonly double BehaviorChangeThreshold;
			[Serial] public readonly double TargetChangeThreshold;

            /// <summary>
            /// Fractional distance this tank can "see".
            /// </summary>
            [Serial, Optional(1.0)] public readonly double RadarFraction;

            [Serial, Optional(1.0)] public readonly double MinHealthRatio;
            [Serial, Optional(1.0)] public readonly double MinAmmoRatio;
            [Serial, Optional(false)] public readonly bool IgnoreInvisibility;
            [Serial, Optional(false)] public readonly bool IgnoreSpeed;
            [Serial, Optional(false)] public readonly bool IgnoreFirepower;
            [Serial, Optional(false)] public readonly bool IgnoreArmor;
		}

		#endregion
		#region Constructors

		/// <summary>
		/// Loads a new CollectPowerupBehavior.
		/// </summary>
		public CollectPowerupBehavior(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
		}

		/// <summary>
		/// Creates a new CollectPowerupBehavior.
		/// </summary>
		public CollectPowerupBehavior(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
		}

		#endregion
		#region Fields

		protected readonly Settings _settings;
		protected IDoodad _target;

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected override double GetBehaviorChangeThreshold()
		{
			return _settings.BehaviorChangeThreshold;
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		protected override void OnSelected()
		{
			_target = null;
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected override void OnTankChanged(ITank tank)
		{
			if(tank != null)
				_target = null;
		}

		/// <summary>
		/// Returns whether the current target should be cleared under the current circumstances.
		/// </summary>
		protected virtual bool ShouldClearTarget()
		{
			return _target != null &&
			       (!AITools.Exists(_target, _game) ||
			        !IsDesirablePowerup(_target) ||
			        _target.Position.DistanceSquaredTo(_curTank.Location) > (_curTank.Rules.RadarRange * _settings.RadarFraction) * (_curTank.Rules.RadarRange * _settings.RadarFraction));
		}

		/// <summary>
		/// Returns the score for the specified doodad.
		/// </summary>
		protected virtual double GetDoodadScore(IDoodad doodad)
		{
			return ((_curTank.Rules.RadarRange * _settings.RadarFraction) - _curTank.Location.DistanceTo(doodad.Location)) / (_curTank.Rules.RadarRange * _settings.RadarFraction);
		}

		/// <summary>
		/// Returns whether the specified doodad is a desirable powerup.
		/// </summary>
		protected virtual bool IsDesirablePowerup(IDoodad doodad)
		{
			return AITools.IsDesirablePowerup(doodad,
			                                  _curTank,
			                                  _settings.MinHealthRatio,
			                                  _settings.MinAmmoRatio,
			                                  _settings.IgnoreInvisibility,
			                                  _settings.IgnoreSpeed,
			                                  _settings.IgnoreFirepower,
			                                  _settings.IgnoreArmor);
		}

		/// <summary>
		/// Returns the valid targets from which the target can be chosen.
		/// </summary>
		protected virtual IEnumerable<IDoodad> GetValidTargets()
		{
			Circle searchCircle = new Circle(_curTank.Location, (_curTank.Rules.RadarRange * _settings.RadarFraction));
			var found = _game.Doodads.FindIntersectsWith(searchCircle);
			foreach(IDoodad doodad in found)
			{
				if(IsDesirablePowerup(doodad))
					yield return doodad;
			}
		}

		/// <summary>
		/// Sets the _target field to the best target.
		/// </summary>
		protected virtual double? PickBestTarget()
		{
			var found = GetValidTargets();

			IDoodad best = _target;
			double bestScore = best != null ? GetDoodadScore(best) : double.MinValue;
			double changeThreshold = best != null ? _settings.TargetChangeThreshold : 0;
			foreach(IDoodad doodad in found)
			{
				if(!_game.Map.IsPassable(doodad.Location))
					continue;
                
                double score = GetDoodadScore(doodad);
				if(score > bestScore + changeThreshold)
				{
					best = doodad;
					bestScore = score;
					changeThreshold = 0;
				}
			}

			_target = best;
			return best != null ? bestScore : (double?)null;
		}

		/// <summary>
		/// Gets the score for this behavior, or null to skip this behavior.
		/// </summary>
		protected override double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			if(ShouldClearTarget())
				_target = null;

			double? score = PickBestTarget();
			if(score != null)
				return AITools.GetRangeBasedScore(score.Value, 0, 1, _settings.ScoreLow, _settings.ScoreHigh, true);
			else
				return null;
		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		protected override MoveAction GetAction(MoveAction prevMoveAction)
		{
			if(ShouldClearTarget())
				_target = null;

			if(_target != null)
				return new MoveAction.Target(_target.Location, null);
			else
				return null;
		}

		#endregion
	}
}