using System;
using System.Collections.Generic;
using Halfling.Geometry;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Doodads;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// A move behavior that causes the tank to seek out and collect powerups.
	/// </summary>
	public class FollowSuperiorBehavior : MoveBehaviorBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for a FollowSuperiorBehavior.
		/// </summary>
		[AutoSerializable, DeserializeOnce] public class Settings
		{
			[Serial] public readonly double ScoreLow;
			[Serial] public readonly double ScoreHigh;
			[Serial] public readonly double BehaviorChangeThreshold;
			[Serial] public readonly double TargetChangeThreshold;

            /// <summary>
            /// Fractional distance this tank can "see".
            /// </summary>
            [Serial, Optional(1.0)] public readonly double RadarFraction;

            [Serial] public readonly Angle TacticalCone;
            [Serial] public readonly Angle TacticalConeSatisfied;
            [Serial] public readonly double TacticalForwardRadius;
            [Serial] public readonly double TacticalReverseRadius;

            [Serial, Optional(0)] public readonly int HesitateFramesBeforeMovingMin;
            [Serial, Optional(0)] public readonly int HesitateFramesBeforeMovingMax;
		}

		#endregion
		#region Constructors

		/// <summary>
		/// Loads a new FollowSuperiorBehavior.
		/// </summary>
		public FollowSuperiorBehavior(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
		}

		/// <summary>
		/// Creates a new FollowSuperiorBehavior.
		/// </summary>
		public FollowSuperiorBehavior(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
		}

		#endregion
		#region Fields

		protected readonly Settings _settings;
		protected ITank _target;
        protected int _hesitateMovingCountdown;
        protected double _milRank;
        
		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected override double GetBehaviorChangeThreshold()
		{
			return _settings.BehaviorChangeThreshold;
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		protected override void OnSelected()
		{
			_target = null;
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected override void OnTankChanged(ITank tank)
		{
			if(tank != null)
				_target = null;
            

		}

		/// <summary>
		/// Returns whether the current target should be cleared under the current circumstances.
		/// </summary>
		protected virtual bool ShouldClearTarget()
		{
			return _target != null &&
			       (!AITools.Exists(_target, _game) ||
			        _target.Position.DistanceSquaredTo(_curTank.Location) > (_curTank.Rules.RadarRange * _settings.RadarFraction) * (_curTank.Rules.RadarRange * _settings.RadarFraction));
		}

		/// <summary>
		/// Returns the score for the specified tank.
		/// </summary>
		protected virtual double GetTankScore(ITank tank)
		{
            // Closer tanks get higher score between 0.0 and 1.0 for tanks inside sight range and negative for those outside.
            double baseScore = ((_curTank.Rules.RadarRange * _settings.RadarFraction) - _curTank.Location.DistanceTo(tank.Location)) / (_curTank.Rules.RadarRange * _settings.RadarFraction);
            
            // A higher ranking tank is better to follow.  Add a score between 0 and 7 for rank difference.
            baseScore += AITools.MilitaryRank(_game, tank) - _milRank;

            // Highest possible score is 8.  Normalize score.
            baseScore /= 8;

            return baseScore;
		}

		/// <summary>
		/// Returns the valid targets from which the target can be chosen.
		/// </summary>
		protected virtual IEnumerable<ITank> GetValidTargets()
		{
			Circle searchCircle = new Circle(_curTank.Location, (_curTank.Rules.RadarRange * _settings.RadarFraction));
			var found = _game.Tanks.FindIntersectsWith(searchCircle);
            foreach(ITank tank in found)
			{
                if (AITools.AreAllied(tank,_curTank) &&             // Tank is friendly.
                    AITools.MilitaryRank(_game,tank) > _milRank)    // Tank has a higher ranking.
                    yield return tank;
			}
		}

		/// <summary>
		/// Sets the _target field to the best target.
		/// </summary>
		protected virtual double? PickBestTarget()
		{
			var found = GetValidTargets();

			ITank best = _target;
			double bestScore = best != null ? GetTankScore(best) : double.MinValue;
			double changeThreshold = best != null ? _settings.TargetChangeThreshold : 0;
			foreach(ITank tank in found)
			{
                if (!_game.Map.IsPassable(tank.Location))
					continue;

                double score = GetTankScore(tank);
				if(score > bestScore + changeThreshold)
				{
                    best = tank;
					bestScore = score;
					changeThreshold = 0;
				}
			}

			_target = best;
			return best != null ? bestScore : (double?)null;
		}

		/// <summary>
		/// Gets the score for this behavior, or null to skip this behavior.
		/// </summary>
		protected override double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
            // Establish rank of this player.
            _milRank = AITools.MilitaryRank(_game, _curTank);
            
            if(ShouldClearTarget())
				_target = null;

			double? score = PickBestTarget();
			if(score != null)
				return AITools.GetRangeBasedScore(score.Value, 0, 1, _settings.ScoreLow, _settings.ScoreHigh, true);
			else
				return null;
		}

        /// <summary>
        /// Gets the move action for this behavior. The returned value will
        /// be either MoveAction.Target or MoveAction.Input depending
        /// on whether the action calls for moving to a point/direction using the
        /// built-in movement and pathfinding or for manually controlling the movement input.
        /// Return null to perform no action.
        /// </summary>
        protected override MoveAction GetAction(MoveAction prevMoveAction)
        {
            if (ShouldClearTarget())
                _target = null;

            // Hesitate moving?
            if (_hesitateMovingCountdown > 0)
            {
                _hesitateMovingCountdown--;
                return null;
            }

            if (_target != null)
            {
                if (AITools.IsLosBlocked(_game, _curTank, _target))
                    return new MoveAction.Target(_target.Location, null);
                else
                    return GetTacticalMoveAction(prevMoveAction);
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Returns the move action for when the target tank is close enough to worry about
        /// about tactical movement instead of pathfinding.
        /// 
        /// Note: Range of weapons is not important in this version of GetTacticalMoveAction.
        /// </summary>
        protected virtual MoveAction GetTacticalMoveAction(MoveAction prevMoveAction)
        {
            bool forward = false;
            bool reverse = false;
            bool left = false;
            bool right = false;

            Direction dirToTarget = _curTank.Location.DirectionTo(_target.Location);
			if(_curTank.ChassisRot.AngularDistanceTo(dirToTarget) > _settings.TacticalConeSatisfied / 2)
            {
                if (_curTank.ChassisRot.AngularSignTo(dirToTarget) < 0)
                    left = true;
                else
                    right = true;
            }

            double distSquaredToTarget = _curTank.Location.DistanceSquaredTo(_target.Location);
            if (distSquaredToTarget <= _settings.TacticalReverseRadius * _settings.TacticalReverseRadius)
            {
                reverse = true;
            }
            else if((distSquaredToTarget >= _settings.TacticalForwardRadius * _settings.TacticalForwardRadius) &&
				_curTank.ChassisRot.AngularDistanceTo(dirToTarget) <= _settings.TacticalCone)
            {
                forward = true;
            }

            return new MoveAction.Input(forward, reverse, left, right);
        }

		#endregion
	}
}