﻿using Halfling.Geometry.Dim2;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// Defines a single scoreable tank attack behavior.
	/// </summary>
	public interface IAttackBehavior
	{
		#region Properties

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		double BehaviorChangeThreshold { get; }

		#endregion
		#region Methods

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		void OnTankChanged(ITank tank);

		/// <summary>
		/// Called when this behavior is selected as the attack behavior.
		/// </summary>
		void OnSelected();

		/// <summary>
		/// Called when this behavior is deselected as the attack behavior.
		/// </summary>
		void OnDeselected();

		/// <summary>
		/// Gets the attack action score for this behavior, or null to skip this behavior.
		/// </summary>
		double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior);

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// Return null to perform no action.
		/// </summary>
		AttackAction GetAction(AttackAction prevAttackAction);

		/// <summary>
		/// Called when this behavior is the current attack behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="attackBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		double? ModifyScore(IAttackBehavior attackBehavior, double? score);

		/// <summary>
		/// Returns the debug info for this behavior.
		/// </summary>
		string GetDebugInfo();

		#endregion
	}

	/// <summary>
	/// Stores information about a behavior's attack action.
	/// This class is abstract. Use either AttackAction.Target or
	/// AttackAction.Input.
	/// </summary>
	public abstract class AttackAction
	{
		#region Constructors

		/// <summary>
		/// Private constructor to prevent extension and instantiation.
		/// </summary>
		private AttackAction() { }

		#endregion
		#region Subtypes

		/// <summary>
		/// Stores a behavior's attack action as a target tank.
		/// </summary>
		public class Target : AttackAction
		{
			#region Constructors

			/// <summary>
			/// Creates a new AttackAction.Target.
			/// </summary>
			/// <param name="tank">The target tank.</param>
			public Target(ITank tank)
			{
				Tank = tank;
			}

			#endregion
			#region Properties

			/// <summary>
			/// Gets the target tank.
			/// </summary>
			public ITank Tank { get; private set; }

			#endregion
		}

		/// <summary>
		/// Stores a behavior's attack action as aim location and firing input.
		/// </summary>
		public class Input : AttackAction
		{
			#region Constructors

			/// <summary>
			/// Creates a new AttackAction.Input.
			/// </summary>
			/// <param name="aimLoc">The location at which the tank should aim.</param>
			/// <param name="fireMouseLeft">Whether the tank should fire its primary weapons.</param>
			/// <param name="fireMouseRight">Whether the tank should fire its secondary weapons.</param>
			public Input(Vector aimLoc, bool fireMouseLeft, bool fireMouseRight)
			{
				AimLoc = aimLoc;
				FireMouseLeft = fireMouseLeft;
				FireMouseRight = fireMouseRight;
			}

			#endregion
			#region Properties

			/// <summary>
			/// Gets the location at which the tank should aim.
			/// </summary>
			public Vector AimLoc { get; private set; }

			/// <summary>
			/// Gets whether the tank should fire its primary weapons.
			/// </summary>
			public bool FireMouseLeft { get; private set; }

			/// <summary>
			/// Gets whether the tank should fire its secondary weapons.
			/// </summary>
			public bool FireMouseRight { get; private set; }

			#endregion
		}

		#endregion
	}
}