﻿using Halfling.Geometry.Dim2;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// Defines a single scoreable tank movement behavior.
	/// </summary>
	public interface IMoveBehavior
	{
		#region Properties

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		double BehaviorChangeThreshold { get; }

		#endregion
		#region Methods

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		void OnTankChanged(ITank tank);

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		void OnSelected();

		/// <summary>
		/// Called when this behavior is deselected as the move behavior.
		/// </summary>
		void OnDeselected();

		/// <summary>
		/// Gets the move action score for this behavior, or null to skip this behavior.
		/// </summary>
		double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior);

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		MoveAction GetAction(MoveAction prevMoveAction);

		/// <summary>
		/// Called when this behavior is the current move behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="moveBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		double? ModifyScore(IMoveBehavior moveBehavior, double? score);

		/// <summary>
		/// Returns the debug info for this behavior.
		/// </summary>
		string GetDebugInfo();

		#endregion
	}

	/// <summary>
	/// Stores information about a behavior's move action.
	/// This class is abstract. Use either MoveAction.Target or
	/// MoveAction.Input.
	/// </summary>
	public abstract class MoveAction
	{
		#region Constructors

		/// <summary>
		/// Private constructor to prevent extension and instantiation.
		/// </summary>
		private MoveAction() { }

		#endregion
		#region Subtypes

		/// <summary>
		/// Stores a behavior's move action as the tank's target location and direction.
		/// </summary>
		public class Target : MoveAction
		{
			#region Constructors

			/// <summary>
			/// Creates a new MoveAction.Target.
			/// </summary>
			/// <param name="location">The target location of the tank, or null if the location doesn't matter.</param>
			/// <param name="direction">The target direction of the tank, or null if the direction doesn't matter.</param>
			public Target(Vector? location, Direction? direction)
			{
				Location = location;
				Direction = direction;
			}

			#endregion
			#region Properties

			/// <summary>
			/// Gets the target location of the tank, or null if the location doesn't matter.
			/// </summary>
			public Vector? Location { get; private set; }

			/// <summary>
			/// Gets the target direction of the tank, or null if the direction doesn't matter.
			/// </summary>
			public Direction? Direction { get; private set; }

			#endregion
		}

		/// <summary>
		/// Stores a behavior's move action as forward, reverse, left, and right input.
		/// </summary>
		public class Input : MoveAction
		{
			#region Constructors

			/// <summary>
			/// Creates a new MoveAction.Input.
			/// </summary>
			/// <param name="forward">Whether the tank should move forward.</param>
			/// <param name="reverse">Whether the tank should move in reverse.</param>
			/// <param name="left">Whether the tank should rotate left.</param>
			/// <param name="right">Whether the tank should rotate right.</param>
			public Input(bool forward, bool reverse, bool left, bool right)
			{
				Forward = forward;
				Reverse = reverse;
				Left = left;
				Right = right;
			}

			#endregion
			#region Properties

			/// <summary>
			/// Gets whether the tank should move forward.
			/// </summary>
			public bool Forward { get; private set; }

			/// <summary>
			/// Gets whether the tank should move in reverse.
			/// </summary>
			public bool Reverse { get; private set; }

			/// <summary>
			/// Gets whether the tank should rotate left.
			/// </summary>
			public bool Left { get; private set; }

			/// <summary>
			/// Gets whether the tank should rotate right.
			/// </summary>
			public bool Right { get; private set; }

			#endregion
		}

		#endregion
	}
}