﻿using System;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;

namespace TankyTank.AIBots
{
	/// <summary>
	/// An attack behavior that causes the tank's turret to aim randomly as if it is looking around.
	/// </summary>
	public class LookAroundBehavior : AttackBehaviorBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for a LookAroundBehavior.
		/// </summary>
		[AutoSerializable, DeserializeOnce] public class Settings
		{
			[Serial] public readonly double Score;
			[Serial] public readonly double BehaviorChangeThreshold;
			[Serial] public readonly int AimLocChangeInterval;

            /// <summary>
            /// Fractional distance this tank can "see".
            /// </summary>
            [Serial, Optional(1.0)] public readonly double RadarFraction;
		}

		#endregion
		#region Constructors

		/// <summary>
		/// Loads a new LookAroundBehavior.
		/// </summary>
		public LookAroundBehavior(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
		}

		/// <summary>
		/// Creates a new LookAroundBehavior.
		/// </summary>
		public LookAroundBehavior(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
		}

		#endregion
		#region Fields

		protected readonly Settings _settings;

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected override double GetBehaviorChangeThreshold()
		{
			return _settings.BehaviorChangeThreshold;
		}

		/// <summary>
		/// Gets the move action score for this behavior, or null to skip this behavior.
		/// </summary>
		protected override double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			return _settings.Score;
		}

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// Return null to perform no action.
		/// </summary>
		protected override AttackAction GetAction(AttackAction prevAttackAction)
		{
			if(_game.Frame % _settings.AimLocChangeInterval == 0)
			{
				Direction dir = _game.Rand.Direction();
				Vector aimLoc = _curTank.Location + Vector.FromAngle(dir) * (_curTank.Rules.RadarRange * _settings.RadarFraction);
				return new AttackAction.Input(aimLoc, false, false);
			}
			else
			{
				return null;
			}
		}

		#endregion
	}
}