﻿using TankyTank.Game;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// Base class for combined move/attack behaviors.
	/// Prevents duplicate calls for some methods and allows optimizations for others.
	/// </summary>
	public abstract class MoveAttackBehaviorBase : IMoveAttackBehavior
	{
		#region Constructors

		/// <summary>
		/// Initializes a new MoveAttackBehaviorBase.
		/// </summary>
		protected MoveAttackBehaviorBase(IGame game)
		{
			_game = game;
		}

		#endregion
		#region Fields

		protected readonly IGame _game;

		protected ITank _curTank;
		private int _tankChangedFrame = -1;

		private bool _selectedAsMove;
		private bool _selectedAsAttack;

		private double _threshold;
		private int _thresholdFrame = -1;

		private double? _score;
		private int _scoreFrame = -1;

		private int _beforeGetActionsFrame = -1;

		#endregion
		#region IMoveAttackBehavior Implementations

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		void IMoveBehavior.OnTankChanged(ITank tank)
		{
			CallOnTankChanged(tank);
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		void IAttackBehavior.OnTankChanged(ITank tank)
		{
			CallOnTankChanged(tank);
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		void IMoveBehavior.OnSelected()
		{
			OnSelectedAsMove();
			CallOnSelected();
			_selectedAsMove = true;
		}

		/// <summary>
		/// Called when this behavior is selected as the attack behavior.
		/// </summary>
		void IAttackBehavior.OnSelected()
		{
			OnSelectedAsAttack();
			CallOnSelected();
			_selectedAsAttack = true;
		}

		/// <summary>
		/// Called when this behavior is deselected as the move behavior.
		/// </summary>
		void IMoveBehavior.OnDeselected()
		{
			OnDeselectedAsMove();
			CallOnDeselected();
			_selectedAsMove = false;
		}

		/// <summary>
		/// Called when this behavior is deselected as the attack behavior.
		/// </summary>
		void IAttackBehavior.OnDeselected()
		{
			OnDeselectedAsAttack();
			CallOnDeselected();
			_selectedAsAttack = false;
		}

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		double IMoveBehavior.BehaviorChangeThreshold
		{
			get { return CallGetBehaviorChangeThreshold(); }
		}

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		double IAttackBehavior.BehaviorChangeThreshold
		{
			get { return CallGetBehaviorChangeThreshold(); }
		}

		/// <summary>
		/// Gets the move action score for this behavior, or null to skip this behavior.
		/// </summary>
		double? IMoveBehavior.GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			return CallGetScore(curMoveBehavior, curAttackBehavior);
		}

		/// <summary>
		/// Gets the attack action score for this behavior, or null to skip this behavior.
		/// </summary>
		double? IAttackBehavior.GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			return CallGetScore(curMoveBehavior, curAttackBehavior);
		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// </summary>
		MoveAction IMoveBehavior.GetAction(MoveAction prevMoveAction)
		{
			if(IsValidBehavior())
			{
				CallBeforeGetActions();
				return GetMoveAction(prevMoveAction);
			}
			else
			{
				return null;
			}
		}

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// </summary>
		AttackAction IAttackBehavior.GetAction(AttackAction prevAttackAction)
		{
			if(IsValidBehavior())
			{
				CallBeforeGetActions();
				return GetAttackAction(prevAttackAction);
			}
			else
			{
				return null;
			}
		}

		/// <summary>
		/// Called when this behavior is the current move behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="moveBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		double? IMoveBehavior.ModifyScore(IMoveBehavior moveBehavior, double? score)
		{
			return ModifyScore(moveBehavior, score);
		}

		/// <summary>
		/// Called when this behavior is the current attack behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="attackBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		double? IAttackBehavior.ModifyScore(IAttackBehavior attackBehavior, double? score)
		{
			return ModifyScore(attackBehavior, score);
		}

		/// <summary>
		/// Returns the debug info for this behavior, or null to print no info.
		/// </summary>
		string IAttackBehavior.GetDebugInfo()
		{
			return GetDebugInfo();
		}

		/// <summary>
		/// Returns the debug info for this behavior, or null to print no info.
		/// </summary>
		string IMoveBehavior.GetDebugInfo()
		{
			return GetDebugInfo();
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Sets the _curTank field and calls the OnTankChanged() method unless it has already been called this frame.
		/// </summary>
		private void CallOnTankChanged(ITank tank)
		{
			if(tank != _curTank || _game.Frame != _tankChangedFrame)
			{
				_curTank = tank;
				_tankChangedFrame = _game.Frame;
				OnTankChanged(tank);
			}
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected virtual void OnTankChanged(ITank tank)
		{

		}

		/// <summary>
		/// Calls the OnSelected() method if this behavior is currently unselected
		/// as both move and attack behaviors.
		/// </summary>
		private void CallOnSelected()
		{
			if(!_selectedAsMove && !_selectedAsAttack)
				OnSelected();
		}

		/// <summary>
		/// Called when this behavior is selected as either a move or attack behavior,
		/// but not if it is already selected.
		/// </summary>
		protected virtual void OnSelected()
		{

		}

		/// <summary>
		/// Called when this behavior is selected as a move behavior.
		/// </summary>
		protected virtual void OnSelectedAsMove()
		{

		}

		/// <summary>
		/// Called when this behavior is selected as an attack behavior.
		/// </summary>
		protected virtual void OnSelectedAsAttack()
		{

		}

		/// <summary>
		/// Calls the OnDeselected() method if this behavior is selected
		/// as one and only one of move and attack behavios.
		/// </summary>
		private void CallOnDeselected()
		{
			if((_selectedAsMove || _selectedAsAttack) && !(_selectedAsMove && _selectedAsAttack))
				OnDeselected();
		}

		/// <summary>
		/// Called when this behavior becomes not selected as either a move or attack behavior,
		/// but not if it is already deselected.
		/// </summary>
		protected virtual void OnDeselected()
		{

		}

		/// <summary>
		/// Called when this behavior is deselected as a move behavior.
		/// </summary>
		protected virtual void OnDeselectedAsMove()
		{

		}

		/// <summary>
		/// Called when this behavior is deselected as an attack behavior.
		/// </summary>
		protected virtual void OnDeselectedAsAttack()
		{

		}

		/// <summary>
		/// Calls the GetBehaviorChangeThreshold() method, but uses the cached threshold if it has already
		/// been called this frame.
		/// </summary>
		private double CallGetBehaviorChangeThreshold()
		{
			if(_game.Frame != _thresholdFrame)
			{
				_threshold = GetBehaviorChangeThreshold();
				_thresholdFrame = _game.Frame;
			}

			return _threshold;
		}

		/// <summary>
		/// Returns the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected virtual double GetBehaviorChangeThreshold()
		{
			return 0;
		}

		/// <summary>
		/// Returns whether this behavior is currently valid.
		/// </summary>
		protected virtual bool IsValidBehavior()
		{
			if(AITools.NeedsAmmo(_curTank))
				return false;
			return true;
		}

		/// <summary>
		/// Calls the GetScore() method, but uses the cached score if GetScore() has already been
		/// called this frame.
		/// </summary>
		private double? CallGetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			if(_game.Frame != _scoreFrame)
			{
				if(IsValidBehavior())
					_score = GetScore(curMoveBehavior, curAttackBehavior);
				else
					_score = null;
				_scoreFrame = _game.Frame;
			}

			return _score;
		}

		/// <summary>
		/// Gets the action score for this behavior, or null to skip this behavior.
		/// </summary>
		protected abstract double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior);

		/// <summary>
		/// Calls the BeforeGetActions() method, but not if it has already been call this frame.
		/// </summary>
		private void CallBeforeGetActions()
		{
			if(_game.Frame != _beforeGetActionsFrame)
			{
				BeforeGetActions();
				_beforeGetActionsFrame = _game.Frame;
			}
		}

		/// <summary>
		/// Called before GetMoveAction() or GetAttackAction() is called, but not if it has already
		/// been called on the other method this frame.
		/// </summary>
		protected virtual void BeforeGetActions()
		{

		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// </summary>
		protected abstract MoveAction GetMoveAction(MoveAction prevMoveAction);

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// </summary>
		protected abstract AttackAction GetAttackAction(AttackAction prevAttackAction);

		/// <summary>
		/// Called when this behavior is the current move behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="moveBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		protected virtual double? ModifyScore(IMoveBehavior moveBehavior, double? score)
		{
			return score;
		}

		/// <summary>
		/// Called when this behavior is the current attack behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="attackBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		protected virtual double? ModifyScore(IAttackBehavior attackBehavior, double? score)
		{
			return score;
		}

		/// <summary>
		/// Returns the debug info for this behavior.
		/// </summary>
		protected virtual string GetDebugInfo()
		{
			string scoreStr = _score != null ? ((int)_score).ToString() : "null";
			return GetType().Name + " (" + scoreStr + "/+" + _threshold + ")";
		}

		#endregion
	}
}