﻿using TankyTank.Game;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// Base class for move behaviors.
	/// </summary>
	public abstract class MoveBehaviorBase : IMoveBehavior
	{
		#region Constructors

		/// <summary>
		/// Initializes a new MoveBehaviorBase.
		/// </summary>
		protected MoveBehaviorBase(IGame game)
		{
			_game = game;
		}

		#endregion
		#region Fields

		protected readonly IGame _game;
		protected ITank _curTank;
		private double? _score;
		private double _behaviorChangeThreshold;

		#endregion
		#region IMoveBehavior Implementations

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		void IMoveBehavior.OnTankChanged(ITank tank)
		{
			_curTank = tank;
			OnTankChanged(tank);
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		void IMoveBehavior.OnSelected()
		{
			OnSelected();
		}

		/// <summary>
		/// Called when this behavior is deselected as the move behavior.
		/// </summary>
		void IMoveBehavior.OnDeselected()
		{
			OnDeselected();
		}

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		double IMoveBehavior.BehaviorChangeThreshold
		{
			get { return _behaviorChangeThreshold = GetBehaviorChangeThreshold(); }
		}

		/// <summary>
		/// Gets the move action score for this behavior, or null to skip this behavior.
		/// </summary>
		double? IMoveBehavior.GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			if(IsValidBehavior())
				return _score = GetScore(curMoveBehavior, curAttackBehavior);
			else
				return _score = null;
		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		MoveAction IMoveBehavior.GetAction(MoveAction prevMoveAction)
		{
			if(IsValidBehavior())
				return GetAction(prevMoveAction);
			else
				return null;
		}

		/// <summary>
		/// Called when this behavior is the current move behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="moveBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		double? IMoveBehavior.ModifyScore(IMoveBehavior moveBehavior, double? score)
		{
			return ModifyScore(moveBehavior, score);
		}

		/// <summary>
		/// Returns the debug info for this behavior, or null to print no info.
		/// </summary>
		string IMoveBehavior.GetDebugInfo()
		{
			return GetDebugInfo();
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected virtual void OnTankChanged(ITank tank)
		{

		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		protected virtual void OnSelected()
		{

		}

		/// <summary>
		/// Called when this behavior is deselected as the move behavior.
		/// </summary>
		protected virtual void OnDeselected()
		{

		}

		/// <summary>
		/// Returns the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected virtual double GetBehaviorChangeThreshold()
		{
			return 0;
		}

		/// <summary>
		/// Returns whether this behavior is currently valid.
		/// </summary>
		protected virtual bool IsValidBehavior()
		{
			return true;
		}

		/// <summary>
		/// Gets the move action score for this behavior, or null to skip this behavior.
		/// </summary>
		protected abstract double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior);

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		protected abstract MoveAction GetAction(MoveAction prevMoveAction);

		/// <summary>
		/// Called when this behavior is the current move behavior to optionally modify the score
		/// of another behavior.
		/// </summary>
		/// <param name="moveBehavior">The other behavior whose score is being modified.</param>
		/// <param name="score">The score produced by the behavior.</param>
		/// <returns>The modified score.</returns>
		protected virtual double? ModifyScore(IMoveBehavior moveBehavior, double? score)
		{
			return score;
		}

		/// <summary>
		/// Returns the debug info for this behavior.
		/// </summary>
		protected virtual string GetDebugInfo()
		{
			string scoreStr = _score != null ? ((int)_score).ToString() : "null";
			return GetType().Name + " (" + scoreStr + "/+" + _behaviorChangeThreshold + ")";
		}

		#endregion
	}
}