﻿using System;
using System.Collections.Generic;
using Halfling.Geometry;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game.Tanks;
using IGame = TankyTank.Game.IGame;

namespace TankyTank.AIBots
{
	/// <summary>
	/// A move/attack behavior that causes the tank to engage in combat with an enemy tank.
	/// </summary>
	public class RepairFriendBehavior : MoveAttackBehaviorBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for an RepairFriendBehavior.
		/// </summary>
		[AutoSerializable, DeserializeOnce] public class Settings
		{
			[Serial] public readonly double ScoreLow;
			[Serial] public readonly double ScoreHigh;
			[Serial] public readonly double BehaviorChangeThreshold;
			[Serial] public readonly double TargetChangeThreshold;

            /// <summary>
            /// Fractional distance this tank can "see".
            /// </summary>
            [Serial, Optional(1.0)] public readonly double RadarFraction;

			[Serial] public readonly Angle TacticalCone;
			[Serial] public readonly Angle TacticalConeSatisfied;
			[Serial] public readonly double TacticalForwardRadius;
			[Serial] public readonly double TacticalReverseRadius;

			[Serial, Optional(0)] public readonly int HesitateFramesBeforeFiringMin;
			[Serial, Optional(0)] public readonly int HesitateFramesBeforeFiringMax;
			[Serial, Optional(0)] public readonly int HesitateFramesBeforeMovingMin;
			[Serial, Optional(0)] public readonly int HesitateFramesBeforeMovingMax;
		}

		#endregion
		#region Constructors

		/// <summary>
		/// Loads a new RepairFriendBehavior.
		/// </summary>
		public RepairFriendBehavior(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
		}

		/// <summary>
		/// Creates a new RepairFriendBehavior.
		/// </summary>
		public RepairFriendBehavior(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
		}

		#endregion
		#region Fields

		protected readonly Settings _settings;
		protected ITank _target;
		protected int _hesitateFiringCountdown;
		protected int _hesitateMovingCountdown;

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected override double GetBehaviorChangeThreshold()
		{
			return _settings.BehaviorChangeThreshold;
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		protected override void OnSelected()
		{
			_target = null;
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected override void OnTankChanged(ITank tank)
		{
			if(tank != null)
				_target = null;
		}

		/// <summary>
		/// Returns whether this behavior is currently valid.
		/// </summary>
		protected override bool IsValidBehavior()
		{
			// Don't try to repair when out of ammo.
            if(AITools.NeedsAmmo(_curTank))
				return false;
			return true;
		}

		/// <summary>
		/// Returns whether the current target should be cleared under the current conditions.
		/// </summary>
		/// <returns>True if the target was cleared.</returns>
		protected virtual bool ShouldClearTarget()
		{
			return _target != null &&
			       (!AITools.Exists(_target, _game) ||
			        !AITools.AreAllied(_curTank, _target) ||
                    _target.Health == _target.Rules.Combat.MaxHealth || // Already at full health.
			        _target.Position.DistanceTo(_curTank.Location) > _curTank.Rules.RadarRange * _settings.RadarFraction ||
					(!_game.Map.IsPassable(_target.Location) && AITools.IsLosBlocked(_game, _curTank, _target)));
		}

		/// <summary>
		/// Returns the number of frames to hesitate firing after a new target is selected.
		/// </summary>
		protected virtual int GetHesitateFramesBeforeFiring()
		{
			return _game.Rand.Int32(_settings.HesitateFramesBeforeFiringMin,
			                        _settings.HesitateFramesBeforeFiringMax);
		}

		/// <summary>
		/// Returns the number of frames to hesitate before moving.
		/// </summary>
		protected virtual int GetHesitateFramesBeforeMoving()
		{
			return _game.Rand.Int32(_settings.HesitateFramesBeforeMovingMin,
			                        _settings.HesitateFramesBeforeMovingMax);
		}

		/// <summary>
		/// Returns the score for the specified tank.
		/// </summary>
		protected virtual double GetTankScore(ITank tank)
		{
			double baseScore = ((_curTank.Rules.RadarRange * _settings.RadarFraction) - _curTank.Location.DistanceTo(tank.Location)) / (_curTank.Rules.RadarRange * _settings.RadarFraction);
			
            // Go after the friendly tank more intently if it is very damaged.
            baseScore /= (double)tank.Health / tank.Rules.Combat.MaxHealth;
			
			return baseScore;
		}

		/// <summary>
		/// Returns the valid targets from which the target can be chosen.
		/// </summary>
		protected virtual IEnumerable<ITank> GetValidTargets()
		{
			Circle searchCircle = new Circle(_curTank.Location, (_curTank.Rules.RadarRange * _settings.RadarFraction));
			var found = _game.Tanks.FindIntersectsWith(searchCircle, _curTank.UniqueID);
			foreach(ITank tank in found)
			{
				// Don't heal enemies.
                if(!AITools.AreAllied(_curTank, tank))
					continue;
                // Don't heal friends you can't see.
				if(AITools.IsLosBlocked(_game, _curTank, tank))
					continue;
                // Don't heal friends that have full health.
                if(tank.Health == tank.Rules.Combat.MaxHealth)
                    continue;

				yield return tank;
			}
		}

		/// <summary>
		/// Sets _target field to the best target.
		/// </summary>
		/// <returns>The score of the target, or null if no target was selected.</returns>
		protected virtual double? PickBestTarget()
		{
			var found = GetValidTargets();

			ITank best = _target;
			double bestScore = best != null ? GetTankScore(best) : double.MinValue;
			double changeThreshold = best != null ? _settings.TargetChangeThreshold : 0;
			foreach(ITank tank in found)
			{
				double score = GetTankScore(tank);
				if(score > bestScore + changeThreshold)
				{
					best = tank;
					bestScore = score;
					changeThreshold = 0;
				}
			}

			if(_target != best)
			{
				_hesitateFiringCountdown = GetHesitateFramesBeforeFiring();
				_hesitateMovingCountdown = GetHesitateFramesBeforeMoving();
				_target = best;
			}

			return best != null ? bestScore : (double?)null;
		}

		/// <summary>
		/// Gets the score for this behavior, or null to skip this behavior.
		/// </summary>
		protected override double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			// Is this even a tank that can repair?
            if (!AITools.IsRepairTank(_curTank))
                return null;

            if(ShouldClearTarget())
				_target = null;

			double? score = PickBestTarget();
			if(score != null)
				return AITools.GetRangeBasedScore(score.Value, 0, 1, _settings.ScoreLow, _settings.ScoreHigh, true);
			else
				return null;
		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		protected override MoveAction GetMoveAction(MoveAction prevMoveAction)
		{
			if(ShouldClearTarget())
				_target = null;

			// Hesitate moving?
			if(_hesitateMovingCountdown > 0)
			{
				_hesitateMovingCountdown--;
				return null;
			}

			if(_target != null)
			{
				if(AITools.IsLosBlocked(_game, _curTank, _target))
					return new MoveAction.Target(_target.Location, null);
				else
					return GetTacticalMoveAction(prevMoveAction);
			}
			else
			{
				return null;
			}
		}

		/// <summary>
		/// Returns the move action for when the target tank is close enough to worry about
		/// about tactical movement instead of pathfinding.
		/// </summary>
		protected virtual MoveAction GetTacticalMoveAction(MoveAction prevMoveAction)
		{
			bool forward = false;
			bool reverse = false;
			bool left = false;
			bool right = false;

			Direction dirToTarget = _curTank.Location.DirectionTo(_target.Location);
			if(_curTank.ChassisRot.AngularDistanceTo(dirToTarget) > _settings.TacticalConeSatisfied / 2)
			{
				if(_curTank.ChassisRot.AngularSignTo(dirToTarget) < 0)
					left = true;
				else
					right = true;
			}

			double distSquaredToTarget = _curTank.Location.DistanceSquaredTo(_target.Location);
			double shortestRange = Math.Max(AITools.ShortestRange(_curTank), AITools.LongestMinRange(_curTank));
			if(distSquaredToTarget <= Math.Min(shortestRange * shortestRange, _settings.TacticalReverseRadius * _settings.TacticalReverseRadius))
			{
				reverse = true;
			}
			else if(distSquaredToTarget >= Math.Min(shortestRange * shortestRange, _settings.TacticalForwardRadius * _settings.TacticalForwardRadius) &&
				_curTank.ChassisRot.AngularDistanceTo(dirToTarget) <= _settings.TacticalCone)
			{
				forward = true;
			}

			return new MoveAction.Input(forward, reverse, left, right);
		}

		/// <summary>
		/// Gets the attack action for this behavior. The returned value will
		/// be either AttackAction.Target or AttackAction.Input depending
		/// on whether the action calls for shooting at a target tank using the
		/// built-in aiming and shooting or for manually controlling the attack input.
		/// Return null to perform no action.
		/// </summary>
		protected override AttackAction GetAttackAction(AttackAction prevAttackAction)
		{
			if(ShouldClearTarget())
				_target = null;

			// Hesitate firing?
			if(_hesitateFiringCountdown > 0)
			{
				_hesitateFiringCountdown--;
				return null;
			}
			
			if(_target != null)
				return new AttackAction.Target(_target);
			else
				return null;
		}

		#endregion
	}
}