﻿using System;
using System.Collections.Generic;
using Halfling.Geometry;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// A move behavior that causes the tank to retreat from enemies.
	/// The health of nearby friendly tanks is compared to the health of nearby enemy tanks
	/// to determine whether it is appropriate or not to retreat.
	/// </summary>
	public class RetreatBehavior : MoveBehaviorBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for an AttackEnemyBehavior.
		/// </summary>
		[AutoSerializable, DeserializeOnce]
		public class Settings
		{
			[Serial] public readonly double ScoreLow;
			[Serial] public readonly double ScoreHigh;
			[Serial] public readonly double BehaviorChangeThreshold;
			[Serial] public readonly double TargetChangeThreshold;

			/// <summary>
			/// Fractional distance this tank can "see".
			/// </summary>
			[Serial, Optional(1.0)] public readonly double RadarFraction;

			/// <summary>
			/// The units of health added to the friendly tank health
			/// calculation.  Use negative values for tanks with low self-esteem.
			/// </summary>
			[Serial] public readonly int TotalHealthComparisonThreshold;

			/// <summary>
			/// The wider region in front of the tank where the tank will attempt
			/// to turn to either the left or the right, away from an oncoming enemy.
			/// </summary>
			[Serial] public readonly Angle RetreatFrontCone;

			/// <summary>
			/// The narrower region in back of the tank where the tank will continue
			/// running forward in retreat without turning left or right.
			/// </summary>
			[Serial] public readonly Angle RetreatRearCone;


			[Serial, Optional(0)] public readonly int HesitateFramesBeforeMovingMin;
			[Serial, Optional(0)] public readonly int HesitateFramesBeforeMovingMax;

		}

		#endregion
		#region Constructors

		/// <summary>
		/// Loads a new RetreatBehavior.
		/// </summary>
		public RetreatBehavior(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
		}

		/// <summary>
		/// Creates a new RetreatBehavior.
		/// </summary>
		public RetreatBehavior(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
		}

		#endregion
		#region Fields

		protected readonly Settings _settings;

		/// <summary>
		/// Tank from which you will retreat.
		/// </summary>
		protected ITank _target;

		protected int _hesitateMovingCountdown;

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected override double GetBehaviorChangeThreshold()
		{
			return _settings.BehaviorChangeThreshold;
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		protected override void OnSelected()
		{
			_target = null;
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected override void OnTankChanged(ITank tank)
		{
			if(tank != null)
				_target = null;
		}

		/// <summary>
		/// Returns whether this behavior is currently valid.
		/// </summary>
		protected override bool IsValidBehavior()
		{
			return !AITools.NeedsAmmo(_curTank);
		}

		/// <summary>
		/// Returns whether the current target should be cleared under the current conditions.
		/// </summary>
		/// <returns>True if the target was cleared.</returns>
		protected virtual bool ShouldClearTarget()
		{
			return _target != null &&
				   (!AITools.Exists(_target, _game) ||
					AITools.AreAllied(_curTank, _target) ||
					_target.Invisibility == TankInvisibility.Full ||
					_target.Position.DistanceTo(_curTank.Location) > _curTank.Rules.RadarRange * _settings.RadarFraction);
		}

		/// <summary>
		/// Returns the score for the specified tank.
		/// </summary>
		protected virtual double GetTankScore(ITank tank)
		{
			// Score of 0 for far away tanks.  Score almost 1 for close tanks.
			return ((_curTank.Rules.RadarRange * _settings.RadarFraction) - _curTank.Location.DistanceTo(tank.Location)) / (_curTank.Rules.RadarRange * _settings.RadarFraction);
		}

		/// <summary>
		/// Returns the valid targets from which the target can be chosen.
		/// Search for enemy tanks to return as possible foes from which to retreat.
		/// </summary>
		protected virtual IEnumerable<ITank> GetValidTargets(IEnumerable<ITank> tanksInRange)
		{
			foreach(ITank tank in tanksInRange)
			{
				if(AITools.AreAllied(_curTank, tank))
					continue;

				// For the moment, bad guys can still be scary behind walls.
				//if(AITools.IsLosBlocked(_game, _curTank, tank))
				//	continue;

				if(tank.Invisibility == TankInvisibility.Full)
					continue;

				yield return tank;
			}
		}

		/// <summary>
		/// Sets _target field to the best target.
		/// </summary>
		/// <returns>The score of the target, or null if no target was selected.</returns>
		protected virtual double? PickBestTarget(IEnumerable<ITank> validTargets)
		{
			// Decide from whom this _curTank should run if indeed _curTank should run.
			ITank best = _target;
			double bestScore = best != null ? GetTankScore(best) : double.MinValue;
			double changeThreshold = best != null ? _settings.TargetChangeThreshold : 0;
			foreach(ITank tank in validTargets)
			{
				double score = GetTankScore(tank);
				if(score > bestScore + changeThreshold)
				{
					best = tank;
					bestScore = score;
					changeThreshold = 0;
				}
			}

			if(_target != best)
			{
				_hesitateMovingCountdown = GetHesitateFramesBeforeMoving();
				_target = best;
			}
			return best != null ? bestScore : (double?)null;
		}

		/// <summary>
		/// Determines whether or not _curTank should attempt to retreat.
		/// </summary>
		protected virtual bool ShouldRetreat(IEnumerable<ITank> tanksInRange)
		{
			// Create a list of all nearby tanks and their associated health points.
			// Sum the health scores of all tanks on the same team.
			Dictionary<int, int> teamHealthScores = new Dictionary<int, int>
			{
				{_curTank.Team, _curTank.Health}
			};

			foreach(ITank tank in tanksInRange)
			{
				if(!teamHealthScores.ContainsKey(tank.Team))
					teamHealthScores.Add(tank.Team, tank.Health);
				else
					teamHealthScores[tank.Team] += tank.Health;
			}

			// Compare our health to each other team.
			// If the check passes for at least one team then we want to retreat.
			int friendlyHealthScore = teamHealthScores[_curTank.Team];
			foreach(var keyAndValue in teamHealthScores)
			{
				int team = keyAndValue.Key;
				int totalHealthScore = keyAndValue.Value;
				if(team != _curTank.Team && totalHealthScore > friendlyHealthScore + _settings.TotalHealthComparisonThreshold)
					return true;
			}

			// If we get here then the check didn't pass for any teams, so we don't want to retreat.
			return false;
		}

		/// <summary>
		/// Gets the score for this behavior, or null to skip this behavior.
		/// </summary>
		protected override double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			if(ShouldClearTarget())
				_target = null;

			Circle searchCircle = new Circle(_curTank.Location, (_curTank.Rules.RadarRange * _settings.RadarFraction));
			var tanksInRange = _game.Tanks.FindIntersectsWith(searchCircle, _curTank.UniqueID);

			if(!ShouldRetreat(tanksInRange))
				return null;

			// PickBestTarget sets _target.
			var score = PickBestTarget(GetValidTargets(tanksInRange));

			if(score != null)
				return AITools.GetRangeBasedScore(score.Value, 0, 1, _settings.ScoreLow, _settings.ScoreHigh, true);
			else
				return null;
		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		protected override MoveAction GetAction(MoveAction prevMoveAction)
		{
			if(ShouldClearTarget())
				_target = null;

			// Hesitate moving?
			if(_hesitateMovingCountdown > 0)
			{
				_hesitateMovingCountdown--;
				return null;
			}

			if(_target != null)
				return GetRetreatMoveAction(prevMoveAction);
			else
				return null;
		}

		/// <summary>
		/// Returns the move action to retreat away from the target tank regardless of
		/// the AI tank's proximity to the enemy tank.
		/// </summary>
		protected virtual MoveAction GetRetreatMoveAction(MoveAction prevMoveAction)
		{
			bool forward = false;
			bool reverse = false;
			bool left = false;
			bool right = false;

			Direction dirToTarget = _curTank.Location.DirectionTo(_target.Location);
			Angle angleDifference = _curTank.ChassisRot.AngularDistanceTo(dirToTarget);
			Angle angleSign = _curTank.ChassisRot.AngularSignTo(dirToTarget);
			
			if(angleDifference < _settings.RetreatFrontCone / 2)
			{
				// The predator is right in front of you.  Back up, but do you turn left or right?
				// (There is no back straight up.)(
				reverse = true;

				// If the predator is to your left, turn right - that is, Clockwise.
				// If the predator is to your right, turn left - that is, Walt-Clockwise.
				if(angleSign < 0)
					right = true;
				else
					left = true;
			}
			else if((Math.PI - angleDifference) < _settings.RetreatRearCone / 2)
			{
				// The predator is directly behind you.  Just run as fast as possible.
				forward = true;
			}
			else
			{
				// The predator is off to the side of you.  Go forward and keep turning.
				forward = true;

				// If the predator is to your left, turn right - that is, Clockwise.
				// If the predator is to your right, turn left - that is, Walt-Clockwise.
				if(angleSign < 0)
					right = true;
				else
					left = true;
			}

			return new MoveAction.Input(forward, reverse, left, right);
		}

		/// <summary>
		/// Returns the number of frames to hesitate before moving.
		/// </summary>
		protected virtual int GetHesitateFramesBeforeMoving()
		{
			return _game.Rand.Int32(_settings.HesitateFramesBeforeMovingMin,
									_settings.HesitateFramesBeforeMovingMax);
		}

		#endregion
	}
}