﻿using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Tanks;

namespace TankyTank.AIBots
{
	/// <summary>
	/// A move behavior that causes the tank to move to random locations on the map.
	/// </summary>
	public class WanderAimlesslyBehavior : MoveBehaviorBase
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for a WanderAimlesslyBehavior.
		/// </summary>
		[AutoSerializable, DeserializeOnce] public class Settings
		{
			[Serial] public readonly double Score;
			[Serial] public readonly double BehaviorChangeThreshold;
			[Serial] public readonly double TargetLocationReachedSatisfied;
		}

		#endregion
		#region Constructors

		/// <summary>
		/// Loads a new WanderAimlesslyBehavior.
		/// </summary>
		public WanderAimlesslyBehavior(IGame game, INode node, Serializer ser)
			: base(game)
		{
			_settings = ser.TreeDeserialize<Settings>(node, "Settings");
		}

		/// <summary>
		/// Creates a new WanderAimlesslyBehavior.
		/// </summary>
		public WanderAimlesslyBehavior(IGame game, Settings settings)
			: base(game)
		{
			_settings = settings;
		}

		#endregion
		#region Fields

		protected readonly Settings _settings;
		protected Vector _targetLoc;

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the threshold by which another behavior's score must exceed this behavior's score
		/// in order to change behaviors.
		/// </summary>
		protected override double GetBehaviorChangeThreshold()
		{
			return _settings.BehaviorChangeThreshold;
		}

		/// <summary>
		/// Gets the move action score for this behavior, or null to skip this behavior.
		/// </summary>
		protected override double? GetScore(IMoveBehavior curMoveBehavior, IAttackBehavior curAttackBehavior)
		{
			return _settings.Score;
		}

		/// <summary>
		/// Called when this behavior is selected as the move behavior.
		/// </summary>
		protected override void OnSelected()
		{
			SelectNewTarget();
		}

		/// <summary>
		/// Called when the tank associated with this behavior is changed.
		/// </summary>
		protected override void OnTankChanged(ITank tank)
		{
			if(tank != null)
				SelectNewTarget();
		}

		/// <summary>
		/// Gets the move action for this behavior. The returned value will
		/// be either MoveAction.Target or MoveAction.Input depending
		/// on whether the action calls for moving to a point/direction using the
		/// built-in movement and pathfinding or for manually controlling the movement input.
		/// Return null to perform no action.
		/// </summary>
		protected override MoveAction GetAction(MoveAction prevMoveAction)
		{
			double targetLocationReachedSatisfiedSquared = _settings.TargetLocationReachedSatisfied * _settings.TargetLocationReachedSatisfied;
			if(_curTank.Location.DistanceSquaredTo(_targetLoc) <= targetLocationReachedSatisfiedSquared)
				SelectNewTarget();

			return new MoveAction.Target(_targetLoc, null);
		}

		/// <summary>
		/// Selects a new target location.
		/// </summary>
		protected virtual void SelectNewTarget()
		{
			// Hack to prevent crashing. A real solution needs pathfinding
			// to handle doodads better.
			if(_game.Map.IsPassable(_curTank.Location))
			{
				Rect mapRect = _game.Map.BoundingRect;
				do
				{
					_targetLoc = new Vector(
						_game.Rand.Double(mapRect.Left, mapRect.Right - 1),
						_game.Rand.Double(mapRect.Top, mapRect.Bottom - 1));
				}
				while(!_game.Map.IsPassable(new Circle(_targetLoc, _curTank.Rules.Radius)) ||
				      !_game.Map.CheckPathExists(_curTank.Location, _targetLoc));
			}
		}

		#endregion
	}
}