// Walt Destler

using System;
using Halfling.Audio;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Bullets;
using TankyTank.Game.Tanks;

namespace TankyTank.Modes
{
	/// <summary>
	/// Base client game mode logic.
	/// </summary>
	public class BaseClient : ModeBase, IClientMode
	{
		#region Strings and Sounds

		/// <summary>
		/// Stores txt-loaded strings for the client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Strings
		{
			[Serial] public readonly string LivesLimitReached;
			[Serial] public readonly string ScoreLimitReached;
			[Serial] public readonly string TimeLimitReached;
			[Serial] public readonly string LivesRemainingFormat;
			[Serial] public readonly string CannotSpawnNoLives;
		}

		/// <summary>
		/// Stores txt-loaded sound effects for the client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Sounds
		{
			[Serial] public readonly Effect TimeLimitReached;
			[Serial] public readonly Effect LivesLimitReached;
			[Serial] public readonly Effect ScoreLimitReached;
		}

		#endregion
		#region Non-Public Fields

		protected new IClientGame game;
		private Strings strings;
		private Sounds sounds;

		private bool playingEndSound = false;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new base game mode.
		/// </summary>
		public BaseClient(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			strings = ser.TreeDeserialize<Strings>(file, "Strings");
			sounds = ser.TreeDeserialize<Sounds>(file, "Sounds");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called once when the game is started.
		/// </summary>
		public virtual void StartGame(IClientGame game, byte[] initData)
		{
			ModeBaseSettings modeSettings = ser.BinaryDeserialize<ModeBaseSettings>(initData);
			base.StartGame(game, modeSettings);
			this.game = game;

			game.InfoBoxOkayVisible = false;
		}

		/// <summary>
		/// Called when a message from the server mode is received.
		/// </summary>
		public virtual void ReceiveServerModeMessage(byte[] message)
		{
			// Do nothing.
		}

		/// <summary>
		/// Updates the game-mode specific logic.
		/// </summary>
		public override void Update()
		{
			base.Update();

			// Show info text?
			string infoText = getInfoText();
			if(infoText != null && infoText != "")
			{
				game.InfoBoxVisible = true;
				game.InfoBoxText = infoText;
			}
			else if(infoText != null)
			{
				game.InfoBoxVisible = false;
			}
		}

		/// <summary>
		/// Adds game mode specific sprites.
		/// </summary>
		public virtual void AddSprites()
		{

		}

		/// <summary>
		/// Draws game mode specific user interface graphics.
		/// </summary>
		public virtual void DrawUI()
		{

		}

		/// <summary>
		/// Draws game mode specific minimap graphics.
		/// </summary>
		public virtual void DrawMinimap()
		{
			
		}

		/// <summary>
		/// Gets the end game text for the specified ID from the server.
		/// </summary>
		public virtual string GetEndText(GameEndCause cause, int endID)
		{
			IEffect endSound = null;
			string ret;

			switch(endID)
			{
			case (int)EndID.LivesLimit:
				ret = strings.LivesLimitReached;
				endSound = sounds.LivesLimitReached;
				break;
			case (int)EndID.ScoreLimit:
				ret = strings.ScoreLimitReached;
				endSound = sounds.ScoreLimitReached;
				break;
			case (int)EndID.TimeLimit:
				endSound = sounds.TimeLimitReached;
				ret = strings.TimeLimitReached;
				break;
			default:
				throw new ApplicationException("Unknown end ID '" + endID + "'.");
			}

			if(endSound != null && !playingEndSound)
			{
				game.PlaySound(endSound, true);
				playingEndSound = true;
			}

			return ret;
		}

		/// <summary>
		/// Returns the instructional text to be displayed on the Infoboard.
		/// </summary>
		public virtual string GetInfoboardInstructions()
		{
			return instructions;
		}

		/// <summary>
		/// Returns the text to be displayed on the left of the Infoboard.
		/// </summary>
		public virtual string GetInfoboardTextLeft()
		{
			int livesLimit = modeSettings.LivesLimitEnable ? modeSettings.LivesLimit : 0;
			int scoreLimit = modeSettings.ScoreLimitEnable ? modeSettings.ScoreLimit : 0;
			int timeLimit = modeSettings.TimeLimitEnable ? modeSettings.TimeLimit : 0;
			string livesLimitText = "Lives Limit: " + (livesLimit != 0 ? livesLimit.ToString() : "None");
			string scoreLimitText = "Score Limit: " + (scoreLimit != 0 ? scoreLimit.ToString() : "None");
			string timeLimitText = "Time Limit: " + (timeLimit != 0 ? timeLimit.ToString() : "None");
			string randomTanksText = "Random Tanks: " + (modeSettings.ForceRandomTanks ? "Yes" : "No");
			string respawnDelayText = "Respawn Delay: " + modeSettings.RespawnDelay + " sec.";
			return livesLimitText + "\n" +
			       scoreLimitText + "\n" +
			       timeLimitText + "\n" +
				   respawnDelayText + "\n" +
			       randomTanksText;
		}

		/// <summary>
		/// Returns the text to be displayed on the right of the Infoboard.
		/// </summary>
		public virtual string GetInfoboardTextRight()
		{
			string killPointsText = "Kill Points: " + modeSettings.KillPoints;
			string deathPointsText = "Death Points: " + modeSettings.DeathPoints;
			string oopsPointsText = "Oops Points: " + modeSettings.OopsPoints;
			return killPointsText + "\n" +
			       deathPointsText + "\n" +
			       oopsPointsText;
		}

		/// <summary>
		/// Called when the user clicks the okay button on the in-game information box.
		/// </summary>
		public virtual void OnInfoBoxOkayClicked()
		{

		}

		/// <summary>
		/// Returns whether the specified tank should be drawn on the minimap.
		/// </summary>
		public virtual bool ShouldDrawMinimapTank(ITank tank)
		{
			// Invisible?
			if(tank.Invisibility != TankInvisibility.None && (tank.Team == 0 || tank.Team != game.UserPlayer.Team))
				return false;

			if(tank.Team == 0 || tank.Team != game.UserPlayer.Team)
			{
				return UnrestrictedMinimap ||
					((game.UserTank != null && tank.Location.DistanceSquaredTo(game.UserTank.Location) <= game.UserTank.Rules.RadarRangeSquared) &&
					!game.Roofs.PointIsRadarBlocked(tank.Location));
			}
			else
			{
				return true;
			}
		}

		/// <summary>
		/// Returns whether the specified bullet should be drawn on the minimap.
		/// </summary>
		public virtual bool ShouldDrawMinimapBullet(IBullet bullet)
		{
			if(!bullet.Rules.ShowOnRadar)
			{
				return false;
			}
			else if(bullet.Team == 0 || bullet.Team != game.UserPlayer.Team)
			{
				return UnrestrictedMinimap ||
					((game.UserTank != null && bullet.Location.DistanceSquaredTo(game.UserTank.Location) <= game.UserTank.Rules.RadarRangeSquared) &&
					!game.Roofs.PointIsRadarBlocked(bullet.Location));
			}
			else
			{
				return true;
			}
		}

		/// <summary>
		/// Gets whether the user is allowed to scroll the view when dead.
		/// </summary>
		public virtual bool AllowScrolling
		{
#if !DEBUG
			get { return !CanSpawn; }
#else
			get { return true; }
#endif
		}

		/// <summary>
		/// Gets whether the player can currently request to spawn.
		/// </summary>
		public virtual bool CanSpawn
		{
			get
			{
				if(modeSettings.LivesLimitEnable)
					return game.UserPlayer.Deaths < modeSettings.LivesLimit;
				else
					return true;
			}
		}

		/// <summary>
		/// Gets whether all tanks are shown on the radar.
		/// </summary>
		public virtual bool UnrestrictedMinimap
		{
			get { return !CanSpawn; }
		}

		/// <summary>
		/// Gets the time limit (in minutes), or null if there is no time limit.
		/// </summary>
		public virtual int? TimeLimit
		{
			get
			{
				if(modeSettings.TimeLimitEnable)
					return modeSettings.TimeLimit;
				else
					return null;
			}
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Returns the text to display in the info box, or null for no text.
		/// </summary>
		protected virtual string getInfoText()
		{
			if(modeSettings.LivesLimitEnable)
			{
				if(CanSpawn)
				{
					int minusOne = game.UserTank != null ? 1 : 0;
					return string.Format(strings.LivesRemainingFormat, modeSettings.LivesLimit - game.UserPlayer.Deaths - minusOne);
				}
				else
				{
					return strings.CannotSpawnNoLives;
				}
			}
			else
			{
				return null;
			}
		}

		#endregion
	}
}