// Walt Destler

using System;
using System.Collections.Generic;
using System.Drawing;
using Halfling.Geometry;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.Timing;
using Halfling.TreeData;
using TankyTank.AI;
using TankyTank.Game;
using TankyTank.Game.Doodads;
using TankyTank.Game.Players;
using TankyTank.Game.Tanks;
using TankyTank.Game.Teams;

namespace TankyTank.Modes
{
	/// <summary>
	/// Base server game mode logic.
	/// </summary>
	public class BaseServer : ModeBase, IServerMode
	{
		#region Non-Public Fields

		protected new IServerGame game;

		protected List<TankRespawn> tankRespawns = new List<TankRespawn>();
		private Time scoreLimitReachedTime = Time.MaxValue;
		private Time livesLimitReachedTime = Time.MaxValue;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new base game mode.
		/// </summary>
		public BaseServer(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			// Do nothing.
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called once when the game is started.
		/// </summary>
		public virtual void StartGame(IServerGame game, INode modeSettingsNode)
		{
			ModeBaseSettings modeSettings = ser.TreeDeserialize<ModeBaseSettings>(modeSettingsNode);
			base.StartGame(game, modeSettings);
			this.game = game;
		}

		/// <summary>
		/// Returns the mode initialization data.
		/// </summary>
		public virtual byte[] GetInitData()
		{
			return ser.BinarySerialize(modeSettings);
		}

		/// <summary>
		/// Called when a message is received from a client mode.
		/// </summary>
		public virtual void ReceiveClientModeMessage(byte[] message, int client)
		{
			// Do nothing.
		}

		/// <summary>
		/// Called when an initial spawn request is received from a client.
		/// </summary>
		public virtual void ReceiveSpawnRequest(int client, int player)
		{
			if(game.Players[player].Owner != client)
				throw new ApplicationException("The specified player is not owned by the specified client.");

			if(CanSpawn(player))
			{
				// Check to see if the player is waiting to respawn.
				foreach(TankRespawn tr in tankRespawns)
				{
					if(tr.Tank.Player == player)
					{
						if(tr.Tank.Owner != client)
							throw new ApplicationException("The specified player is not owned by the specified client.");

						tr.Requested = true;
						return;
					}
				}

				// Make sure the specified player isn't in the game.
				foreach(ITank tank in game.Tanks)
				{
					if(tank.Player == player)
						return; // Simply throw the request out.
				}

				// If we get here, then the player has not yet ever spawned, so create a new tank immediately.
				int team = game.Players[player].Team;
				Color color = game.Players[player].Color;
				int type = getNewTankType(player, team, false);
				TankPos pos = getNewTankPos(player, team, type);
				game.SpawnNewTank(type, client, color, player, team, pos);
			}
		}

		/// <summary>
		/// Called when a respawn request is received from a client.
		/// </summary>
		public virtual void ReceiveRespawnRequest(int client, int tankID)
		{
			foreach(TankRespawn tr in tankRespawns)
			{
				if(tr.Tank.UniqueID == tankID)
				{
					if(tr.Tank.Owner != client)
						throw new ApplicationException("The specified tank is not owned by the specified client.");

					tr.Requested = true;
					return;
				}
			}
		}

		/// <summary>
		/// Called to spawn a new player.
		/// </summary>
		public virtual void SpawnNewPlayer(int owner, string name, Color color, int player, int team, int preferredTank)
		{
			game.CreateNewPlayer(player, owner, name, color, team, false, preferredTank);
		}

		/// <summary>
		/// Called to spawn a new AI player.
		/// </summary>
		public virtual void SpawnAIPlayer(IBot bot, string name, Color color, int team, int preferredTank)
		{
			if(game.Teams.Count == 0)
				team = 0;

			int player = game.Rand.Int32();
			game.CreateNewPlayer(player, game.Owner, name, color, team, true, preferredTank);
			int type = getNewTankType(player, team, false);
			TankPos pos = getNewTankPos(player, team, type);
			int id = game.SpawnNewTank(type, game.Owner, color, player, team, pos);
			game.AttachAIBot(id, bot);
		}

		/// <summary>
		/// Updates the game-mode specific logic.
		/// </summary>
		public override void Update()
		{
			base.Update();
			respawnTanks();
			checkLivesLimit();
			checkScoreLimit();
			checkTimeLimit();
		}

		/// <summary>
		/// Called when a tank is killed.
		/// </summary>
		/// <param name="killedByZero">Whether the tank was killed by an inconsequential player.</param>
		public override void OnTankKilled(ITank tank, bool killedByZero)
		{
			base.OnTankKilled(tank, killedByZero);

			// Respawn the tank.
			if(game.Players.Exists(tank.Player) && CanSpawn(tank.Player))
				tankRespawns.Add(new TankRespawn(tank, RespawnDelay));

			// Drop a doodad?
			if(!killedByZero && game.Rules.General.DoodadDropChance > 0 && game.Rand.Double(0, 1) <= game.Rules.General.DoodadDropChance)
			{
				int type = pickDroppableDoodad();
				DoodadRules dRules = game.Rules.Doodads[type];
				game.SpawnNewDoodad(
					type,
					tank.Location,
					false,
					dRules.GetRandomSpriteIndex(game.Rand),
					dRules.GetRandomScale(game.Rand),
					dRules.GetRandomOrientation(game.Rand));
			}
		}

		/// <summary>
		/// Called to verify that the specified doodad should be collected by the specified tank.
		/// </summary>
		public virtual bool VerifyDoodadCollect(ITank tank, IDoodad doodad)
		{
			return true;
		}

		/// <summary>
		/// Called when damage is inflicted on a tank. Returns the damage to actually inflict.
		/// </summary>
		public virtual int ModifyDamage(int damage, int victimTank, int sourceTank, int sourcePlayer, int sourceTeam)
		{
			ITank victim = game.Tanks[victimTank];

			// Don't damage if teams are same and friendly fire is off.
			if(!modeSettings.FriendlyFire && victim.Team != 0 && victim.Team == sourceTeam)
				return 0;
			else
				return damage;
		}

		/// <summary>
		/// Returns whether the specified player can currently request to spawn.
		/// </summary>
		public virtual bool CanSpawn(int player)
		{
			if(modeSettings.LivesLimitEnable)
				return game.Players[player].Deaths < modeSettings.LivesLimit;
			else
				return true;
		}

		#endregion
		#region Protected Methods

		/// <summary>
		/// Respawns any tanks that need respawning.
		/// </summary>
		private void respawnTanks()
		{
			List<TankRespawn> respawned = new List<TankRespawn>();
			foreach(TankRespawn tr in tankRespawns)
			{
				if(tr.Countdown <= 0)
				{
					if(tr.Requested)
					{
						respawned.Add(tr);
						int team = game.Players[tr.Tank.Player].Team;
						int type = getNewTankType(tr.Tank.Player, team, false);
						TankPos pos = getNewTankPos(tr.Tank.Player, team, type);
						game.RespawnTank(tr.Tank, type, pos);
					}
				}
				else
				{
					tr.Countdown--;
				}
			}

			foreach(TankRespawn tr in respawned)
				tankRespawns.Remove(tr);
		}

		/// <summary>
		/// Returns whether the scenario was completed and the GameEndCause should indicate thusly.
		/// </summary>
		protected virtual bool wasScenarioCompleted()
		{
			return false;
		}

		/// <summary>
		/// Checks to see if the lives limit has been reached.
		/// </summary>
		protected virtual void checkLivesLimit()
		{
			bool livesLimitReached = false;
			bool?[] teams = new bool?[Constants.MaxTeams + 1];
			for(int i = 0; i < teams.Length; i++)
				teams[i] = null;
			int livingPlayers = 0;
			int deadPlayers = 0;

			// Check each player.
			foreach(IPlayer player in game.Players)
			{
				if(CanSpawn(player.UniqueID))
				{
					livingPlayers++;
					teams[player.Team] = true;
				}
				else
				{
					deadPlayers++;
					if(teams[player.Team] == null)
						teams[player.Team] = false;
				}
			}

			// FFA?
			if(game.Teams.Count == 0)
			{
				if(livingPlayers <= 1 && deadPlayers > 0)
					livesLimitReached = true;
			}

			// Teams.
			else
			{
				int livingTeams = 0;
				int deadTeams = 0;
				foreach(bool? status in teams)
				{
					if(status == true)
						livingTeams++;
					else if(status == false)
						deadTeams++;
				}

				if(livingTeams <= 1 && deadTeams > 0)
					livesLimitReached = true;
			}

			if(livesLimitReached)
			{
				// Set time at which game should end.
				if(livesLimitReachedTime == Time.MaxValue)
					livesLimitReachedTime = game.ClockTime;

				// End game if necessary.
				if(livesLimitReachedTime < game.ClockTime - new Time(.5))
					game.EndGame(wasScenarioCompleted() ? GameEndCause.ScenarioCompleted : GameEndCause.Mode, (int)EndID.LivesLimit);
			}
			else
			{
				livesLimitReachedTime = Time.MaxValue;
			}
		}

		/// <summary>
		/// Checks to see if the kill limit has been reached.
		/// </summary>
		protected virtual void checkScoreLimit()
		{
			if(modeSettings.ScoreLimitEnable)
			{
				bool scoreLimitReached = false;

				if(game.Teams.Count > 0 && teamScoreMatters())
				{
					// Check teams' scores.
					foreach(ITeam t in game.Teams)
					{
						if(t.Score >= modeSettings.ScoreLimit)
							scoreLimitReached = true;
					}
				}
				else
				{
					// Check players' scores.
					foreach(IPlayer p in game.Players)
					{
						if(p.Score >= modeSettings.ScoreLimit)
							scoreLimitReached = true;
					}
				}

				if(scoreLimitReached)
				{
					// Set time at which game should end.
					if(scoreLimitReachedTime == Time.MaxValue)
						scoreLimitReachedTime = game.ClockTime;

					// End game if necessary.
					if(scoreLimitReachedTime < game.ClockTime - new Time(.5))
						game.EndGame(wasScenarioCompleted() ? GameEndCause.ScenarioCompleted : GameEndCause.Mode, (int)EndID.ScoreLimit);
				}
				else
				{
					scoreLimitReachedTime = Time.MaxValue;
				}
			}
		}

		/// <summary>
		/// Returns whether the team score should be taken into consideration when checking for game end.
		/// </summary>
		protected virtual bool teamScoreMatters()
		{
			return game.Teams.Count > 0;
		}

		/// <summary>
		/// Checks to see if the time limit has been reached.
		/// </summary>
		protected virtual void checkTimeLimit()
		{
			if(modeSettings.TimeLimitEnable && game.ElapsedTime >= modeSettings.TimeLimit * Time.FromMinutes(1))
				game.EndGame(wasScenarioCompleted() ? GameEndCause.ScenarioCompleted : GameEndCause.Mode, (int)EndID.TimeLimit);
		}

		/// <summary>
		/// Gets a new tank type.
		/// </summary>
		protected virtual int getNewTankType(int player, int team, bool forceRandom)
		{
			if(!forceRandom && !modeSettings.ForceRandomTanks && game.Players.Exists(player) && game.GetPreferredTank(player) != -1)
				return game.GetPreferredTank(player);
			else
				return game.Rand.Int32(0, game.Rules.Tanks.Length - 1);
		}

		/// <summary>
		/// Gets the position of a new tank, but with the specified turret rot.
		/// </summary>
		protected virtual TankPos getNewTankPos(int player, int team, int tankType)
		{
			// Find spawn points...
			List<IDoodad> spawnPoints = new List<IDoodad>();

			// Find team spawn point.
			if(team != 0 && teamSpawnsMatter())
			{
				foreach(IDoodad d in game.Doodads)
				{
					if(d.Rules.SpawnPoint && d.Rules.SpawnTeam == team)
						spawnPoints.Add(d);
				}
			}

			// Find generic spawn point if no team spawn was found.
			if(spawnPoints.Count == 0)
			{
				foreach(IDoodad d in game.Doodads)
				{
					if(d.Rules.SpawnPoint && d.Rules.SpawnTeam == 0)
						spawnPoints.Add(d);
				}
			}

			// Get location.
			Vector loc;
			if(spawnPoints.Count > 0)
			{
				// Choose a random spawn point.
				//Doodad sp = spawnPoints[app.Rand.Int32(0, spawnPoints.Count - 1)];
				IDoodad sp = chooseBestSpawnPoint(spawnPoints, team);

				// Randomly pick a location with that spawn point.
				do
				{
					Direction dir = game.Rand.Direction();
					double mag = game.Rand.Double(0, sp.Rules.SpawnRadius);
					loc = sp.Location + Vector.FromAngle(dir) * mag;
				}
				while(!isValidSpawnLoc(loc, tankType));
			}
			else
			{
				// No spawn points were found, so choose anywhere on the map.
				do
				{
					double radius = game.Rules.Tanks[tankType].TileCollisionRadius;
					loc = new Vector(
							game.Rand.Double(game.Map.BoundingRect.Left + radius, game.Map.BoundingRect.Right - radius),
							game.Rand.Double(game.Map.BoundingRect.Top + radius, game.Map.BoundingRect.Bottom - radius));
				}
				while(!isValidSpawnLoc(loc, tankType));
			}

			// Return.
			return new TankPos(
				loc,
				Vector.Zero,
				//app.Rand.PickSingle(Direction.MinValue, Direction.MaxValue),
				chooseBestSpawnDirection(loc, team, tankType),
				Angle.Zero,
				loc);
		}

		/// <summary>
		/// Returns whether tanks should spawn on their team spawn points.
		/// </summary>
		protected virtual bool teamSpawnsMatter()
		{
			return true;
		}

		/// <summary>
		/// Chooses the best spawn point from the specified list of spawn points.
		/// </summary>
		protected virtual IDoodad chooseBestSpawnPoint(List<IDoodad> spawnPoints, int team)
		{
			// Find the spawn points with the least number of other tanks around them.
			List<IDoodad> best = new List<IDoodad>();
			int bestEnemies = int.MaxValue;
			foreach(IDoodad sp in spawnPoints)
			{
				// Count number of other tanks.
				int count = 0;
				foreach(ITank t in game.Tanks.FindIntersectsWith(new Circle(sp.Location, game.Rules.General.SpawnLogicProximity)))
				{
					//if(team == 0 || t.Team != team)
						count++;
				}

				if(count < bestEnemies)
				{
					best = new List<IDoodad>();
					best.Add(sp);
					bestEnemies = count;
				}
				else if(count == bestEnemies)
				{
					best.Add(sp);
				}
			}

			// Return a random spawn point from the best spawn points found.
			return best[game.Rand.Int32(0, best.Count - 1)];
		}

		/// <summary>
		/// Chooses the best direction for a tank to face when it spawns.
		/// </summary>
		protected virtual Direction chooseBestSpawnDirection(Vector spawnLoc, int team, int tankType)
		{
			// If tank is close, spawn facing it.
			ITank closest = null;
			double closestDistSquared = double.MaxValue;
			Circle searchCircle = new Circle(spawnLoc, game.Rules.Tanks[tankType].RadarRange);
			foreach(ITank t in game.Tanks.FindIntersectsWith(searchCircle))
			{
				if(team != 0 && team == t.Team)
					continue;

				double distSquared = spawnLoc.DistanceSquaredTo(t.Location);
				if(distSquared < closestDistSquared)
				{
					closest = t;
					closestDistSquared = distSquared;
				}
			}

			if(closest != null)
				return spawnLoc.DirectionTo(closest.Location);
			else
				return spawnLoc.DirectionTo(game.Map.BoundingRect.Center);
		}

		/// <summary>
		/// Returns whether the specified location is a valid spawn point.
		/// </summary>
		protected virtual bool isValidSpawnLoc(Vector loc, int tankType)
		{
			Circle pos = new Circle(loc, game.Rules.Tanks[tankType].TileCollisionRadius);

			if(!game.Map.BoundingRect.Contains(pos))
				return false;

			return game.Map.IsPassable(pos);
		}

		/// <summary>
		/// Picks a random droppable doodad.
		/// </summary>
		protected virtual int pickDroppableDoodad()
		{
			int index;

			do
			{
				index = game.Rand.Int32(0, game.Rules.Doodads.Count - 1);
			}
			while(!game.Rules.Doodads.GetAtIndex(index).Droppable);

			return game.Rules.Doodads.KeyOf(index);
		}

		#endregion
		#region Private Types

		/// <summary>
		/// Keeps track of a respawning tank.
		/// </summary>
		protected class TankRespawn
		{
			public readonly ITank Tank;
			public int Countdown;
			public bool Requested = false;

			/// <summary>
			/// Initializes a new TankRespawn.
			/// </summary>
			public TankRespawn(ITank tank, int countdown)
			{
				this.Tank = tank;
				this.Countdown = countdown;
			}
		}

		#endregion
	}

	/// <summary>
	/// Specifies to the client how the game was ended.
	/// </summary>
	public enum EndID
	{
		LivesLimit,
		ScoreLimit,
		TimeLimit,
		Last = TimeLimit,
	}
}