// Walt Destler

using System;
using System.Collections.Generic;
using System.Drawing;
using Halfling.Audio;
using Halfling.Geometry.Dim2;
using Halfling.Graphics;
using Halfling.Interpolation;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game.Doodads;
using TankyTank.Game.Tanks;

namespace TankyTank.Modes
{
	/// <summary>
	/// CTF client game mode.
	/// </summary>
	public class CTFClient : BaseClient
	{
		#region Strings & Sounds

		/// <summary>
		/// Stores txt-loaded strings for the CTF client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Strings
		{
			[Serial] public readonly string FriendlyFlagCaptured;
			[Serial] public readonly string FriendlyFlagDelivered;
			[Serial] public readonly string FriendlyFlagRescued;
			[Serial] public readonly string EnemyFlagCaptured;
			[Serial] public readonly string EnemyFlagDelivered;
			[Serial] public readonly string EnemyFlagRescued;
		}

		/// <summary>
		/// Stores txt-loaded graphics for the CTF client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Graphics
		{
			[Serial] public readonly Sprite Team1Flag;
			[Serial] public readonly Sprite Team2Flag;
			[Serial] public readonly Sprite Team1RP;
			[Serial] public readonly Sprite Team2RP;
		}

		/// <summary>
		/// Stores txt-loaded sound effects for the CTF client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Sounds
		{
			[Serial] public readonly Effect FriendlyFlagCaptured;
			[Serial] public readonly Effect FriendlyFlagDelivered;
			[Serial] public readonly Effect FriendlyFlagRescued;
			[Serial] public readonly Effect EnemyFlagCaptured;
			[Serial] public readonly Effect EnemyFlagDelivered;
			[Serial] public readonly Effect EnemyFlagRescued;
		}

		#endregion
		#region Private Fields

		private Strings strings;
		private Graphics graphics;
		private Sounds sounds;

		private Dictionary<ITank, FlaggedTankData> tanksWithFlags1 = new Dictionary<ITank, FlaggedTankData>();
		private Dictionary<ITank, FlaggedTankData> tanksWithFlags2 = new Dictionary<ITank, FlaggedTankData>();

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new CTF client game mode.
		/// </summary>
		public CTFClient(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			strings = ser.TreeDeserialize<Strings>(file, "Strings");
			graphics = ser.TreeDeserialize<Graphics>(file, "Graphics");
			sounds = ser.TreeDeserialize<Sounds>(file, "Sounds");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called when the game is started.
		/// </summary>
		public override void StartGame(IClientGame game, byte[] initData)
		{
			// Unpack initialization data.
			CTFInitData init = ser.BinaryDeserialize<CTFInitData>(initData);
			foreach(int id in init.TanksWithFlags1)
			{
				ITank tank = game.Tanks[id];
				FlaggedTankData ftd = new FlaggedTankData();
				ftd.InterpolatedLoc = new TerpVector2D(tank.Location);
				tanksWithFlags1.Add(tank, ftd);
			}
			foreach(int id in init.TanksWithFlags2)
			{
				ITank tank = game.Tanks[id];
				FlaggedTankData ftd = new FlaggedTankData();
				ftd.InterpolatedLoc = new TerpVector2D(tank.Location);
				tanksWithFlags2.Add(tank, ftd);
			}

			base.StartGame(game, init.BaseInitData);
		}

		/// <summary>
		/// Adds CTF specific sprites.
		/// </summary>
		public override void AddSprites()
		{
			base.AddSprites();

			// Draw flag on each tank.
			foreach(KeyValuePair<ITank, FlaggedTankData> kvp in tanksWithFlags1)
				addFlagSpriteOnTank(game.Rules.Doodads[CTFServer.CTFFlag1Key], kvp.Value.InterpolatedLoc.Interpolate(game.ClockTime));
			foreach(KeyValuePair<ITank, FlaggedTankData> kvp in tanksWithFlags2)
				addFlagSpriteOnTank(game.Rules.Doodads[CTFServer.CTFFlag2Key], kvp.Value.InterpolatedLoc.Interpolate(game.ClockTime));
		}

		/// <summary>
		/// Draws CTF specific minimap graphics.
		/// </summary>
		public override void DrawMinimap()
		{
			base.DrawMinimap();

			// Draw each flag and return point.
			foreach(IDoodad doodad in game.Doodads)
			{
				int key = game.Rules.Doodads.KeyOf(doodad.Rules);
				if(key == CTFServer.CTFFlag1Key)
					drawMinimapFlag(1, doodad.Location);
				else if(key == CTFServer.CTFFlag2Key)
					drawMinimapFlag(2, doodad.Location);
				else if(key == CTFServer.CTFReturn1Key)
					drawMinimapRP(1, doodad.Location);
				else if(key == CTFServer.CTFReturn2Key)
					drawMinimapRP(2, doodad.Location);
			}
			foreach(var kvp in tanksWithFlags1)
				drawMinimapFlag(1, kvp.Value.InterpolatedLoc.Interpolate(game.ClockTime));
			foreach(var kvp in tanksWithFlags2)
				drawMinimapFlag(2, kvp.Value.InterpolatedLoc.Interpolate(game.ClockTime));
		}

		/// <summary>
		/// Updates the game mode.
		/// </summary>
		public override void Update()
		{
			base.Update();
			
			// Remove any tanks with flags that no longer exist.
			List<ITank> toRemove = new List<ITank>();
			foreach(ITank tank in tanksWithFlags1.Keys)
			{
				if(!game.Tanks.Exists(tank.UniqueID))
					toRemove.Add(tank);
			}
			foreach(ITank tank in toRemove)
				tanksWithFlags1.Remove(tank);
			toRemove.Clear();
			foreach(ITank tank in tanksWithFlags2.Keys)
			{
				if(!game.Tanks.Exists(tank.UniqueID))
					toRemove.Add(tank);
			}
			foreach(ITank tank in toRemove)
				tanksWithFlags2.Remove(tank);

			// Update interpolated locations.
			foreach(var kvp in tanksWithFlags1)
				kvp.Value.InterpolatedLoc.SetValue(game.ClockTime, kvp.Key.Location, game.NextUpdateTime);
			foreach(var kvp in tanksWithFlags2)
				kvp.Value.InterpolatedLoc.SetValue(game.ClockTime, kvp.Key.Location, game.NextUpdateTime);
		}

		/// <summary>
		/// Called when a doodad is collected.
		/// </summary>
		public override void OnDoodadCollected(ITank tank, IDoodad doodad)
		{
			base.OnDoodadCollected(tank, doodad);

			// Did the tank collect a flag?
			if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFServer.CTFFlag1Key)
			{
				if(tank.Team == 2)
				{
					FlaggedTankData ftd = new FlaggedTankData();
					ftd.InterpolatedLoc = new TerpVector2D(tank.Location);
					tanksWithFlags1.Add(tank, ftd);

					if(game.Players.Exists(tank.Player))
					{
						if(game.UserPlayer.Team == tank.Team)
						{
							game.ShowMessage(
								string.Format(strings.EnemyFlagCaptured, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.EnemyFlagCaptured, true);
						}
						else
						{
							game.ShowMessage(
								string.Format(strings.FriendlyFlagCaptured, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.FriendlyFlagCaptured, true);
						}
					}
				}
				else
				{
					if(game.Players.Exists(tank.Player))
					{
						if(game.UserPlayer.Team == tank.Team)
						{
							game.ShowMessage(
								string.Format(strings.FriendlyFlagRescued, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.FriendlyFlagRescued, true);
						}
						else
						{
							game.ShowMessage(
								string.Format(strings.EnemyFlagRescued, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.EnemyFlagRescued, true);
						}
					}

					onFlagRescued(tank, doodad);
				}
			}
			else if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFServer.CTFFlag2Key)
			{
				if(tank.Team == 1)
				{
					FlaggedTankData ftd = new FlaggedTankData();
					ftd.InterpolatedLoc = new TerpVector2D(tank.Location);
					tanksWithFlags2.Add(tank, ftd);

					if(game.Players.Exists(tank.Player))
					{
						if(game.UserPlayer.Team == tank.Team)
						{
							game.ShowMessage(
								string.Format(strings.EnemyFlagCaptured, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.EnemyFlagCaptured, true);
						}
						else
						{
							game.ShowMessage(
								string.Format(strings.FriendlyFlagCaptured, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.FriendlyFlagCaptured, true);
						}
					}
				}
				else
				{
					if(game.Players.Exists(tank.Player))
					{
						if(game.UserPlayer.Team == tank.Team)
						{
							game.ShowMessage(
								string.Format(strings.FriendlyFlagRescued, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.FriendlyFlagRescued, true);
						}
						else
						{
							game.ShowMessage(
								string.Format(strings.EnemyFlagRescued, game.Players[tank.Player].PrintedName),
								Color.Magenta);
							game.PlaySound(sounds.EnemyFlagRescued, true);
						}
					}

					onFlagRescued(tank, doodad);
				}
			}
		}

		/// <summary>
		/// Called when a message from the server mode has been received.
		/// </summary>
		public override void ReceiveServerModeMessage(byte[] message)
		{
			base.ReceiveServerModeMessage(message);

			// This means that a flag has been returned...

			// Get tank that returned the flag.
			int tankID = ser.BinaryDeserialize<int>(message);
			ITank tank = game.Tanks[tankID];

			// Remove flag from tank.
			tanksWithFlags1.Remove(tank);
			tanksWithFlags2.Remove(tank);

			// Perform flag delivery actions.
			onFlagDelivered(tank);

			// Show message to player.
			if(tank.Team != game.UserPlayer.Team)
			{
				game.ShowMessage(
					string.Format(strings.FriendlyFlagDelivered, game.Players[tank.Player].PrintedName, 2),
					Color.Magenta);
				game.PlaySound(sounds.FriendlyFlagDelivered, true);
			}
			else
			{
				game.ShowMessage(
					string.Format(strings.EnemyFlagDelivered, game.Players[tank.Player].PrintedName, 2),
					Color.Magenta);
				game.PlaySound(sounds.EnemyFlagDelivered, true);
			}
		}

		/// <summary>
		/// Returns the text to be displayed on the left of the Infoboard.
		/// </summary>
		public override string GetInfoboardTextLeft()
		{
			string baseStr = base.GetInfoboardTextLeft();
			string teamsText = null;
			if(game.Teams.Count > 0)
			{
				string teamsCountText = "Teams: " + (game.Teams.Count - 1);
				string lockTeamsText = "Lock Teams: " + (modeSettings.LockTeams ? "Yes" : "No");
				string friendFireText = "Friendly Fire: " + (modeSettings.FriendlyFire ? "Yes" : "No");
				teamsText = "\n" + teamsCountText + "\n" +
							lockTeamsText + "\n" +
							friendFireText;
			}
			return baseStr + (teamsText ?? "");
		}

		/// <summary>
		/// Returns the text to be displayed on the right of the Infoboard.
		/// </summary>
		public override string GetInfoboardTextRight()
		{
			string baseStr = base.GetInfoboardTextRight();
			string flagDeliveredPoints = "Flag Points: " + modeSettings.FlagDeliveredPoints;
			string flagLostPoints = "Flag Lost Points: " + modeSettings.FlagLostPoints;
			string flagRescuedPoints = "Flag Rescued Points: " + modeSettings.FlagRescuedPoints;
			return baseStr + "\n" +
			       flagDeliveredPoints + "\n" +
			       flagLostPoints + "\n" +
			       flagRescuedPoints;
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Called when a flag is delivered.
		/// </summary>
		protected virtual void onFlagDelivered(ITank tank)
		{
			// Award points.
			game.AddPoints(tank.Player, tank.Team, modeSettings.FlagDeliveredPoints);
			if(tank.Team == 1)
				game.AddPoints(0, 2, modeSettings.FlagLostPoints);
			else
				game.AddPoints(0, 1, modeSettings.FlagLostPoints);

			if(game.UserPlayer.Team == tank.Team)
			{
				if(modeSettings.FlagDeliveredPoints != 0)
					game.CreateFloatingText("+" + modeSettings.FlagDeliveredPoints, tank.Location, new IntColor(0, 255, 0));
			}
			else
			{
				if(modeSettings.FlagLostPoints != 0)
					game.CreateFloatingText(modeSettings.FlagLostPoints.ToString(), tank.Location, new IntColor(255, 0, 0));
			}
		}

		/// <summary>
		/// Called when a flag is rescued.
		/// </summary>
		protected virtual void onFlagRescued(ITank tank, IDoodad doodad)
		{
			if(modeSettings.FlagRescuedPoints != 0)
			{
				game.AddPoints(tank.Player, tank.Team, modeSettings.FlagRescuedPoints);
				if(tank.Team == game.UserPlayer.Team)
					game.CreateFloatingText("+" + modeSettings.FlagRescuedPoints, doodad.Location, new IntColor(0, 255, 0));
			}
		}

		/// <summary>
		/// Draws a flag on the specified tank.
		/// </summary>
		private void addFlagSpriteOnTank(DoodadRules flagRules, Vector interpolatedLoc)
		{
			if(flagRules.DrawInGame)
			{
				foreach(DoodadSpriteRules sprite in flagRules.Graphics.DoodadSprites[0])
				{
					/*if(sprite.HighDetail && !game.App.Settings.HighDetail)
						continue;*/

					int frameNum = sprite.Animation.GetFrame(game.ClockTime);
					Sprite frame = sprite.Animation.Frames[frameNum];

					// Get rect.
					Rect r = new Rect();
					r.Size = (Vector)frame.Size;
					r.Center = interpolatedLoc +
						sprite.Offset.Rotate(flagRules.Orientation) +
						sprite.ShadowOffset;

					// Draw.
					Quad q = new Quad(
						r,
						sprite.DepthLevel.ToRenderable(),
						0,
						0,
						flagRules.Orientation,
						new Vector(.5f, .5f),
						frame.Color,
						frame.Source);
					game.AddSprite(q, frame, sprite.DepthLevel);
				}
			}
		}

		/// <summary>
		/// Draws the specified flag onto the minimap.
		/// </summary>
		private void drawMinimapFlag(int team, Vector gameLoc)
		{
			ISprite sprite;
			if(team == 1)
				sprite = graphics.Team1Flag;
			else if(team == 2)
				sprite = graphics.Team2Flag;
			else
				throw new ArgumentException("team");

			game.DrawOnMinimap(sprite, gameLoc, null);
		}

		/// <summary>
		/// Draws the specified return point on the minimap.
		/// </summary>
		private void drawMinimapRP(int team, Vector gameLoc)
		{
			ISprite sprite;
			if(team == 1)
				sprite = graphics.Team1RP;
			else if(team == 2)
				sprite = graphics.Team2RP;
			else
				throw new ArgumentException("team");

			game.DrawOnMinimap(sprite, gameLoc, null);
		}

		#endregion
		#region Private Types

		/// <summary>
		/// Stores data about a tank with a flag.
		/// </summary>
		private class FlaggedTankData
		{
			public TerpVector2D InterpolatedLoc;
		}

		#endregion
	}
}