// Walt Destler

using System;
using System.Collections.Generic;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game.Doodads;
using TankyTank.Game.Maps;
using TankyTank.Game.Tanks;

namespace TankyTank.Modes
{
	/// <summary>
	/// Stores CTF initialization data.
	/// </summary>
	[AutoSerializable] public struct CTFInitData
	{
		[Serial] public byte[] BaseInitData;
		[Serial] public List<int> TanksWithFlags1;
		[Serial] public List<int> TanksWithFlags2;
	}

	/// <summary>
	/// CTF server game mode.
	/// </summary>
	public class CTFServer : BaseServer
	{
		#region Strings

		/// <summary>
		/// Stores txt-loaded strings for the CTF client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class CheckMapStrings
		{
			[Serial] public readonly string MapErrorFlag1;
			[Serial] public readonly string MapErrorFlag2;
			[Serial] public readonly string MapErrorReturn1;
			[Serial] public readonly string MapErrorReturn2;
		}

		#endregion
		#region Doodad Constants

		public static readonly int CTFFlag1Key = StringHasher.Instance.GetPortableHash("CTFFlag1");
		public static readonly int CTFFlag2Key = StringHasher.Instance.GetPortableHash("CTFFlag2");
		public static readonly int CTFReturn1Key = StringHasher.Instance.GetPortableHash("CTFReturn1");
		public static readonly int CTFReturn2Key = StringHasher.Instance.GetPortableHash("CTFReturn2");

		#endregion
		#region Private Fields

		private List<Vector> originalFlags1 = new List<Vector>();
		private List<Vector> originalFlags2 = new List<Vector>();

		private List<ITank> tanksWithFlags1 = new List<ITank>();
		private List<ITank> tanksWithFlags2 = new List<ITank>();

		private readonly CTFInfo _info;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new CTF server game mode.
		/// </summary>
		public CTFServer(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			_info = new CTFInfo(this);
		}

		#endregion
		#region Map Validation

		/// <summary>
		/// Checks the specified map for compatibility with this game mode.
		/// </summary>
		public static new bool CheckMap(TreeFile file, Serializer ser, IMap map, IList<string> errors)
		{
			CheckMapStrings strings = ser.TreeDeserialize<CheckMapStrings>(file, "Strings");

			// Verify that map has at proper doodads.
			bool foundCTFFlag1 = false;
			bool foundCTFFlag2 = false;
			bool foundCTFReturn1 = false;
			bool foundCTFReturn2 = false;
			foreach(MapDoodad d in map.Doodads)
			{
				if(d.Key == CTFFlag1Key)
					foundCTFFlag1 = true;
				else if(d.Key == CTFFlag2Key)
					foundCTFFlag2 = true;
				else if(d.Key == CTFReturn1Key)
					foundCTFReturn1 = true;
				else if(d.Key == CTFReturn2Key)
					foundCTFReturn2 = true;
			}

			// Add any found errors.
			if(!foundCTFFlag1)
				errors.Add(strings.MapErrorFlag1);
			if(!foundCTFFlag2)
				errors.Add(strings.MapErrorFlag2);
			if(!foundCTFReturn1)
				errors.Add(strings.MapErrorReturn1);
			if(!foundCTFReturn2)
				errors.Add(strings.MapErrorReturn2);

			// Return.
			return errors.Count == 0;
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Starts the game.
		/// </summary>
		public override void StartGame(IServerGame game, INode modeSettingsNode)
		{
			base.StartGame(game, modeSettingsNode);

			// Record original flag locations.
			foreach(IDoodad doodad in game.Doodads)
			{
				if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFFlag1Key)
					originalFlags1.Add(doodad.Location);
				else if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFFlag2Key)
					originalFlags2.Add(doodad.Location);
			}
		}

		/// <summary>
		/// Returns the mode initialization data.
		/// </summary>
		public override byte[] GetInitData()
		{
			CTFInitData init = new CTFInitData();
			init.BaseInitData = base.GetInitData();
			init.TanksWithFlags1 = new List<int>(tanksWithFlags1.Count);
			init.TanksWithFlags2 = new List<int>(tanksWithFlags2.Count);
			foreach(ITank t in tanksWithFlags1)
				init.TanksWithFlags1.Add(t.UniqueID);
			foreach(ITank t in tanksWithFlags2)
				init.TanksWithFlags2.Add(t.UniqueID);

			return ser.BinarySerialize(init);
		}

		/// <summary>
		/// Called to update game mode logic.
		/// </summary>
		public override void Update()
		{
			base.Update();

			// Remove any tanks with flags that no longer exist.
			List<ITank> toRemove = new List<ITank>();
			foreach(ITank tank in tanksWithFlags1)
			{
				if(!game.Tanks.Exists(tank.UniqueID))
					toRemove.Add(tank);
			}
			foreach(ITank tank in toRemove)
				tanksWithFlags1.Remove(tank);
			toRemove.Clear();
			foreach(ITank tank in tanksWithFlags2)
			{
				if(!game.Tanks.Exists(tank.UniqueID))
					toRemove.Add(tank);
			}
			foreach(ITank tank in toRemove)
				tanksWithFlags2.Remove(tank);

			// Return flags.
			returnFlags(tanksWithFlags1, originalFlags1, CTFFlag1Key, CTFReturn2Key);
			returnFlags(tanksWithFlags2, originalFlags2, CTFFlag2Key, CTFReturn1Key);
		}

		/// <summary>
		/// Called when a tank is killed.
		/// </summary>
		public override void OnTankKilled(ITank tank, bool killedByZero)
		{
			base.OnTankKilled(tank, killedByZero);

			// Drop the flag if it was carrying one.
			if(tanksWithFlags1.Contains(tank))
				game.SpawnNewDoodad(CTFFlag1Key, tank.Location, false, 0, 1, game.Rules.Doodads[CTFFlag1Key].Orientation);
			else if(tanksWithFlags2.Contains(tank))
				game.SpawnNewDoodad(CTFFlag2Key, tank.Location, false, 0, 1, game.Rules.Doodads[CTFFlag2Key].Orientation);
		}

		/// <summary>
		/// Called when a doodad is collected.
		/// </summary>
		public override void OnDoodadCollected(ITank tank, IDoodad doodad)
		{
			base.OnDoodadCollected(tank, doodad);

			// Did the tank collect a flag?
			if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFFlag1Key)
			{
				if(tank.Team == 2)
					tanksWithFlags1.Add(tank);
				else
					onFlagRescued(tank, doodad);
			}
			else if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFFlag2Key)
			{
				if(tank.Team == 1)
					tanksWithFlags2.Add(tank);
				else
					onFlagRescued(tank, doodad);
			}
		}

		/// <summary>
		/// Called to verify that the specified doodad should be collected by the specified tank.
		/// </summary>
		public override bool VerifyDoodadCollect(ITank tank, IDoodad doodad)
		{
			int flagTeam;

			// Get flag team of doodad or return normally if doodad isn't a CTF flag.
			if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFFlag1Key)
				flagTeam = 1;
			else if(game.Rules.Doodads.KeyOf(doodad.Rules) == CTFFlag2Key)
				flagTeam = 2;
			else
				return base.VerifyDoodadCollect(tank, doodad);

			// Teamless tanks aren't allowed to collect flags.
			if(tank.Team == 0)
				return false;

			// Tanks can only collect flags of the opposite team.
			// (unless a flag is being liberated)
			if(tank.Team == flagTeam && (originalFlags1.Contains(doodad.Location) || originalFlags2.Contains(doodad.Location)))
				return false;

			// Only collect flag if this tank doesn't already have a flag. (and it's not being liberated)
			if(tank.Team != flagTeam && (tanksWithFlags1.Contains(tank) || tanksWithFlags2.Contains(tank)))
				return false;

			return true;
		}

		/// <summary>
		/// Returns an ICTFInfo object describing the current state of the capture-the-flag game.
		/// </summary>
		public override object ModeInfo
		{
			get { return _info; }
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Checks to see if any flags have been returned and returns them if so.
		/// </summary>
		private void returnFlags(List<ITank> tanksWithFlags, List<Vector> originalFlags, int flagKey, int returnKey)
		{
			List<ITank> remove = new List<ITank>();

			foreach(ITank tank in tanksWithFlags)
			{
				// Get any doodads this tank is over.
				var doodads = game.Doodads.FindIntersectsWith(tank.Location);
				foreach(IDoodad doodad in doodads)
				{
					// Is this doodad a team X return point?
					if(game.Rules.Doodads.KeyOf(doodad.Rules) == returnKey)
					{
						// Remove flag and add point.
						remove.Add(tank);

						// Send message to clients.
						game.MessageClientModes(ser.BinarySerialize(tank.UniqueID));

						// Perform flag delivery actions.
						onFlagDelivered(tank, originalFlags, flagKey, returnKey);
					}
				}
			}

			foreach(ITank tank in remove)
				tanksWithFlags.Remove(tank);
		}

		/// <summary>
		/// Called when a flag has been delivered.
		/// </summary>
		protected virtual void onFlagDelivered(ITank tank, List<Vector> originalFlags, int flagKey, int returnKey)
		{
			// Award points.
			game.AddPoints(tank.Player, tank.Team, modeSettings.FlagDeliveredPoints);
			if(tank.Team == 1)
				game.AddPoints(0, 2, modeSettings.FlagLostPoints);
			else
				game.AddPoints(0, 1, modeSettings.FlagLostPoints);

			// Respawn flag.
			respawnFlag(originalFlags, flagKey);
		}

		/// <summary>
		/// Called when a flag is rescued.
		/// </summary>
		protected virtual void onFlagRescued(ITank tank, IDoodad doodad)
		{
			// Award points.
			game.AddPoints(tank.Player, tank.Team, modeSettings.FlagRescuedPoints);

			// Respawn flag.
			if(tank.Team == 1)
				respawnFlag(originalFlags1, CTFFlag1Key);
			else if(tank.Team == 2)
				respawnFlag(originalFlags2, CTFFlag2Key);
		}

		/// <summary>
		/// Respawns a flag.
		/// </summary>
		protected void respawnFlag(List<Vector> locations, int flagKey)
		{
			List<Vector> found = new List<Vector>();

			// Find empty locations.
			foreach(Vector loc in locations)
			{
				var doodads = game.Doodads.FindIntersectsWith(loc);
				bool exists = false;
				foreach(IDoodad doodad in doodads)
				{
					if(game.Rules.Doodads.KeyOf(doodad.Rules) == flagKey && doodad.Location == loc)
					{
						exists = true;
						break;
					}
				}

				if(!exists)
					found.Add(loc);
			}

			// Choose a random location.
			if(found.Count == 0)
				throw new ApplicationException("Flag respawn location not found.");
			Vector newLoc = found[game.Rand.Int32(0, found.Count - 1)];

			// Spawn new flag.
			game.SpawnNewDoodad(flagKey, newLoc, false, 0, 1, game.Rules.Doodads[flagKey].Orientation);
		}

		#endregion
		#region Types

		/// <summary>
		/// Allows information about a capture-the-flag game to be retrieved.
		/// </summary>
		private class CTFInfo : ICTFInfo
		{
			#region Constructors

			/// <summary>
			/// Creates a new CTFInfo object.
			/// </summary>
			public CTFInfo(CTFServer ctf)
			{
				_ctf = ctf;
			}

			#endregion
			#region Fields

			private readonly CTFServer _ctf;

			#endregion
			#region Public Methods

			/// <summary>
			/// Returns the flag doodads.
			/// </summary>
			/// <param name="flagTeam">The team to which the returned flags belong.</param>
			public IEnumerable<IDoodad> GetFlags(int flagTeam)
			{
				if(flagTeam != 1 && flagTeam != 2)
					throw new ArgumentException("Flag team must be either 1 or 2.", "flagTeam");

				int flagKey = flagTeam == 1 ? CTFFlag1Key : CTFFlag2Key;
				foreach(IDoodad doodad in _ctf.game.Doodads)
				{
					if(_ctf.game.Rules.Doodads.KeyOf(doodad.Rules) == flagKey)
						yield return doodad;
				}
			}

			/// <summary>
			/// Returns the flag doodads that have been not been captured or moved by the enemy.
			/// </summary>
			/// <param name="flagTeam">The team to which the returned flags belong.</param>
			public IEnumerable<IDoodad> GetUncapturedFlags(int flagTeam)
			{
				if(flagTeam != 1 && flagTeam != 2)
					throw new ArgumentException("Flag team must be either 1 or 2.", "flagTeam");

				int flagKey = flagTeam == 1 ? CTFFlag1Key : CTFFlag2Key;
				var originalFlags = flagTeam == 1 ? _ctf.originalFlags1 : _ctf.originalFlags2;
				foreach(IDoodad doodad in _ctf.game.Doodads)
				{
					if(_ctf.game.Rules.Doodads.KeyOf(doodad.Rules) == flagKey && originalFlags.Contains(doodad.Location))
						yield return doodad;
				}
			}

			/// <summary>
			/// Returns the flag doodads that were captured by an enemy who was then destroyed
			/// and thus the flag needs to be rescued.
			/// </summary>
			/// <param name="flagTeam">The team to which the returned flags belong.</param>
			public IEnumerable<IDoodad> GetRescuableFlags(int flagTeam)
			{
				if(flagTeam != 1 && flagTeam != 2)
					throw new ArgumentException("Flag team must be either 1 or 2.", "flagTeam");

				int flagKey = flagTeam == 1 ? CTFFlag1Key : CTFFlag2Key;
				var originalFlags = flagTeam == 1 ? _ctf.originalFlags1 : _ctf.originalFlags2;
				foreach(IDoodad doodad in _ctf.game.Doodads)
				{
					if(_ctf.game.Rules.Doodads.KeyOf(doodad.Rules) == flagKey && !originalFlags.Contains(doodad.Location))
						yield return doodad;
				}
			}

			/// <summary>
			/// Returns the tanks currently carrying flags.
			/// </summary>
			/// <param name="flagTeam">The team to which the flags belong.</param>
			public IEnumerable<ITank> GetTanksWithFlags(int flagTeam)
			{
				if(flagTeam != 1 && flagTeam != 2)
					throw new ArgumentException("Flag team must be either 1 or 2.", "flagTeam");

				var tanksWithFlags = flagTeam == 1 ? _ctf.tanksWithFlags1 : _ctf.tanksWithFlags2;
				foreach(ITank tank in tanksWithFlags)
					yield return tank;
			}

			/// <summary>
			/// Returns whether the specified tank is currently carrying a flag.
			/// </summary>
			public bool IsTankCarryingFlag(ITank tank)
			{
				return _ctf.tanksWithFlags1.Contains(tank) || _ctf.tanksWithFlags2.Contains(tank);
			}

			/// <summary>
			/// Returns the flag return point doodads to which captured flags must be delivered.
			/// </summary>
			/// <param name="rpTeam">The team to which the returned flag return points belong.</param>
			public IEnumerable<IDoodad> GetReturnPoints(int rpTeam)
			{
				if(rpTeam != 1 && rpTeam != 2)
					throw new ArgumentException("Flag return point team must be either 1 or 2.", "rpTeam");

				int rpKey = rpTeam == 1 ? CTFReturn1Key : CTFReturn2Key;
				foreach(IDoodad doodad in _ctf.game.Doodads)
				{
					if(_ctf.game.Rules.Doodads.KeyOf(doodad.Rules) == rpKey)
						yield return doodad;
				}
			}

			#endregion
		}

		#endregion
	}
}