// Walt Destler

using System.Drawing;
using Halfling.Audio;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game.Players;
using TankyTank.Game.Tanks;

namespace TankyTank.Modes
{
	/// <summary>
	/// Juggernaut client game mode.
	/// </summary>
	public class JugClient : BaseClient
	{
		#region Strings & Sounds

		/// <summary>
		/// Stores txt-loaded strings for the CTF client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Strings
		{
			[Serial] public readonly string NewJuggernaut;
			[Serial] public readonly string JuggernautInfo;
			[Serial] public readonly string NoJuggernautInfo;
		}

		/// <summary>
		/// Stores txt-loaded sound effects for the CTF client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Sounds
		{
			[Serial] public readonly Effect NewJuggernaut;
		}

		#endregion
		#region Private Fields

		private Strings strings;
		private Sounds sounds;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new Juggernaut client game mode.
		/// </summary>
		public JugClient(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			strings = ser.TreeDeserialize<Strings>(file, "Strings");
			sounds = ser.TreeDeserialize<Sounds>(file, "Sounds");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called when a player is killed.
		/// </summary>
		public override void OnPlayerKilled(int player, int team, int sourcePlayer, int sourceTeam, Vector loc)
		{
			base.OnPlayerKilled(player, team, sourcePlayer, sourceTeam, loc);

			// Did someone kill the juggernaut?
			if(((team == 2 && sourceTeam == 1) || (team == 0 && sourceTeam == 0)) && sourcePlayer != 0)
			{
				game.ShowMessage(
					string.Format(strings.NewJuggernaut, game.Players[sourcePlayer].Name),
					Color.Magenta);
				game.PlaySound(sounds.NewJuggernaut, true);
			}
		}

		/// <summary>
		/// Returns whether the specified tank should be drawn on the minimap.
		/// Always true if the tank is the juggernaut.
		/// </summary>
		public override bool ShouldDrawMinimapTank(ITank tank)
		{
			if(tank.Team == 2)
				return true;
			else
				return base.ShouldDrawMinimapTank(tank);
		}

		/// <summary>
		/// Gets whether team scores should be shown.
		/// </summary>
		public override bool TeamScoresMatter
		{
			get { return false; }
		}

		/// <summary>
		/// Gets whether the player has no control over his initial team.
		/// </summary>
		public override bool DisallowInitTeamSelect
		{
			get { return true; }
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Gets the info text to show to the user.
		/// </summary>
		protected override string getInfoText()
		{
			string ret = base.getInfoText();
			if(ret != null)
				ret += "\n\n";
			else
				ret = "";

			IPlayer player = getJuggernautPlayer();
			if(player != null)
			{
				ITank tank = getJuggernautTank();
				int damage = tank != null ? (int)((double)tank.Health / tank.Rules.Combat.MaxHealth * 100) : 0;
				ret += string.Format(strings.JuggernautInfo, player.Name, damage);
			}
			else
			{
				ret += string.Format(strings.NoJuggernautInfo);
			}

			return ret;
		}

		/// <summary>
		/// Returns the current juggernaut player, or null if there is no juggernaut.
		/// </summary>
		private IPlayer getJuggernautPlayer()
		{
			foreach(IPlayer p in game.Players)
			{
				if(p.Team == 2)
					return p;
			}

			return null;
		}

		/// <summary>
		/// Returns the current juggernaut tank, or null if there is no juggernaut.
		/// </summary>
		private ITank getJuggernautTank()
		{
			foreach(ITank t in game.Tanks)
			{
				if(t.Team == 2)
					return t;
			}

			return null;
		}

		#endregion
	}
}