// Walt Destler

using System.Collections.Generic;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game.Players;
using TankyTank.Game.Tanks;

namespace TankyTank.Modes
{
	/// <summary>
	/// Juggernaut server game mode.
	/// </summary>
	public class JugServer : BaseServer
	{
		#region Settings

		/// <summary>
		/// Stores txt-loaded settings for the Juggernaut server mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Settings
		{
			[Serial] public readonly double PowerfulFactor;
			[Serial] public readonly double ArmoredFactor;
		}

		#endregion
		#region Fields

		private readonly Settings _settings;

		private bool _teamsActive = false;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new Juggernaut server game mode.
		/// </summary>
		public JugServer(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			_settings = ser.TreeDeserialize<Settings>(file, "Settings");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called to update mode-specific game logic.
		/// </summary>
		public override void Update()
		{
			base.Update();

			// If teams are active but there is no juggernaut, revert to FFA.
			if(_teamsActive && getJuggernautPlayer() == null)
			{
				_teamsActive = false;
				foreach(IPlayer p in game.Players)
					game.ChangeTeam(p.UniqueID, 0);
			}
		}

		/// <summary>
		/// Called to spawn a new player.
		/// </summary>
		public override void SpawnNewPlayer(int owner, string name, System.Drawing.Color color, int player, int team, int preferredTank)
		{
			team = _teamsActive ? 1 : 0;
			base.SpawnNewPlayer(owner, name, color, player, team, preferredTank);
		}

		/// <summary>
		/// Called to spawn a new AI player.
		/// </summary>
		public override void SpawnAIPlayer(TankyTank.AI.IBot bot, string name, System.Drawing.Color color, int team, int preferredTank)
		{
			team = _teamsActive ? 1 : 0;
			base.SpawnAIPlayer(bot, name, color, team, preferredTank);
		}

		/// <summary>
		/// Called when a player is killed.
		/// </summary>
		public override void OnPlayerKilled(int player, int team, int sourcePlayer, int sourceTeam, Vector loc)
		{
			base.OnPlayerKilled(player, team, sourcePlayer, sourceTeam, loc);

			if(_teamsActive)
			{
				// Did someone kill the juggernaut?
				if(team == 2 && (sourceTeam == 1 || sourceTeam == 0))
				{
					// Make new player the juggernaut.
					game.ChangeTeam(player, 1);
					game.ChangeTeam(sourcePlayer, 2);

					// Heal the new juggernaut and apply powerups.
					ITank tank = getJuggernautTank();
					if(tank != null)
					{
						game.DamageTank(tank.UniqueID, tank.Health - tank.Rules.Combat.MaxHealth, 0, 0, 0, Vector.Zero);
						game.ApplyPowerups(
							tank.UniqueID,
							0,
							0,
							1,
							-1,
							_settings.PowerfulFactor,
							-1,
							_settings.ArmoredFactor);
					}
				}
			}
			else if(player != sourcePlayer && sourcePlayer != 0)
			{
				// Turn on teams and make the source player the juggernaut.
				_teamsActive = true;
				game.ChangeTeam(sourcePlayer, 2);
				foreach(IPlayer p in game.Players)
				{
					if(p.Team == 0)
						game.ChangeTeam(p.UniqueID, 1);
				}

				// Heal the new juggernaut and apply powerups.
				ITank tank = getJuggernautTank();
				if(tank != null)
				{
					game.DamageTank(tank.UniqueID, tank.Health - tank.Rules.Combat.MaxHealth, 0, 0, 0, Vector.Zero);
					game.ApplyPowerups(
						tank.UniqueID,
						0,
						0,
						1,
						-1,
						_settings.PowerfulFactor,
						-1,
						_settings.ArmoredFactor);
				}
			}
		}

		/// <summary>
		/// Gets whether the player has no control over his initial team.
		/// </summary>
		public override bool DisallowInitTeamSelect
		{
			get { return true; }
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Returns whether the team score should be taken into consideration when checking for game end.
		/// </summary>
		protected override bool teamScoreMatters()
		{
			return false;
		}

		/// <summary>
		/// Returns whether tanks should spawn on their team spawn points.
		/// </summary>
		protected override bool teamSpawnsMatter()
		{
			return false;
		}

		/// <summary>
		/// Returns the current juggernaut player, or null if there is no juggernaut.
		/// </summary>
		private IPlayer getJuggernautPlayer()
		{
			foreach(IPlayer p in game.Players)
			{
				if(p.Team == 2)
					return p;
			}

			return null;
		}

		/// <summary>
		/// Returns the current juggernaut tank, or null if there is no juggernaut.
		/// </summary>
		private ITank getJuggernautTank()
		{
			foreach(ITank t in game.Tanks)
			{
				if(t.Team == 2)
					return t;
			}

			return null;
		}

		#endregion
	}
}