// Walt Destler

using System.Collections.Generic;
using Halfling.Audio;
using Halfling.Cursors;
using Halfling.Geometry.Dim2;
using Halfling.Graphics;
using Halfling.Gui;
using Halfling.Input;
using Halfling.Serialization;
using Halfling.Timing;
using Halfling.TreeData;
using TankyTank.Game.Bullets;
using TankyTank.Game.Doodads;
using TankyTank.Game.Maps;
using TankyTank.Game.Players;
using TankyTank.Game.Tanks;

namespace TankyTank.Modes
{
	/// <summary>
	/// Base game mode logic.
	/// </summary>
	public class ModeBase : IMode
	{
		#region Private Fields

		protected IGame game;
		protected Serializer ser;
		protected ModeBaseSettings modeSettings;
		protected string instructions;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new base game mode.
		/// </summary>
		public ModeBase(TreeFile file, Serializer ser)
		{
			this.ser = ser;
			instructions = ser.TreeDeserialize<string>(file, "Instructions");
		}

		#endregion
		#region Static Methods

		/// <summary>
		/// Checks the specified map for compatibility with this game mode.
		/// </summary>
		public static bool CheckMap(TreeFile file, Serializer ser, IMap map, IList<string> errors)
		{
			return true;
		}

		/// <summary>
		/// Creates the settings GUI for this mode.
		/// </summary>
		public static ISettingsGui CreateSettingsGui(TreeFile file, Serializer ser, Display dis, Mouse mouse,
			Keyboard kb, Clock clock, CursorManager cm, Squawker squawker, IVolumeController volController, IGuiRenderer guiRenderer,
			bool isMultiplayer)
		{
			SettingsGuiConfig config = ser.TreeDeserialize<SettingsGuiConfig>(file.FindAtPath("GuiSettingsConfig"));

			return new SettingsGui(
				file.FindAtPath("SettingsGui"),
				ser,
				dis,
				mouse,
				kb,
				clock,
				cm,
				squawker,
				volController,
				guiRenderer,
				config,
				isMultiplayer);
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called once when the game is started.
		/// </summary>
		public virtual void StartGame(IGame game, ModeBaseSettings modeSettings)
		{
			this.game = game;
			this.modeSettings = modeSettings;
		}

		/// <summary>
		/// Updates the game-mode specific logic.
		/// </summary>
		public virtual void Update()
		{
			
		}

		/// <summary>
		/// Called when a tank is killed.
		/// </summary>
		/// <param name="killedByZero">Whether the tank was killed by an inconsequential player.</param>
		public virtual void OnTankKilled(ITank tank, bool killedByZero)
		{
			
		}

		/// <summary>
		/// Called when a player is killed.
		/// </summary>
		public virtual void OnPlayerKilled(int player, int team, int sourcePlayer, int sourceTeam, Vector loc)
		{
			// Modify player statistics...

			// Suicide?
			if(player == sourcePlayer)
			{
				game.AddOops(player, team);
				game.AddDeath(player, team);
				if(modeSettings.OopsPoints != 0)
				{
					game.AddPoints(player, team, modeSettings.OopsPoints);
					if(game is IClientGame)
					{
						IPlayer userPlayer = ((IClientGame)game).UserPlayer;
						if(userPlayer.UniqueID == player || (userPlayer.Team != 0 && userPlayer.Team == team && TeamScoresMatter))
							((IClientGame)game).CreateFloatingText(modeSettings.OopsPoints.ToString(), loc, new IntColor(255, 0, 0));
					}
				}
			}

			// Friendly fire?
			else if(team == sourceTeam && team != 0)
			{
				game.AddDeath(player, team);
				game.AddOops(sourcePlayer, sourceTeam);
				if(modeSettings.OopsPoints != 0)
				{
					game.AddPoints(sourcePlayer, sourceTeam, modeSettings.OopsPoints);
					if(game is IClientGame)
					{
						IPlayer userPlayer = ((IClientGame)game).UserPlayer;
						if(userPlayer.UniqueID == sourcePlayer || (userPlayer.Team != 0 && userPlayer.Team == sourceTeam && TeamScoresMatter))
							((IClientGame)game).CreateFloatingText(modeSettings.OopsPoints.ToString(), loc, new IntColor(255, 0, 0));
					}
				}
			}

			else
			{
				game.AddDeath(player, team);
				if(modeSettings.DeathPoints != 0)
				{
					game.AddPoints(player, team, modeSettings.DeathPoints);
					if(game is IClientGame)
					{
						IPlayer userPlayer = ((IClientGame)game).UserPlayer;
						if(userPlayer.UniqueID == player || (userPlayer.Team != 0 && userPlayer.Team == team && TeamScoresMatter))
							((IClientGame)game).CreateFloatingText(modeSettings.DeathPoints.ToString(), loc, new IntColor(255, 0, 0));
					}
				}
				game.AddKill(sourcePlayer, sourceTeam);
				if(modeSettings.KillPoints != 0)
				{
					game.AddPoints(sourcePlayer, sourceTeam, modeSettings.KillPoints);
					if(game is IClientGame)
					{
						IPlayer userPlayer = ((IClientGame)game).UserPlayer;
						if(userPlayer.UniqueID == sourcePlayer || (userPlayer.Team != 0 && userPlayer.Team == sourceTeam && TeamScoresMatter))
							((IClientGame)game).CreateFloatingText("+" + modeSettings.KillPoints, loc, new IntColor(0, 255, 0));
					}
				}
			}
		}

		/// <summary>
		/// Called when a doodad is collected.
		/// </summary>
		public virtual void OnDoodadCollected(ITank tank, IDoodad doodad)
		{

		}

		/// <summary>
		/// Gets whether the user is not allowed to select his tank.
		/// </summary>
		public bool ForceRandomTanks
		{
			get { return modeSettings.ForceRandomTanks; }
		}

		/// <summary>
		/// Gets whether teams are locked once chosen.
		/// </summary>
		public bool LockTeams
		{
			get { return modeSettings.LockTeams; }
		}

		/// <summary>
		/// Gets whether the player has no control over his initial team.
		/// </summary>
		public virtual bool DisallowInitTeamSelect
		{
			get { return false; }
		}

		/// <summary>
		/// Gets the limit on the number of human players in the game, or 0 for unlimited.
		/// </summary>
		public int PlayerLimit
		{
			get { return modeSettings.PlayerLimitEnable ? modeSettings.PlayerLimit : 0; }
		}

		/// <summary>
		/// Gets an object possibly containing the current mode information.
		/// </summary>
		public virtual object ModeInfo
		{
			get { return null; }
		}

		/// <summary>
		/// Gets the number of frames before a tank can respawn.
		/// </summary>
		public virtual int RespawnDelay
		{
			get { return game.Rules.General.UpdatesPerSecond * modeSettings.RespawnDelay; }
		}

		/// <summary>
		/// Gets whether team scores should be drawn.
		/// </summary>
		public virtual bool TeamScoresMatter
		{
			get { return game.Teams.Count > 0; }
		}

		#endregion
	}

	/// <summary>
	/// Stores the settings for a base game mode.
	/// </summary>
	[AutoSerializable] public class ModeBaseSettings
	{
		[Serial, Optional(false)] public bool LivesLimitEnable;
		[Serial, Optional(5)] public int LivesLimit;

		[Serial, Optional(true)] public bool ScoreLimitEnable;
		[Serial, Optional(25)] public int ScoreLimit;

		[Serial, Optional(false)] public bool TimeLimitEnable;
		[Serial, Optional(10)] public int TimeLimit;

		[Serial, Optional(false)] public bool LockTeams;
		[Serial, Optional(false)] public bool FriendlyFire;

		[Serial, Optional(false)] public bool PlayerLimitEnable;
		[Serial, Optional(16)] public int PlayerLimit;

		[Serial, Optional(false)] public bool ForceRandomTanks;

		[Serial, Optional(3)] public int RespawnDelay;

		[Serial, Optional(1)] public int KillPoints;
		[Serial, Optional(0)] public int DeathPoints;
		[Serial, Optional(-1)] public int OopsPoints;
		[Serial, Optional(5)] public int FlagDeliveredPoints;
		[Serial, Optional(0)] public int FlagLostPoints;
		[Serial, Optional(0)] public int FlagRescuedPoints;
	}
}