// Walt Destler

using System;
using Halfling.Audio;
using Halfling.Cursors;
using Halfling.Graphics;
using Halfling.Gui;
using Halfling.Input;
using Halfling.Serialization;
using Halfling.Timing;
using Halfling.TreeData;

namespace TankyTank.Modes
{
	/// <summary>
	/// The base GUI for game mode settings.
	/// </summary>
	public class SettingsGui : ISettingsGui
	{
		#region Events

		/// <summary>
		/// Invoked when the number of teams is changed.
		/// </summary>
		public event EventHandler<EventArgs> TeamsChanged;

		/// <summary>
		/// Invoked when allowing tank selection is toggled.
		/// </summary>
		public event EventHandler<EventArgs> AllowTankSelectionChanged;

		#endregion
		#region Non-Public Fields

		protected readonly bool isMultiplayer;
		protected readonly SettingsGuiConfig config;

		protected IGui root;
		protected CheckButton livesLimitEnable;
		protected Scrollbar livesLimit;
		protected CheckButton scoreLimitEnable;
		protected Scrollbar scoreLimit;
		protected CheckButton timeLimitEnable;
		protected Scrollbar timeLimit;
		protected CheckButton teamsEnable;
		protected Scrollbar teams;
		protected CheckButton teamsLock;
		protected CheckButton teamsFriendlyFire;
		protected CheckButton playerLimitEnable;
		protected Scrollbar playerLimit;
		protected CheckButton forceRandomTanks;
		protected Scrollbar respawnDelay;
		protected Label killPointsLabel;
		protected Scrollbar killPoints;
		protected Label deathPointsLabel;
		protected Scrollbar deathPoints;
		protected Label oopsPointsLabel;
		protected Scrollbar oopsPoints;
		protected Label flagDeliveredPointsLabel;
		protected Scrollbar flagDeliveredPoints;
		protected Label flagLostPointsLabel;
		protected Scrollbar flagLostPoints;
		protected Label flagRescuedPointsLabel;
		protected Scrollbar flagRescuedPoints;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new game mode settings GUI.
		/// </summary>
		public SettingsGui(INode node, Serializer ser, Display dis, Mouse mouse, Keyboard kb,
			Clock clock, CursorManager cm, Squawker squawker, IVolumeController volController, IGuiRenderer guiRenderer,
			SettingsGuiConfig config, bool isMultiplayer)
		{
			AutoMenu menu = new AutoMenu(ser, node, dis, guiRenderer);
			
			// Get tagged controls.
			livesLimitEnable = (CheckButton)menu.GetTaggedGui("liveslimitenable");
			livesLimit = (Scrollbar)menu.GetTaggedGui("liveslimit");
			scoreLimitEnable = (CheckButton)menu.GetTaggedGui("scorelimitenable");
			scoreLimit = (Scrollbar)menu.GetTaggedGui("scorelimit");
			timeLimitEnable = (CheckButton)menu.GetTaggedGui("timelimitenable");
			timeLimit = (Scrollbar)menu.GetTaggedGui("timelimit");
			teamsEnable = (CheckButton)menu.GetTaggedGui("teamsenable");
			teams = (Scrollbar)menu.GetTaggedGui("teams");
			teamsLock = (CheckButton)menu.GetTaggedGui("teamslock");
			teamsFriendlyFire = (CheckButton)menu.GetTaggedGui("teamsfriendlyfire");
			playerLimitEnable = (CheckButton)menu.GetTaggedGui("playerlimitenable");
			playerLimit = (Scrollbar)menu.GetTaggedGui("playerlimit");
			forceRandomTanks = (CheckButton)menu.GetTaggedGui("forcerandomtanks");
			respawnDelay = (Scrollbar)menu.GetTaggedGui("respawndelay");
			killPointsLabel = (Label)menu.GetTaggedGui("killpointslabel");
			killPoints = (Scrollbar)menu.GetTaggedGui("killpoints");
			deathPointsLabel = (Label)menu.GetTaggedGui("deathpointslabel");
			deathPoints = (Scrollbar)menu.GetTaggedGui("deathpoints");
			oopsPointsLabel = (Label)menu.GetTaggedGui("oopspointslabel");
			oopsPoints = (Scrollbar)menu.GetTaggedGui("oopspoints");
			flagDeliveredPointsLabel = (Label)menu.GetTaggedGui("flagdeliveredpointslabel");
			flagDeliveredPoints = (Scrollbar)menu.GetTaggedGui("flagdeliveredpoints");
			flagLostPointsLabel = (Label)menu.GetTaggedGui("flaglostpointslabel");
			flagLostPoints = (Scrollbar)menu.GetTaggedGui("flaglostpoints");
			flagRescuedPointsLabel = (Label)menu.GetTaggedGui("flagrescuedpointslabel");
			flagRescuedPoints = (Scrollbar)menu.GetTaggedGui("flagrescuedpoints");

			// Handle events.
			livesLimitEnable.SelectedChanged += onLivesLimitEnableChanged;
			scoreLimitEnable.SelectedChanged += onScoreLimitEnableChanged;
			timeLimitEnable.SelectedChanged += onTimeLimitEnableChanged;
			teamsEnable.SelectedChanged += onTeamsEnableChanged;
			teams.ValueChanged += onTeamsChanged;
			playerLimitEnable.SelectedChanged += onPlayerLimitEnableChanged;
			forceRandomTanks.SelectedChanged += onForceRandomTanksChanged;

			// Get root and detach it from the menu.
			root = menu.Root;
			root.SettableParent = null;
			
			// Configure gui.
			this.config = config;
			this.isMultiplayer = isMultiplayer;
			configureGui();
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Loads the default settings from the specified node.
		/// </summary>
		public virtual void LoadSettings(INode node, Serializer ser)
		{
			SettingsGuiSettings settings = new SettingsGuiSettings();
			settings.LivesLimitEnable = config.LivesLimitOnValue;
			settings.LivesLimit = config.LivesLimitCountValue;
			settings.ScoreLimitEnable = config.ScoreLimitOnValue;
			settings.ScoreLimit = config.ScoreLimitCountValue;
			settings.TimeLimitEnable = config.TimeLimitOnValue;
			settings.TimeLimit = config.TimeLimitCountValue;
			settings.TeamsEnable = config.TeamsOnValue;
			settings.Teams = config.TeamsCountValue;
			settings.LockTeams = config.TeamsLockValue;
			settings.FriendlyFire = config.TeamsFriendlyFireValue;
			settings.PlayerLimitEnable = config.PlayerLimitOnValue;
			settings.PlayerLimit = config.PlayerLimitCountValue;
			settings.ForceRandomTanks = config.ForceRandomTanksValue;
			settings.RespawnDelay = config.RespawnDelayValue;
			settings.KillPoints = config.KillPointsValue;
			settings.DeathPoints = config.DeathPointsValue;
			settings.OopsPoints = config.OopsPointsValue;
			settings.FlagDeliveredPoints = config.FlagDeliveredPointsValue;
			settings.FlagLostPoints = config.FlagLostPointsValue;
			settings.FlagRescuedPoints = config.FlagRescuedPointsValue;
			ser.TreeFill(settings, node, false);

			if(config.LivesLimitOnModifiable)
				livesLimitEnable.Selected = settings.LivesLimitEnable;
			if(config.LivesLimitCountModifiable)
				livesLimit.Value = settings.LivesLimit;
			if(config.ScoreLimitOnModifiable)
				scoreLimitEnable.Selected = settings.ScoreLimitEnable;
			if(config.ScoreLimitCountModifiable)
				scoreLimit.Value = settings.ScoreLimit;
			if(config.TimeLimitOnModifiable)
				timeLimitEnable.Selected = settings.TimeLimitEnable;
			if(config.TimeLimitCountModifiable)
				timeLimit.Value = settings.TimeLimit;
			if(config.TeamsOnModifiable)
				teamsEnable.Selected = settings.TeamsEnable;
			if(config.TeamsCountModifiable)
				teams.Value = settings.Teams;
			if(config.TeamsLockModifiable)
				teamsLock.Selected = settings.LockTeams;
			if(config.TeamsFriendlyFireModifiable)
				teamsFriendlyFire.Selected = settings.FriendlyFire;
			if(config.PlayerLimitOnModifiable)
				playerLimitEnable.Selected = settings.PlayerLimitEnable;
			if(config.PlayerLimitCountModifiable)
				playerLimit.Value = settings.PlayerLimit;
			if(config.ForceRandomTanksModifiable)
				forceRandomTanks.Selected = settings.ForceRandomTanks;
			if(config.RespawnDelayModifiable)
				respawnDelay.Value = settings.RespawnDelay;
			if(config.KillPointsModifiable)
				killPoints.Value = settings.KillPoints;
			if(config.DeathPointsModifiable)
				deathPoints.Value = settings.DeathPoints;
			if(config.OopsPointsModifiable)
				oopsPoints.Value = settings.OopsPoints;
			if(config.FlagDeliveredPointsModifiable)
				flagDeliveredPoints.Value = settings.FlagDeliveredPoints;
			if(config.FlagLostPointsModifiable)
				flagLostPoints.Value = settings.FlagLostPoints;
			if(config.FlagRescuedPointsModifiable)
				flagRescuedPoints.Value = settings.FlagRescuedPoints;
		}

		/// <summary>
		/// Saves the settings to the specified node.
		/// </summary>
		public virtual void SaveSettings(INode node, Serializer ser)
		{
			SettingsGuiSettings settings = new SettingsGuiSettings();

			settings.LivesLimitEnable = livesLimitEnable.Selected;
			settings.LivesLimit = livesLimit.Value;
			settings.ScoreLimitEnable = scoreLimitEnable.Selected;
			settings.ScoreLimit = scoreLimit.Value;
			settings.TimeLimitEnable = timeLimitEnable.Selected;
			settings.TimeLimit = timeLimit.Value;
			settings.TeamsEnable = teamsEnable.Selected;
			settings.Teams = teams.Value;
			settings.PlayerLimitEnable = playerLimitEnable.Selected;
			settings.PlayerLimit = playerLimit.Value;
			settings.LockTeams = teamsLock.Selected;
			settings.FriendlyFire = teamsFriendlyFire.Selected;
			settings.ForceRandomTanks = forceRandomTanks.Selected;
			settings.RespawnDelay = respawnDelay.Value;
			settings.KillPoints = killPoints.Value;
			settings.DeathPoints = deathPoints.Value;
			settings.OopsPoints = oopsPoints.Value;
			settings.FlagDeliveredPoints = flagDeliveredPoints.Value;
			settings.FlagLostPoints = flagLostPoints.Value;
			settings.FlagRescuedPoints = flagRescuedPoints.Value;

			ser.TreeSerialize(node, settings);
		}

		/// <summary>
		/// Gets the root of this settings gui.
		/// </summary>
		public virtual IGui Root
		{
			get { return root; }
		}

		/// <summary>
		/// Gets the number of teams specified by this gui.
		/// </summary>
		public virtual int Teams
		{
			get
			{
				if(teamsEnable.Selected)
					return teams.Value;
				else
					return 0;
			}
		}

		/// <summary>
		/// Gets whether the user is allowed to select tanks for AI players.
		/// </summary>
		public virtual bool AllowTankSelection
		{
			get { return !forceRandomTanks.Selected; }
		}

		/// <summary>
		/// Gets whether the user is allowed to select teams for AI players.
		/// </summary>
		public virtual bool AllowTeamSelection
		{
			get { return Teams > 0 && config.AllowTeamSelection; }
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Sets the user interface according to the specified host options object.
		/// </summary>
		private void configureGui()
		{
			// Lives limit...

			livesLimitEnable.Visible = config.LivesLimitVisible;
			livesLimit.Visible = config.LivesLimitVisible;

			livesLimitEnable.Enabled = config.LivesLimitOnModifiable;
			if(!config.LivesLimitOnModifiable)
				livesLimitEnable.Selected = config.LivesLimitOnValue;

			livesLimit.Enabled = config.LivesLimitCountModifiable && livesLimitEnable.Selected;
			livesLimit.MinValue = config.LivesLimitCountMinValue;
			livesLimit.MaxValue = config.LivesLimitCountMaxValue;
			if(!config.LivesLimitCountModifiable)
				livesLimit.Value = config.LivesLimitCountValue;

			// Score limit...

			scoreLimitEnable.Visible = config.ScoreLimitVisible;
			scoreLimit.Visible = config.ScoreLimitVisible;

			scoreLimitEnable.Enabled = config.ScoreLimitOnModifiable;
			if(!config.ScoreLimitOnModifiable)
				scoreLimitEnable.Selected = config.ScoreLimitOnValue;

			scoreLimit.Enabled = config.ScoreLimitCountModifiable && scoreLimitEnable.Selected;
			scoreLimit.MinValue = config.ScoreLimitCountMinValue;
			scoreLimit.MaxValue = config.ScoreLimitCountMaxValue;
			if(!config.ScoreLimitCountModifiable)
				scoreLimit.Value = config.ScoreLimitCountValue;

			// Time limit...

			timeLimitEnable.Visible = config.TimeLimitVisible;
			timeLimit.Visible = config.TimeLimitVisible;

			timeLimitEnable.Enabled = config.TimeLimitOnModifiable;
			if(!config.TimeLimitOnModifiable)
				timeLimitEnable.Selected = config.TimeLimitOnValue;

			timeLimit.Enabled = config.TimeLimitCountModifiable && timeLimitEnable.Selected;
			timeLimit.MinValue = config.TimeLimitCountMinValue;
			timeLimit.MaxValue = config.TimeLimitCountMaxValue;
			if(!config.TimeLimitCountModifiable)
				timeLimit.Value = config.TimeLimitCountValue;

			// Player limit...

			playerLimitEnable.Visible = config.PlayerLimitVisible && isMultiplayer;
			playerLimit.Visible = config.PlayerLimitVisible && isMultiplayer;

			playerLimitEnable.Enabled = config.PlayerLimitOnModifiable;
			if(!config.PlayerLimitOnModifiable)
				playerLimitEnable.Selected = config.PlayerLimitOnValue;

			playerLimit.Enabled = config.PlayerLimitCountModifiable && playerLimitEnable.Selected;
			playerLimit.MinValue = config.PlayerLimitCountMinValue;
			playerLimit.MaxValue = config.PlayerLimitCountMaxValue;
			if(!config.PlayerLimitCountModifiable)
				playerLimit.Value = config.PlayerLimitCountValue;

			// Teams...

			teamsEnable.Visible = config.TeamsVisible;
			teams.Visible = config.TeamsVisible;
			teamsLock.Visible = config.TeamsVisible;
			teamsFriendlyFire.Visible = config.TeamsVisible;

			teamsEnable.Enabled = config.TeamsOnModifiable;
			if(!config.TeamsOnModifiable)
				teamsEnable.Selected = config.TeamsOnValue;

			teams.Enabled = config.TeamsCountModifiable && teamsEnable.Selected;
			teams.MinValue = config.TeamsCountMinValue;
			teams.MaxValue = Math.Min(config.TeamsCountMaxValue, Constants.MaxTeams);
			if(!config.TeamsCountModifiable)
				teams.Value = config.TeamsCountValue;

			teamsLock.Enabled = config.TeamsLockModifiable && teamsEnable.Selected;
			if(!config.TeamsLockModifiable)
				teamsLock.Selected = config.TeamsLockValue;

			teamsFriendlyFire.Enabled = config.TeamsFriendlyFireModifiable && teamsEnable.Selected;
			if(!config.TeamsFriendlyFireModifiable)
				teamsFriendlyFire.Selected = config.TeamsFriendlyFireValue;

			// Force random tanks...

			forceRandomTanks.Visible = config.ForceRandomTanksVisible;
			forceRandomTanks.Enabled = config.ForceRandomTanksModifiable;
			if(!config.ForceRandomTanksModifiable)
				forceRandomTanks.Selected = config.ForceRandomTanksValue;

			// Respawn delay...

			respawnDelay.Visible = config.RespawnDelayVisible;
			respawnDelay.Enabled = config.RespawnDelayModifiable;
			respawnDelay.MinValue = config.RespawnDelayMinValue;
			respawnDelay.MaxValue = config.RespawnDelayMaxValue;
			if(!config.RespawnDelayModifiable)
				respawnDelay.Value = config.RespawnDelayValue;

			// Kill points...

			killPointsLabel.Visible = config.KillPointsVisible;
			killPoints.Visible = config.KillPointsVisible;
			killPoints.Enabled = config.KillPointsModifiable;
			killPoints.MinValue = config.KillPointsMinValue;
			killPoints.MaxValue = config.KillPointsMaxValue;
			if(!config.KillPointsModifiable)
				killPoints.Value = config.KillPointsValue;

			// Death points...

			deathPointsLabel.Visible = config.DeathPointsVisible;
			deathPoints.Visible = config.DeathPointsVisible;
			deathPoints.Enabled = config.DeathPointsModifiable;
			deathPoints.MinValue = config.DeathPointsMinValue;
			deathPoints.MaxValue = config.DeathPointsMaxValue;
			if(!config.DeathPointsModifiable)
				deathPoints.Value = config.DeathPointsValue;

			// Oops points...

			oopsPointsLabel.Visible = config.OopsPointsVisible;
			oopsPoints.Visible = config.OopsPointsVisible;
			oopsPoints.Enabled = config.OopsPointsModifiable;
			oopsPoints.MinValue = config.OopsPointsMinValue;
			oopsPoints.MaxValue = config.OopsPointsMaxValue;
			if(!config.OopsPointsModifiable)
				oopsPoints.Value = config.OopsPointsValue;

			// Flag delivered points...

			flagDeliveredPointsLabel.Visible = config.FlagDeliveredPointsVisible;
			flagDeliveredPoints.Visible = config.FlagDeliveredPointsVisible;
			flagDeliveredPoints.Enabled = config.FlagDeliveredPointsModifiable;
			flagDeliveredPoints.MinValue = config.FlagDeliveredPointsMinValue;
			flagDeliveredPoints.MaxValue = config.FlagDeliveredPointsMaxValue;
			if(!config.FlagDeliveredPointsModifiable)
				flagDeliveredPoints.Value = config.FlagDeliveredPointsValue;

			// Flag lost points...

			flagLostPointsLabel.Visible = config.FlagLostPointsVisible;
			flagLostPoints.Visible = config.FlagLostPointsVisible;
			flagLostPoints.Enabled = config.FlagLostPointsModifiable;
			flagLostPoints.MinValue = config.FlagLostPointsMinValue;
			flagLostPoints.MaxValue = config.FlagLostPointsMaxValue;
			if(!config.FlagLostPointsModifiable)
				flagLostPoints.Value = config.FlagLostPointsValue;

			// Flag rescued points...

			flagRescuedPointsLabel.Visible = config.FlagRescuedPointsVisible;
			flagRescuedPoints.Visible = config.FlagRescuedPointsVisible;
			flagRescuedPoints.Enabled = config.FlagRescuedPointsModifiable;
			flagRescuedPoints.MinValue = config.FlagRescuedPointsMinValue;
			flagRescuedPoints.MaxValue = config.FlagRescuedPointsMaxValue;
			if(!config.FlagRescuedPointsModifiable)
				flagRescuedPoints.Value = config.FlagRescuedPointsValue;
		}

		/// <summary>
		/// Called when the user enables or disables the lives limit.
		/// </summary>
		private void onLivesLimitEnableChanged(object sender, EventArgs e)
		{
			livesLimit.Enabled = livesLimitEnable.Selected && config.LivesLimitCountModifiable;
		}

		/// <summary>
		/// Called when the user enables or disables the score limit.
		/// </summary>
		private void onScoreLimitEnableChanged(object sender, EventArgs e)
		{
			scoreLimit.Enabled = scoreLimitEnable.Selected && config.ScoreLimitCountModifiable;
		}

		/// <summary>
		/// Called when the user enables or disables the time limit.
		/// </summary>
		private void onTimeLimitEnableChanged(object sender, EventArgs e)
		{
			timeLimit.Enabled = timeLimitEnable.Selected && config.TimeLimitCountModifiable;
		}

		/// <summary>
		/// Called when the user enables or disables teams.
		/// </summary>
		private void onTeamsEnableChanged(object sender, EventArgs e)
		{
			teams.Enabled = teamsEnable.Selected && config.TeamsCountModifiable;
			teamsLock.Enabled = teamsEnable.Selected && config.TeamsLockModifiable;
			teamsFriendlyFire.Enabled = teamsEnable.Selected && config.TeamsFriendlyFireModifiable;

			invokeTeamsChanged();
		}

		/// <summary>
		/// Called when the user changes the number of teams.
		/// </summary>
		private void onTeamsChanged(object sender, EventArgs e)
		{
			invokeTeamsChanged();
		}

		/// <summary>
		/// Called when the user enables or disables player limits.
		/// </summary>
		private void onPlayerLimitEnableChanged(object sender, EventArgs e)
		{
			playerLimit.Enabled = playerLimitEnable.Selected && config.ScoreLimitCountModifiable;
		}

		/// <summary>
		/// Called when the user changes the force random tanks option.
		/// </summary>
		private void onForceRandomTanksChanged(object sender, EventArgs e)
		{
			invokeAllowTankSelectionChanged();
		}

		/// <summary>
		/// Invokes the TeamsChanged event.
		/// </summary>
		private void invokeTeamsChanged()
		{
			if(TeamsChanged != null)
				TeamsChanged(this, EventArgs.Empty);
		}

		/// <summary>
		/// Invokes the AllowTankSelectionChanged event.
		/// </summary>
		private void invokeAllowTankSelectionChanged()
		{
			if(AllowTankSelectionChanged != null)
				AllowTankSelectionChanged(this, EventArgs.Empty);
		}

		#endregion
	}

	/// <summary>
	/// Defines what options for a game mode the host may change and the possible ranges of values.
	/// </summary>
	[AutoSerializable] public class SettingsGuiConfig
	{
		[Serial] public readonly bool LivesLimitVisible;
		[Serial] public readonly bool LivesLimitOnModifiable;
		[Serial] public readonly bool LivesLimitOnValue;
		[Serial] public readonly bool LivesLimitCountModifiable;
		[Serial] public readonly int LivesLimitCountValue;
		[Serial] public readonly int LivesLimitCountMinValue;
		[Serial] public readonly int LivesLimitCountMaxValue;

		[Serial] public readonly bool ScoreLimitVisible;
		[Serial] public readonly bool ScoreLimitOnModifiable;
		[Serial] public readonly bool ScoreLimitOnValue;
		[Serial] public readonly bool ScoreLimitCountModifiable;
		[Serial] public readonly int ScoreLimitCountValue;
		[Serial] public readonly int ScoreLimitCountMinValue;
		[Serial] public readonly int ScoreLimitCountMaxValue;

		[Serial] public readonly bool TimeLimitVisible;
		[Serial] public readonly bool TimeLimitOnModifiable;
		[Serial] public readonly bool TimeLimitOnValue;
		[Serial] public readonly bool TimeLimitCountModifiable;
		[Serial] public readonly int TimeLimitCountValue;
		[Serial] public readonly int TimeLimitCountMinValue;
		[Serial] public readonly int TimeLimitCountMaxValue;

		[Serial] public readonly bool PlayerLimitVisible;
		[Serial] public readonly bool PlayerLimitOnModifiable;
		[Serial] public readonly bool PlayerLimitOnValue;
		[Serial] public readonly bool PlayerLimitCountModifiable;
		[Serial] public readonly int PlayerLimitCountValue;
		[Serial] public readonly int PlayerLimitCountMinValue;
		[Serial] public readonly int PlayerLimitCountMaxValue;

		[Serial] public readonly bool TeamsVisible;
		[Serial] public readonly bool TeamsOnModifiable;
		[Serial] public readonly bool TeamsOnValue;
		[Serial] public readonly bool TeamsCountModifiable;
		[Serial] public readonly int TeamsCountValue;
		[Serial] public readonly int TeamsCountMinValue;
		[Serial] public readonly int TeamsCountMaxValue;
		[Serial] public readonly bool TeamsLockModifiable;
		[Serial] public readonly bool TeamsLockValue;
		[Serial] public readonly bool TeamsFriendlyFireModifiable;
		[Serial] public readonly bool TeamsFriendlyFireValue;

		[Serial] public readonly bool ForceRandomTanksVisible;
		[Serial] public readonly bool ForceRandomTanksModifiable;
		[Serial] public readonly bool ForceRandomTanksValue;

		[Serial] public readonly bool RespawnDelayVisible;
		[Serial] public readonly bool RespawnDelayModifiable;
		[Serial] public readonly int RespawnDelayValue;
		[Serial] public readonly int RespawnDelayMinValue;
		[Serial] public readonly int RespawnDelayMaxValue;

		[Serial] public readonly bool KillPointsVisible;
		[Serial] public readonly bool KillPointsModifiable;
		[Serial] public readonly int KillPointsValue;
		[Serial] public readonly int KillPointsMinValue;
		[Serial] public readonly int KillPointsMaxValue;
		
		[Serial] public readonly bool DeathPointsVisible;
		[Serial] public readonly bool DeathPointsModifiable;
		[Serial] public readonly int DeathPointsValue;
		[Serial] public readonly int DeathPointsMinValue;
		[Serial] public readonly int DeathPointsMaxValue;
		
		[Serial] public readonly bool OopsPointsVisible;
		[Serial] public readonly bool OopsPointsModifiable;
		[Serial] public readonly int OopsPointsValue;
		[Serial] public readonly int OopsPointsMinValue;
		[Serial] public readonly int OopsPointsMaxValue;
		
		[Serial] public readonly bool FlagDeliveredPointsVisible;
		[Serial] public readonly bool FlagDeliveredPointsModifiable;
		[Serial] public readonly int FlagDeliveredPointsValue;
		[Serial] public readonly int FlagDeliveredPointsMinValue;
		[Serial] public readonly int FlagDeliveredPointsMaxValue;
		
		[Serial] public readonly bool FlagLostPointsVisible;
		[Serial] public readonly bool FlagLostPointsModifiable;
		[Serial] public readonly int FlagLostPointsValue;
		[Serial] public readonly int FlagLostPointsMinValue;
		[Serial] public readonly int FlagLostPointsMaxValue;

		[Serial] public readonly bool FlagRescuedPointsVisible;
		[Serial] public readonly bool FlagRescuedPointsModifiable;
		[Serial] public readonly int FlagRescuedPointsValue;
		[Serial] public readonly int FlagRescuedPointsMinValue;
		[Serial] public readonly int FlagRescuedPointsMaxValue;

		[Serial] public readonly bool AllowTeamSelection;
	}

	/// <summary>
	/// Stores the settings for a base game mode.
	/// </summary>
	[AutoSerializable] public class SettingsGuiSettings
	{
		[Serial, Optional] public bool LivesLimitEnable;
		[Serial, Optional] public int LivesLimit;

		[Serial, Optional] public bool ScoreLimitEnable;
		[Serial, Optional] public int ScoreLimit;

		[Serial, Optional] public bool TimeLimitEnable;
		[Serial, Optional] public int TimeLimit;

		[Serial, Optional] public bool TeamsEnable;
		[Serial, Optional] public int Teams;
		[Serial, Optional] public bool LockTeams;
		[Serial, Optional] public bool FriendlyFire;

		[Serial, Optional] public bool PlayerLimitEnable;
		[Serial, Optional] public int PlayerLimit;

		[Serial, Optional] public bool ForceRandomTanks;

		[Serial, Optional] public int RespawnDelay;

		[Serial, Optional] public int KillPoints;
		[Serial, Optional] public int DeathPoints;
		[Serial, Optional] public int OopsPoints;
		[Serial, Optional] public int FlagDeliveredPoints;
		[Serial, Optional] public int FlagLostPoints;
		[Serial, Optional] public int FlagRescuedPoints;
	}
}