// Walt Destler

using System;
using Halfling.Audio;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Doodads;
using TankyTank.Game.Tanks;
using TankyTank.Modes;

namespace TankyTank.Scenarios
{
	/// <summary>
	/// Raid On Metropolis client _game mode.
	/// </summary>
	public class ROMClient : CTFClient
	{
		#region Strings and Sounds

		/// <summary>
		/// Stores txt-loaded strings for the Raid On Metropolis scenario.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Strings
		{
			[Serial] public readonly string Victory;
			[Serial] public readonly string Defeat;
		}

		/// <summary>
		/// Stores txt-loaded sound effects for the client mode.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Sounds
		{
			[Serial] public readonly Effect MissionAccomplished;
			[Serial] public readonly Effect MissionFailed;
		}

		#endregion
		#region Private Fields

		private Strings strings;
		private Sounds sounds;
		private readonly string infoText;
		private int startingFlags;
		private int flagsCaptured = 0;

		private bool playingEndSound = false;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new ROMClient.
		/// </summary>
		public ROMClient(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			strings = ser.TreeDeserialize<Strings>(file, "Strings");
			sounds = ser.TreeDeserialize<Sounds>(file, "Sounds");
			infoText = ser.TreeDeserialize<string>(file, "InfoText");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called when the _game is started.
		/// </summary>
		public override void StartGame(IClientGame game, byte[] initData)
		{
			base.StartGame(game, initData);

			// Count number of flags.
			startingFlags = 0;
			foreach(IDoodad d in game.Doodads)
			{
				if(d.Rules == game.Rules.Doodads[StringHasher.Instance.GetPortableHash("CTFFlag2")])
					startingFlags++;
			}
		}

		/// <summary>
		/// Gets the _game end text.
		/// </summary>
		public override string GetEndText(GameEndCause cause, int endID)
		{
			IEffect endSound = null;
			string ret;

			switch(cause)
			{
			case GameEndCause.Mode:
				ret = strings.Defeat;
				endSound = sounds.MissionFailed;
				break;
			case GameEndCause.ScenarioCompleted:
				ret = strings.Victory;
				endSound = sounds.MissionAccomplished;
				break;
			default:
				throw new ApplicationException("Unknown _game end cause '" + cause + "'.");
			}

			if(endSound != null && !playingEndSound)
			{
				game.PlaySound(endSound, true);
				playingEndSound = true;
			}

			return ret;
		}

		/// <summary>
		/// Returns the text to be displayed on the left of the Infoboard.
		/// </summary>
		public override string GetInfoboardTextLeft()
		{
			return null;
		}

		/// <summary>
		/// Returns the text to be displayed on the right of the Infoboard.
		/// </summary>
		public override string GetInfoboardTextRight()
		{
			return null;
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Called when a flag is delivered.
		/// </summary>
		protected override void onFlagDelivered(ITank tank)
		{
			base.onFlagDelivered(tank);

			flagsCaptured++;
			//printInfoText();
		}

		/*/// <summary>
		/// Prints the current info text.
		/// </summary>
		private void printInfoText()
		{
			_game.InfoBoxVisible = true;
			_game.InfoBoxOkayVisible = false;
			_game.InfoBoxText = string.Format(infoText, flagsCaptured, startingFlags);
		}*/

		/// <summary>
		/// Returns the text to display in the info box, or null for no text.
		/// </summary>
		protected override string getInfoText()
		{
			return string.Format(infoText, flagsCaptured, startingFlags, base.getInfoText());
		}

		#endregion
	}
}