// Walt Destler

using System;
using Halfling.Audio;
using Halfling.Geometry.Dim2;
using Halfling.Graphics;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.Game;
using TankyTank.Game.Bullets;
using TankyTank.Game.Doodads;
using TankyTank.Game.Tanks;
using TankyTank.Modes;

namespace TankyTank.Scenarios
{
	/// <summary>
	/// Client _game mode for the Tutorial scenario.
	/// </summary>
	public class TutorialClient : BaseClient
	{
		#region Strings and Sounds

		/// <summary>
		/// Stores txt-loaded strings for the Tutorial.
		/// </summary>
		[AutoSerializable, DeserializeOnce] private class Strings
		{
			[Serial] public readonly string TutorialCompleted;
		}
		
		#endregion
		#region Private Fields

		private Strings strings;

		private string infoBoxOkayText;

		private TutorialMessage[] tutorialMessages;
		private int curMessage = 0;

		private KillMessage[] userKillMessages;
		private int userKills = 0;

		private KillMessage[] instructorKillMessages;
		private int instructorKills = -1; // start at -1 because when user blows up to move to north island it counts as 1.

		private SoundInstance curSound = null;

		private bool finalPhase = false;
		private bool draw4Arrows = false;
		private bool draw1Arrow = false;

		private ISprite enemyArrow;
		private ISprite enemyRearArrow;
		private ISprite flagArrow;
		private ISprite uiArrow;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new Tutorial client _game mode.
		/// </summary>
		public TutorialClient(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			strings = ser.TreeDeserialize<Strings>(file, "Strings");

			// Load tutorial messages.
			tutorialMessages = ser.TreeDeserialize<TutorialMessage[]>(file, "TutorialMessages");
			userKillMessages = ser.TreeDeserialize<KillMessage[]>(file, "UserKillMessages");
			instructorKillMessages = ser.TreeDeserialize<KillMessage[]>(file, "InstructorKillMessages");
			infoBoxOkayText = ser.TreeDeserialize<string>(file, "InfoBoxOkayText");

			// Load arrow picture.
			enemyArrow = ser.TreeDeserialize<ISprite>(file, "EnemyArrow");
			enemyRearArrow = ser.TreeDeserialize<ISprite>(file, "EnemyRearArrow");
			flagArrow = ser.TreeDeserialize<ISprite>(file, "FlagArrow");
			uiArrow = ser.TreeDeserialize<ISprite>(file, "UIArrow");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called to start the _game.
		/// </summary>
		public override void StartGame(IClientGame game, byte[] initData)
		{
			base.StartGame(game, initData);

			// Show info box.
			game.InfoBoxOkayText = infoBoxOkayText;
			game.InfoBoxOkayVisible = true;
			game.InfoBoxVisible = true;
			showCurMessage();
		}

		/// <summary>
		/// Called every frame to update mode logic.
		/// </summary>
		public override void Update()
		{
			base.Update();

			// Check for special events.
			switch(tutorialMessages[curMessage].Special)
			{
			case "getflags": specialGetFlags(); break;
			case "fireprimary": specialFirePrimary(); break;
			case "firesecondary": specialFireSecondary(); break;
			case "maketank0":
				game.MessageServerMode(ser.BinarySerialize("maketank0"));
				showNextMessage();
				draw4Arrows = true;
				break;
			case "maketank2":
				game.MessageServerMode(ser.BinarySerialize("maketank2"));
				showNextMessage();
				draw1Arrow = true;
				break;
			case "makedrone0":
				game.MessageServerMode(ser.BinarySerialize("makedrone0"));
				showNextMessage();
				break;
			case "damageuser":
				game.MessageServerMode(ser.BinarySerialize("damageuser"));
				showNextMessage();
				break;
			case "makedrone2":
				game.MessageServerMode(ser.BinarySerialize("makedrone2"));
				showNextMessage();
				break;
			case "makepowerup":
				game.MessageServerMode(ser.BinarySerialize("makepowerup"));
				showNextMessage();
				break;
			case "finalphase":
				game.MessageServerMode(ser.BinarySerialize("finalphase"));
				game.SetPlayerScore(game.UserTank.Player, 0);
				finalPhase = true;
				showNextMessage();
				break;
			}
		}

		/// <summary>
		/// Adds any extra tutorial sprites.
		/// </summary>
		public override void AddSprites()
		{
			base.AddSprites();

			const double arrowDist = 64;

			// Draw arrows on tanks.
			if(draw4Arrows)
			{
				ITank enemy = findEnemyTank();
				if(enemy != null)
				{
					Vector loc;

					loc = enemy.Location + new Vector(-arrowDist, 0);
					addArrowSprite(enemyArrow, loc, loc.DirectionTo(enemy.Location));

					loc = enemy.Location + new Vector(0, -arrowDist);
					addArrowSprite(enemyArrow, loc, loc.DirectionTo(enemy.Location));

					loc = enemy.Location + new Vector(arrowDist, 0);
					addArrowSprite(enemyArrow, loc, loc.DirectionTo(enemy.Location));

					loc = enemy.Location + new Vector(0, arrowDist);
					addArrowSprite(enemyArrow, loc, loc.DirectionTo(enemy.Location));
				}
			}
			if(draw1Arrow)
			{
				ITank enemy = findEnemyTank();
				if(enemy != null)
				{
					Vector loc = enemy.Location + Vector.FromAngle(enemy.ChassisRot) * -arrowDist;
					addArrowSprite(enemyRearArrow, loc, enemy.ChassisRot);
				}
			}

			// Draw arrows on flags.
			if(tutorialMessages[curMessage].Special == "getflags")
			{
				// Get closest flag.
				IDoodad closestFlag = null;
				double closestDistance = double.MaxValue;
				foreach(IDoodad doodad in game.Doodads)
				{
					if(doodad.Rules == game.Rules.Doodads[StringHasher.Instance.GetPortableHash("CTFFlag1")])
					{
						double d = game.UserTank.Location.DistanceSquaredTo(doodad.Location);
						if(d < closestDistance)
						{
							closestDistance = d;
							closestFlag = doodad;
						}
					}
				}

				// Draw arrow on flags.
				if(closestFlag != null)
				{
					Vector loc = closestFlag.Location + Vector.FromAngle(closestFlag.Location.DirectionTo(game.UserTank.Location)) * arrowDist;
					addArrowSprite(flagArrow, loc, loc.DirectionTo(closestFlag.Location));
				}
			}
		}

		/// <summary>
		/// Draws mode-specific UI elements.
		/// </summary>
		public override void DrawUI()
		{
			base.DrawUI();

			// Draw arrow on ammo bar.
			if(tutorialMessages[curMessage].Special == "showammo")
			{
				Rect ammoRect = game.GetAmmoRect();
				Vector loc = new Vector(
					ammoRect.Center.X,
					ammoRect.Top - uiArrow.Size.Y / 2);
				drawUIArrow(uiArrow, loc, (Direction)0);
			}

			// Draw arrow on health bar.
			else if(tutorialMessages[curMessage].Special == "showhealth")
			{
				Rect healthRect = game.GetHealthRect();
				Vector loc = new Vector(
					healthRect.Center.X,
					healthRect.Top - uiArrow.Size.Y / 2);
				drawUIArrow(uiArrow, loc, (Direction)0);
			}

			// Draw arrow on minimap.
			else if(tutorialMessages[curMessage].Special == "showminimap")
			{
				Rect mmRect = game.GetMinimapRect();
				Vector loc = new Vector(
					mmRect.Center.X,
					mmRect.Top - uiArrow.Size.Y / 2);
				drawUIArrow(uiArrow, loc, (Direction)0);
			}

			// Draw arrows on reload indicators.
			else if(tutorialMessages[curMessage].Special == "showreload")
			{
				Rect pRect = game.GetPrimaryReloadRect();
				Rect sRect = game.GetSecondaryReloadRect();
				Vector pLoc = new Vector(
					pRect.Left - uiArrow.Size.X / 2,
					pRect.Center.Y);
				Vector sLoc = new Vector(
					sRect.Right + uiArrow.Size.X / 2,
					sRect.Center.Y);
				drawUIArrow(uiArrow, pLoc, (Direction)(3 * Math.PI / 2));
				drawUIArrow(uiArrow, sLoc, (Direction)(Math.PI / 2));
			}

			// Draw arrow on score box.
			else if(tutorialMessages[curMessage].Special == "showscore")
			{
				Rect scoreRect = game.GetScoreTextRect();
				Vector loc = new Vector(
					scoreRect.Center.X,
					scoreRect.Top - uiArrow.Size.Y / 2);
				drawUIArrow(uiArrow, loc, (Direction)0);
			}
		}

		/// <summary>
		/// Called when a tank is killed.
		/// </summary>
		public override void OnTankKilled(ITank tank, bool killedByZero)
		{
			switch(tutorialMessages[curMessage].Special)
			{
			case "killtank":
				if(tank != game.UserTank)
				{
					showNextMessage();
					draw4Arrows = false;
					draw1Arrow = false;
				}
				break;
			}

			// Show kill messages?
			if(finalPhase)
			{
				if(tank != game.UserTank)
				{
					showKillMessage(userKillMessages[userKills]);
					userKills++;
				}
				else
				{
					if(instructorKills >= 0)
						showKillMessage(instructorKillMessages[instructorKills]);
					instructorKills++;
				}
			}
		}

		/// <summary>
		/// Called when a doodad is collected.
		/// </summary>
		public override void OnDoodadCollected(ITank tank, IDoodad doodad)
		{
			switch(tutorialMessages[curMessage].Special)
			{
			case "getammo":
				if(doodad.Rules == game.Rules.Doodads[StringHasher.Instance.GetPortableHash("Ammo")])
					showNextMessage();
				break;
			case "gethealth":
				if(doodad.Rules == game.Rules.Doodads[StringHasher.Instance.GetPortableHash("Health")])
					showNextMessage();
				break;
			}
		}

		/// <summary>
		/// Called when the user clicks the okay button in the info box.
		/// </summary>
		public override void OnInfoBoxOkayClicked()
		{
			showNextMessage();
		}

		/// <summary>
		/// Gets the _game end text.
		/// </summary>
		public override string GetEndText(GameEndCause cause, int endID)
		{
			return strings.TutorialCompleted;
		}

		/// <summary>
		/// Returns the text to be displayed on the left of the Infoboard.
		/// </summary>
		public override string GetInfoboardTextLeft()
		{
			return null;
		}

		/// <summary>
		/// Returns the text to be displayed on the right of the Infoboard.
		/// </summary>
		public override string GetInfoboardTextRight()
		{
			return null;
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Finds the enemy tank.
		/// </summary>
		private ITank findEnemyTank()
		{
			foreach(ITank tank in game.Tanks)
			{
				if(tank != game.UserTank)
					return tank;
			}

			return null;
		}

		/// <summary>
		/// Draws an arrow at the specified location and direction.
		/// </summary>
		private void addArrowSprite(ISprite pic, Vector loc, Direction dir)
		{
			byte alpha = (byte)((Math.Sin((double)game.ClockTime / 200) + 1) / 2 * 155 + 100);

			DepthLevel dl = new DepthLevel(0);

			Rect r = new Rect();
			r.Size = pic.Size;
			r.Center = loc;

			Quad q = new Quad(
				r,
				dl.ToRenderable(),
				0,
				0,
				dir + Direction.FromDegrees(90),
				new Vector(.5, .5),
				pic.Color * new IntColor(alpha, 255, 255, 255),
				pic.Source);

			game.AddSprite(q, pic, dl);
		}

		/// <summary>
		/// Draws an arrow at the specified location and direction on top of the user interface..
		/// </summary>
		private void drawUIArrow(ISprite pic, Vector loc, Direction dir)
		{
			byte alpha = (byte)((Math.Sin((double)game.ClockTime / 125) + 1) / 2 * 155 + 100);

			Rect r = new Rect();
			r.Size = pic.Size;
			r.Center = loc;

			Quad quad = new Quad(
				r,
				0,
				0,
				0,
				dir + Direction.FromDegrees(180),
				new Vector(.5, .5),
				pic.Color * new IntColor(alpha, 255, 255, 255),
				pic.Source);

			game.Draw(ref quad, pic);
		}

		/// <summary>
		/// Shows the next message.
		/// </summary>
		private void showNextMessage()
		{
			curMessage = (curMessage + 1) % tutorialMessages.Length;
			showCurMessage();
		}

		/// <summary>
		/// Shows the current message.
		/// </summary>
		private void showCurMessage()
		{
			game.InfoBoxText = tutorialMessages[curMessage].Message;
			game.InfoBoxOkayVisible = tutorialMessages[curMessage].ShowNext;

			if(curSound != null && !curSound.IsStopped)
			{
				curSound.Stop();
				curSound = null;
			}

			if(tutorialMessages[curMessage].Effect != null)
				curSound = game.PlaySound(tutorialMessages[curMessage].Effect, true);
		}

		/// <summary>
		/// Shows a kill message.
		/// </summary>
		private void showKillMessage(KillMessage km)
		{
			game.InfoBoxText = km.Message;
			game.InfoBoxOkayVisible = false;

			if(curSound != null && !curSound.IsStopped)
			{
				curSound.Stop();
				curSound = null;
			}

			if(km.Effect != null)
				curSound = game.PlaySound(km.Effect, true);
		}

		/// <summary>
		/// Checks to see if the user has collected all of the flags.
		/// </summary>
		private void specialGetFlags()
		{
			foreach(IDoodad d in game.Doodads)
			{
				if(d.Rules == game.Rules.Doodads[StringHasher.Instance.GetPortableHash("CTFFlag1")])
					return;
			}

			showNextMessage();
		}

		/// <summary>
		/// Checks to see if the user fired his primary weapon.
		/// </summary>
		private void specialFirePrimary()
		{
			foreach(IBullet b in game.Bullets)
			{
				if(b.UniqueID.Tank == game.UserTank.UniqueID)
				{
					foreach(TurretRules tr in game.UserTank.Rules.Turrets)
					{
						if(tr.Bullet == b.Rules && tr.FireButton == FireButton.MouseLeft)
						{
							showNextMessage();
							return;
						}
					}
				}
			}
		}

		/// <summary>
		/// Checks to see if the user fired his secondary weapon.
		/// </summary>
		private void specialFireSecondary()
		{
			foreach(IBullet b in game.Bullets)
			{
				if(b.UniqueID.Tank == game.UserTank.UniqueID)
				{
					foreach(TurretRules tr in game.UserTank.Rules.Turrets)
					{
						if(tr.Bullet == b.Rules && tr.FireButton == FireButton.MouseRight)
						{
							showNextMessage();
							return;
						}
					}
				}
			}
		}

		#endregion
	}

	/// <summary>
	/// Defines a message shown to the user.
	/// </summary>
	[AutoSerializable] struct TutorialMessage
	{
		[Serial] public readonly string Message;
		[Serial, Optional(null)] public readonly IEffect Effect;
		[Serial] public readonly bool ShowNext;
		[Serial, Optional("")] public readonly string Special;
	}

	/// <summary>
	/// Defines a message shown to the user when the user kills the instructor or vice-versa.
	/// </summary>
	[AutoSerializable] struct KillMessage
	{
		[Serial] public readonly string Message;
		[Serial, Optional(null)] public readonly IEffect Effect;
	}
}