// Walt Destler

using System;
using System.Drawing;
using Halfling.Geometry.Dim2;
using Halfling.Serialization;
using Halfling.TreeData;
using TankyTank.AI;
using TankyTank.AIBots;
using TankyTank.Game.Doodads;
using TankyTank.Game.Tanks;
using TankyTank.Modes;

namespace TankyTank.Scenarios
{
	/// <summary>
	/// Server _game mode for the Tutorial scenario.
	/// </summary>
	public class TutorialServer : BaseServer
	{
		#region Private Fields

		private readonly INode droneNode;
		private readonly INode instructorNode;
		private readonly string dronePlayerName;
		private readonly string instructorPlayerName;

		private int userTankID;
		private bool finalPhase = false;

		#endregion
		#region Constructors

		/// <summary>
		/// Creates a new Tutorial server _game mode.
		/// </summary>
		public TutorialServer(TreeFile file, Serializer ser)
			: base(file, ser)
		{
			droneNode = file.FindAtPath("Drone");
			instructorNode = file.FindAtPath("Instructor");
			dronePlayerName = ser.TreeDeserialize<string>(file, "DronePlayerName");
			instructorPlayerName = ser.TreeDeserialize<string>(file, "InstructorPlayerName");
		}

		#endregion
		#region Public Methods

		/// <summary>
		/// Called to spawn the player.
		/// </summary>
		public override void SpawnNewPlayer(int owner, string name, System.Drawing.Color color, int player, int team, int preferredTank)
		{
			int type = 0; // Assault tank.
			game.CreateNewPlayer(player, owner, name, color, team, false, 0);
			TankPos pos = findSpawnLoc(1);
			userTankID = game.SpawnNewTank(type, owner, color, player, team, pos);
		}

		/// <summary>
		/// Called when a message from the client mode is received.
		/// </summary>
		public override void ReceiveClientModeMessage(byte[] message, int client)
		{
			string special = ser.BinaryDeserialize<string>(message);
			switch(special)
			{
			case "maketank0": specialMakeTank(0); break;
			case "maketank2": specialMakeTank(2); break;
			case "makedrone0": specialMakeDrone(0); break;
			case "makedrone2": specialMakeDrone(2); break;
			case "makepowerup": specialMakePowerup(); break;
			case "finalphase": specialFinalPhase(); break;
			case "damageuser": specialDamageUser(); break;
			}
		}

		/// <summary>
		/// Called when a tank is killed.
		/// </summary>
		public override void OnTankKilled(ITank tank, bool killedByZero)
		{
			if(!finalPhase)
			{
				if(tank.UniqueID == userTankID)
					userTankID = game.RespawnTank(tank, 0, findSpawnLoc(1));
			}
			else
			{
				base.OnTankKilled(tank, killedByZero);
			}
		}

		/// <summary>
		/// Called to modify tank damage.
		/// </summary>
		public override int ModifyDamage(int damage, int victimTank, int sourceTank, int sourcePlayer, int sourceTeam)
		{
			damage = base.ModifyDamage(damage, victimTank, sourceTank, sourcePlayer, sourceTeam);
			if(victimTank == userTankID && !finalPhase)
			{
				return Math.Min(
					damage,
					game.Tanks[userTankID].Health - 1000);
			}
			else
			{
				return damage;
			}
		}

		#endregion
		#region Non-Public Methods

		/// <summary>
		/// Returns whether the scenario was successfully completed.
		/// </summary>
		protected override bool wasScenarioCompleted()
		{
			return true;
		}

		/// <summary>
		/// Finds the dummy's spawn loaction.
		/// </summary>
		private TankPos findSpawnLoc(int num)
		{
			foreach(IDoodad doodad in game.Doodads)
			{
				if(doodad.Rules.SpawnPoint && doodad.Rules.SpawnTeam == num)
				{
					Direction dir;
					if(game.Tanks.Exists(userTankID))
						dir = doodad.Location.DirectionTo(game.Tanks[userTankID].Location);
					else
						dir = (Direction)Math.PI;

					return new TankPos(doodad.Location, Vector.Zero, dir, 0, doodad.Location);
				}
			}

			throw new ApplicationException("Unable to find spawn point " + num + ".");
		}

		/// <summary>
		/// Finds the powerup's spawn loaction.
		/// </summary>
		private Vector findPowerupSpawnLoc()
		{
			foreach(IDoodad doodad in game.Doodads)
			{
				if(doodad.Rules.SpawnPoint && doodad.Rules.SpawnTeam == 8)
					return doodad.Location;
			}

			throw new ApplicationException("Unable to find spawn point 2.");
		}

		/// <summary>
		/// Creates a dummy tank on the island.
		/// </summary>
		private void specialMakeTank(int type)
		{
			TankPos pos = findSpawnLoc(2);
			game.SpawnNewTank(type, game.Owner, Color.White, 0, 0, pos);
		}

		/// <summary>
		/// Creates a dummy tank on the island.
		/// </summary>
		private void specialMakeDrone(int type)
		{
			// Create player.
			int player;
			game.CreateNewPlayer(player = game.Rand.Int32(), game.Owner, dronePlayerName, Color.White, 0, true, type);

			// Spawn tank.
			TankPos pos = findSpawnLoc(3);
			int tank = game.SpawnNewTank(type, game.Owner, Color.White, player, 0, pos);
			
			// Create AI bot.
			IBot bot = new CustomBehaviorBot(game, droneNode, ser);
			game.AttachAIBot(tank, bot);
		}

		/// <summary>
		/// Creates a new powerup.
		/// </summary>
		private void specialMakePowerup()
		{
			game.SpawnNewDoodad(
				StringHasher.Instance.GetPortableHash("Firepower"),
				findPowerupSpawnLoc(),
				false,
				0,
				1,
				(Direction)0);
		}

		/// <summary>
		/// Starts the final phase of the _game.
		/// </summary>
		private void specialFinalPhase()
		{
			finalPhase = true;

			// Kill player.
			ITank userTank = game.Tanks[userTankID];
			game.DamageTank(userTank.UniqueID, userTank.Health, 0, 0, 0, Vector.Zero);
			// Player should respawn normally on the top island.

			// Reset player's score.
			game.SetPlayerScore(userTank.Player, 0);

			// Create instructor player.
			int instructor;
			int instructorTankType = 4;
			Color instructorColor = Color.Red;
			game.CreateNewPlayer(instructor = game.Rand.Int32(), game.Owner, instructorPlayerName, instructorColor, 0, true, instructorTankType);

			// Create instructor tank.
			TankPos pos = getNewTankPos(instructor, 0, instructorTankType);
			int tank = game.SpawnNewTank(instructorTankType, game.Owner, instructorColor, instructor, 0, pos);

			// Create instructor AI bot.
			IBot bot = new CustomBehaviorBot(game, instructorNode, ser);
			game.AttachAIBot(tank, bot);
		}

		/// <summary>
		/// Inflicts a point of damage on the user.
		/// </summary>
		private void specialDamageUser()
		{
			int damage = ModifyDamage(1, userTankID, 0, 0, 0);
			game.DamageTank(
				userTankID,
				damage,
				0,
				0,
				0,
				Vector.Zero);
		}

		/// <summary>
		/// Gets a new tank type.
		/// </summary>
		protected override int getNewTankType(int player, int team, bool forceRandom)
		{
			if(game.Players.Exists(player) && game.GetPreferredTank(player) != -1)
				return game.GetPreferredTank(player);
			else
				return game.Rand.Int32(0, game.Rules.Tanks.Length - 1);
		}

		#endregion
	}
}